#!/bin/bash
#This script expects the master's noderestore nfs export to be mounted at /drive


# check that nfs mount is present before continuing
grep drive /proc/mounts # | egrep -q 'nfs|bind'
if [ $? -ne 0 ] ; then
  cat > /tmp/backup-progress-temp.html <<EOF
<HTML>
<HEAD>
<TITLE>Backup failed.</TITLE>
</HEAD>
<BODY>The backup has failed due to a missing nfs mount.  /drive should be nfs
mounted from the master's backup repository</BODY>
</HTML>
EOF
  MASTER=`head -n1 /etc/nodes`
  rcp /tmp/backup-progress-temp.html $MASTER:/tmp/backup-progress.html
  exit 10
fi



. /drive/backups/$1/config

reporterror() {
cat > /tmp/backup-progress-temp.html <<EOF
<HTML>
<HEAD>
<TITLE>Backup failed</TITLE>
</HEAD>
<BODY>$1</BODY>
</HTML>
EOF
echo dev/shm > /etc/noderestore.exclude
if grep -q -i opensuse.*12 /etc/issue ; then
  echo run >> /etc/noderestore.exclude
  echo var/run >> /etc/noderestore.exclude
fi
rcp /tmp/backup-progress-temp.html $MASTER:/tmp/backup-progress.html
exit 1
}

rm /drive/backups/$1/done

cat >/tmp/backup-progress-temp.html <<EOF
<HTML>
<HEAD>
<TITLE>Backup in progress</TITLE>
</HEAD>
<BODY>The node backup is still in progress</BODY>
</HTML>
EOF
rcp /tmp/backup-progress-temp.html $MASTER:/tmp/backup-progress.html


partedprog=/usr/local/sbin/parted-static
partedscript="/drive/backups/$1/restorepartitions.sh"

rm -f /drive/backups/$1/mdadm.conf

mdconf="none"
[ -f /etc/mdadm.conf ] && mdconf="/etc/mdadm.conf"
[ -f /etc/mdadm/mdadm.conf ] && mdconf="/etc/mdadm/mdadm.conf"
if [ "$mdconf" != "none" ] ; then
  grep -q /dev/md $mdconf && cp $mdconf /drive/backups/$1/mdadm.conf
fi

echo "#!/bin/bash" > $partedscript
echo "export PATH=\$PATH:/bin:/sbin:/usr/bin:/usr/sbin" >> $partedscript
for device in $DEVICES ; do
   $partedprog -s $device p || reporterror "Failed to retrieve partition table for $device"

  if $partedprog -s $device p | grep -q gpt ; then

    echo "parted -s $device mklabel gpt" >> $partedscript

    $partedprog $device unit kb p | grep primary | \
    while read num start end size fstype primary flags ; do
      [ $fstype == "linux-swap" ] && \
        echo "parted -s $device mkpart primary linux-swap $start $end" >> $partedscript || \
        echo "parted -s $device mkpart primary ext3 $start $end" >> $partedscript
    done
    echo "parted -s $device set 1 boot on" >> $partedscript
    echo "hdparm -z $device" >> $partedscript
    chmod 755 $partedscript

  elif $partedprog -s $device p | grep -q msdos ; then

    echo "parted -s $device mklabel msdos" >> $partedscript

    $partedprog $device unit kb p | egrep "primary|extended|logical" | grep -v Sector | \
    while read num start end size primary fstype flags ; do
      if [ $primary == "extended" ] ; then
        echo "parted -s $device mkpart $primary $start $end" >> $partedscript
        echo $fstype $flags | grep -q raid && \
          echo "parted -s $device set $num raid on"  >> $partedscript
      elif [ "x$fstype" == "xlinux-swap" ] ; then
        echo "parted -s $device mkpart $primary linux-swap $start $end" >> $partedscript
      else
        echo "parted -s $device mkpart $primary ext3 $start $end" >> $partedscript
        echo $fstype $flags | grep -q raid && \
          echo "parted -s $device set $num raid on"  >> $partedscript
      fi
    done
    echo "parted -s $device set 1 boot on" >> $partedscript
    echo "hdparm -z $device" >> $partedscript

  fi
done

chmod 755 $partedscript

[ -f /etc/raidtab ] && ( cp /etc/raidtab /drive/backups/$1/ || rwwerror "Unable to backup raidtab" )
cp /etc/fstab /drive/backups/$1/ || rwwerror "Unable to backup fstab"

cd /
TARLIST=`ls | grep -w -v proc | grep -w -v sys | grep -w -v drive | grep -w -v media`

#exclude /run and /var/run to prevent errors - these are dynamically generated
#on opensuse 12.x
if grep -q -i opensuse.*12 /etc/issue ; then
  echo run >> /etc/noderestore.exclude
  echo var/run >> /etc/noderestore.exclude
fi

# generate list of nfs mounted filesystems to be excluded
mount | egrep "type\ nfs|type\ lustre" | grep -v -w /drive | awk '{print $3}' |\
sed 's/^\/*//g' | while read nfsexclude
 do
  [ "x$nfsexclude" == "x" ] || echo $nfsexclude >> /etc/noderestore.exclude
 done

#Handle udev persistent names for eth devices
if [ -f /etc/udev/rules.d/30-net_persistent_names.rules ] ; then
  grep -v SUBSYSTEM /etc/udev/rules.d/30-net_persistent_names.rules > /etc/udev/rules.d/30-net_persistent_names.rules-emptied
  cp /etc/udev/rules.d/30-net_persistent_names.rules{,-backup}
  mv /etc/udev/rules.d/30-net_persistent_names.rules-emptied /etc/udev/rules.d/30-net_persistent_names.rules
elif [ -f /etc/udev/rules.d/70-persistent-net.rules ] ; then
  grep -v SUBSYSTEM /etc/udev/rules.d/70-persistent-net.rules > /etc/udev/rules.d/70-persistent-net.rules-emptied
  cp /etc/udev/rules.d/70-persistent-net.rules{,-backup}
  mv /etc/udev/rules.d/70-persistent-net.rules-emptied /etc/udev/rules.d/70-persistent-net.rules
fi

if [ -f /etc/udev/rules.d/70-persistent-cd.rules ] ; then
  grep -v SUBSYSTEM /etc/udev/rules.d/70-persistent-cd.rules > /etc/udev/rules.d/70-persistent-cd.rules-emptied
  cp /etc/udev/rules.d/70-persistent-cd.rules{,-backup}
  mv /etc/udev/rules.d/70-persistent-cd.rules-emptied /etc/udev/rules.d/70-persistent-cd.rules
fi

tar --create --gzip --file=/drive/backups/$1/backup.tgz --sparse --numeric-owner \
    --anchored --preserve --exclude-from /etc/noderestore.exclude $TARLIST >& /tmp/taroutput.txt

if [ $? -ne 0 ] ; then
  taroutput=`cat /tmp/taroutput.txt`
  #put udev rules back in place for suse10 now that backup failed
  [ -f /etc/udev/rules.d/30-net_persistent_names.rules-backup ] && mv /etc/udev/rules.d/30-net_persistent_names.rules-backup /etc/udev/rules.d/30-net_persistent_names.rules
  [ -f /etc/udev/rules.d/70-persistent-net.rules-backup ] && mv /etc/udev/rules.d/70-persistent-net.rules-backup /etc/udev/rules.d/70-persistent-net.rules
  [ -f /etc/udev/rules.d/70-persistent-cd.rules-backup ] && mv /etc/udev/rules.d/70-persistent-cd.rules-backup /etc/udev/rules.d/70-persistent-cd.rules
  reporterror "Tar exitted with an error.  The output is as follows:<BR>$taroutput"
fi

#put udev rules back in place for suse10 now that backup is done
[ -f /etc/udev/rules.d/30-net_persistent_names.rules-backup ] && mv /etc/udev/rules.d/30-net_persistent_names.rules-backup /etc/udev/rules.d/30-net_persistent_names.rules
[ -f /etc/udev/rules.d/70-persistent-net.rules-backup ] && mv /etc/udev/rules.d/70-persistent-net.rules-backup /etc/udev/rules.d/70-persistent-net.rules
[ -f /etc/udev/rules.d/70-persistent-cd.rules-backup ] && mv /etc/udev/rules.d/70-persistent-cd.rules-backup /etc/udev/rules.d/70-persistent-cd.rules
rsh $MASTER /usr/local/sbin/preparerestore.sh /drive/backups/$1/config $1

touch /drive/backups/$1/done
rm /drive/backups/current
ln -s $1 /drive/backups/current

rm /etc/noderestore.exclude

cat >/tmp/backup-progress-temp.html <<EOF
<HTML>
<HEAD>
<TITLE>Backup completed</TITLE>
</HEAD>
<BODY>The node backup in $1 has completed successfully</BODY>
</HTML>
EOF
rcp /tmp/backup-progress-temp.html $MASTER:/tmp/backup-progress.html
