#!/bin/bash

myhostname=`hostname -s`
SERIALDEV="ttyS1"

dmesg -s 1000000 | grep -q ttyS2 && SERIALDEV="ttyS2"
if [ $SERIALDEV == "ttyS1" ] ; then
  SERIALNUM=1
else
  SERIALNUM=2
fi

### grub portion

if [ -f /boot/grub/grub.cfg -o -f /boot/grub2/grub.cfg ] ; then     ## grub2
  if [ -f /etc/default/grub ] ; then
    if grep -q console=${SERIALDEV} /etc/default/grub ; then
      echo "/etc/default/grub already has console=${SERIALDEV} in it."
    else
      sed -i /etc/default/grub -e 's/^\(GRUB_CMDLINE_LINUX=\"[^\"]*\)\"/\1 console='${SERIALDEV}',115200n8 console=tty1\"/'
    fi
    if grep -q GRUB_TERMINAL=console /etc/default/grub ; then
      echo "/etc/default/grub already has GRUB_TERMINAL=console in it."
    else
      echo "GRUB_TERMINAL=console" >> /etc/default/grub
    fi
    if [ -x /usr/sbin/update-grub2 ] ; then 
      /usr/sbin/update-grub2
    elif [ -x /sbin/grub2-mkconfig ] ; then
      /sbin/grub2-mkconfig > /boot/grub2/grub.cfg
    elif [ -x /usr/sbin/grub2-mkconfig ] ; then
      /usr/sbin/grub2-mkconfig > /boot/grub2/grub.cfg
    else
      echo Could not find method to update grub, exiting.
      exit 1
    fi
  else
    echo "MCMS does not know how to add console=${SERIALDEV} on ${myhostname} to grub2 since /etc/default/grub does not exist."
  fi
elif [ -f /boot/grub/grub.conf -o -f /boot/grub/menu.lst ] ; then   #grub1
  GRUBCONF=/boot/grub/menu.lst
  [ -f /boot/grub/menu.lst ] || GRUBCONF=/boot/grub/grub.conf
  [ -h /boot/grub/menu.lst ] && GRUBCONF=/boot/grub/grub.conf
 
  if grep -q console=${SERIALDEV} ${GRUBCONF} ; then
    echo "${GRUBCONF} on ${myhostname} already has console=${SERIALDEV} in it."
  else
      sed -i ${GRUBCONF} -e "s/\(kernel.*vmlinuz.*\)/\1\ console=${SERIALDEV},115200n8\ console=tty1/"
  fi

else  #can't find a grub config
    myhostname=`hostname -s`
    echo "**********Can't find a grub config on ${myhostname} to add serial console too.  Continuing anyway.**********"
fi
  

### securetty
grep -q ${SERIALDEV} /etc/securetty || echo ${SERIALDEV} >> /etc/securetty



### console login

[ -x /sbin/agetty ] && GETTY="/sbin/agetty" || GETTY="/sbin/getty"

if egrep -q "openSUSE 12.1|openSUSE 12.2" /etc/issue ; then #systemd
  sed -i /lib/systemd/system/getty\@.service -e s/38400/115200/
  ln -s /lib/systemd/system/getty\@.service /etc/systemd/system/getty.target.wants/getty\@${SERIALDEV}.service

elif grep -q "openSUSE 12.3" /etc/issue ; then #systemd different directory and new file
  cp /usr/lib/systemd/system/getty\@.service /usr/lib/systemd/system/gettyserial\@.service
  sed -i /usr/lib/systemd/system/gettyserial\@.service -e "s/%I 38400/-L -w %I 115200/"
  ln -s /usr/lib/systemd/system/gettyserial\@.service /etc/systemd/system/getty.target.wants/getty\@${SERIALDEV}.service

elif grep -q "Fedora release 17" /etc/issue ; then #systemd
  sed -i /lib/systemd/system/getty\@.service -e s/38400/115200/
  ln -s /lib/systemd/system/getty\@.service /etc/systemd/system/getty.target.wants/getty\@${SERIALDEV}.service


elif [ -d /etc/init -o -d /etc/event.d ] ; then   #upstart
  [ -d /etc/init ] && UPDIR="/etc/init" || UPDIR="/etc/event.d"
  
  if [ -f ${UPDIR}/${SERIALDEV}.conf ] ; then
    echo "${myhostname} already has a ${SERIALDEV}.conf file for upstart, skipping."
  else
    cat > ${UPDIR}/${SERIALDEV}.conf << EOFSERIAL
# ${SERIALDEV} - getty
#
# This service maintains a getty on ${SERIALDEV} from the point the system is
# started until it is shut down again.

start on runlevel [2345]
stop on runlevel [!2345]

respawn
exec ${GETTY} -w -L 115200 ${SERIALDEV}
EOFSERIAL
  fi

else   #regular inittab
  if grep -q ^[^#].*getty.*${SERIALDEV} /etc/inittab ; then
    echo "The inittab on ${myhostname} already has entry for running a getty on ${SERIALDEV}, not modifying"
  else
    sed -i /etc/inittab -e /tty6/aS${SERIALNUM}:2345:respawn:${GETTY}\ -w\ -L\ 115200\ ${SERIALDEV}\ vt102
  fi
fi

