<?php
// Service functions for Microway Cluster Control Library.
//
// Copyright (c) 2002 by Microway Inc., All rights reserved.
//
//
// Major modifications and rewrites by Rick Warner, from 2002-2007
//
// Modifications:
//    20Sep02 MJNewman: Report errors from command execution to output of command.
//    13Sep02 MJNewman: Add default nodes to commandNodes
include_once ('microwaytoganglia.php');


// Header elements appropriate to a PHP generated page.
//

function microwayHeaderStuff () {
    header ("Expires: Mon, 26 Jul 1997 05:00:00 GMT");  // Date in the past
    header ("Last-Modified: ".gmdate ("D, d M Y H:i:s")." GMT");        // always modified
    header ("Cache-Control: no-cache, must-revalidate");        // HTTP/1.1
    header ("Pragma: no-cache");        // HTTP/1.0
};


// Emit a Select list of nodes of this cluster
//
// Inputs:
//      defaultNode     name of the default node no default if blank
//      includeAll      true to include a psudo node 'All'
//      selectName      name of this selection
// Returns:
//      <SELECT> list of nodes
function microwayNodeSelect ($defaultNode, $includeAll, $includeMaster, $selectName) {
    $result = "<SELECT name=$selectName>\n";
    if ($includeAll) {
        $result .= "<option VALUE=All" . (($defaultNode != "All") ? "" : " SELECTED") . ">All</option>\n";
    };
    $mastername = "invalid-host name entry";
    if (!$includeMaster)
        $mastername = exec ("hostname -f");
    $hostList = mw2gHostsUpOfCluster();
    if (is_array ($hostList)) {
        foreach ($hostList as $k => $v) {
            if ($includeMaster or $k != $mastername) {
                $result .= "<OPTION VALUE=\"$k\"" . (($defaultNode != $k) ? "" : " SELECTED") . ">$k\n";
            }
        };
    };
    $result .= "</SELECT>";
    return $result;
}

function microwaySlaveSelect ($defaultNode, $includeAll, $selectName) {
    $result = "<SELECT name=$selectName>\n";
    $result .= "<option VALUE=none" . (($defaultNode != "none") ? "" : " SELECTED")
        .">None</option>\n";

    if ($includeAll) {
        $result .= "<option VALUE=All".(($defaultNode != "All") ? "" : " SELECTED") . ">All slaves</option>\n";
    };
    $result .= "<option VALUE=Master".(($defaultNode != "Master") ? "" : " SELECTED") . ">Master</option>\n";
    $hostList = mw2gHostsUpOfCluster ();
    if (is_array ($hostList)) {
        foreach ($hostList as $k => $v) {
            $result .= "<OPTION VALUE=\"$k\"".(($defaultNode != $k) ? "" : " SELECTED") . ">$k\n";
        };
    };
    $result .= "</SELECT>";
    return $result;
}


// Emit Microway banner
//
function microwayBanner ($pageTitle) {
    echo "<table width=100%>\n";
    echo "<tr>\n";

    echo "<td width=20%><table width=100%>\n";
    echo "<tr>\n";
    echo "<td  width=5% align=center><big><big>$pageTitle</big></big></td>\n";
    echo "</tr>\n";
    echo "<tr><td>\n";

    if (isset ($_SESSION['currentUser']))
        $currentUser = $_SESSION['currentUser'];
    else
        $currentUser = "SearchableUserAccountEntry123";

    if (($currentUser == "") || ($currentUser == "SearchableUserAccountEntry123")) {
        echo "<tr><td align=center><big>Not Logged in.</big></td>\n";
        //echo "<tr><td width=5% align=center><big>Not Logged in.</big></td>\n";
    } else {
        echo "<tr><td align=center><big>Logged in as: <b>$currentUser</b></big></td>\n";
        echo "<tr><td align=center><font size=\"-1\"><A HREF=\"login.php?change=yes\">Change User</A>&nbsp;/&nbsp;<A HREF=\"login.php?change=yes\">Logout</A></font></td></tr>";
    };

    echo "</tr>\n";
    echo "</table></td>\n";

    echo "<td width=\"50%\" VALIGN=middle align=center>\n";
    echo "<A HREF=\"http://www.microway.com\">";
    echo "<IMG SRC=microway.gif ALT=\"Microway (R) Technology you can count on.\" BORDER=0><BR></A>\n";
    echo "</td>\n";

    echo "<td>\n";
    // Eventually pass in a function call that emits these top right buttons
    echo "<table>\n";
    echo "<tr><td>";
    $servername = $_SERVER['HTTP_HOST'];
    echo "<A HREF=\"http://$servername/ganglia/\">";
    echo "<IMG SRC=ganglia-button.jpg ALT=Ganglia border=0>";
    echo "</A>";
    echo "</td></tr>\n";
    echo "</table>\n";
    echo "</td>\n";

    echo "</tr>\n";
    echo "<tr><td colspan=10><hr></td></tr>\n";
    echo "</table>\n";
    return;
}

// Echo a cluster control menu as a series of paragraphs
function microwayClusterControlMenu () {
    echo "<br>\n";
    $servername = $_SERVER['HTTP_HOST'];
    echo "&nbsp;&nbsp;<a HREF=\"http://$servername/ganglia/\">Ganglia&nbsp;Monitoring</a>&nbsp;&nbsp;&nbsp;<BR><BR>\n";
    echo "&nbsp;&nbsp;<a HREF=clustercontrol.php>Cluster&nbsp;Control</a><BR><BR>\n";
    $usesNodewatch = $_SESSION['usesNodewatch'];
    $usesIPMI = $_SESSION['usesIPMI'];
    if ($usesNodewatch == "yes" || $usesIPMI == "yes") {
        echo "&nbsp;&nbsp;<a HREF=clusterstatus.php>Cluster&nbsp;Status</A><BR><BR>\n";
    }
    echo "&nbsp;&nbsp;<a HREF=useraccounts.php>Account&nbsp;Admin</A><BR><BR>\n";
    echo "&nbsp;&nbsp;<a HREF=nodebackup.php>Node&nbsp;Backup</A><BR><BR>\n";
    echo "&nbsp;&nbsp;<a HREF=noderestore.php>Node&nbsp;Restore</A><BR><BR>\n";
    echo "<br>\n";
    return;
}

// emit a visible control that causes the cluster to shutdown
function microwayShutdownCluster () {
    echo "<form ACTION=shutdown_cluster.php METHOD=get>";
    echo "<table cellpadding=0 cellspacing=0><tr>\n";
    echo "<td><input  TYPE=submit VALUE=\"Shutdown All Slaves\"></td>\n";
    echo "</tr></table>";
    echo "</form>\n";
    return;
}

// emit a visible control that causes the cluster to reboot
function microwayRebootCluster () {
    echo "<form ACTION=reboot_cluster.php METHOD=get>";
    echo "<table cellpadding=0 cellspacing=0><tr>";
    echo "<td><input  TYPE=submit VALUE=\"Reboot All Slaves\"></td>\n";
    echo "</tr></table>";
    echo "</form>";
    return;
}


// Returns the appropriate limit for the request
//inputs
//      node - node to report
//      param - the parameter whose limit we''re interested in
//      id - of the X number of params, which do we want
//      type - MIN / MAX
// ex -- node=1, param=FAN, id=1, type=MAX
//  returns value of this limit setting

function getLimitSetting ($node, $param, $id, $type) {
    global $microwayAlertLimits;
    $limits = $microwayAlertLimits[$node][$param.$type];
    if (isset ($limits[$id]))
        return $limits[$id];
    else
        return 0;
}

// Report status of a particular NODEWATCH element.
// Reported as a set of elements in sequence.
//
// Inputs:
//      node    node to report
//      name    name of element to report
//      count   at least this number of elements are put out, &nbsp; for others
//      places  decimal places for the result
//      before  string to place prior to each element
//      beforeHighlight string to place prior if value not between min and max
//      after   string to place after each element
//      values  true to report values, false reports names of elements
//      thisMin highlight before if < min
//      thisMax highlight before if > max
//      blank   report blank string if equal to this value
//      hblank  report highlighted blank string if equal to this value 
// Returns:
//      string in the form:
//              before value after
function microwayNodewatchElements ($node, $name, $count, $places, $before, $beforeHighlight, $beforeCritical, $after, $values, $blank, $hblank) {
    global $microwayNodewatchData;
    global $microwayAlertLimits;
    $result = "";
    $thisNode = $microwayNodewatchData[$node];
    $number = 0;
    if ($thisNode != "") {
        $thisName = $thisNode[$name];
        if (is_array ($thisName)) {
            foreach ($thisName as $k => $v) {
                $number += 1;
                if ($number > $count)
                    break;

                $min = getLimitSetting ($node, $name, $number - 1, 'MIN');
                $shutmin = getLimitSetting ($node, $name, $number - 1, 'SHUTMIN');
                $max = getLimitSetting ($node, $name, $number - 1, 'MAX');
                $shutmax = getLimitSetting ($node, $name, $number - 1, 'SHUTMAX');

                if ($values) {
                    if (($v < $shutmin) || ($v > $shutmax)) {
                        $result .= $beforeCritical;
                    } else if (($v == $hblank) || ((($v < $min) || ($v > $max)))) {
                        $result .= $beforeHighlight;
                    } else {
                        $result .= $before;
                    };
                    if (($v == $blank) || ($v == $hblank)) {
                        $result .= "0";
                    } else {
                        $result .= decimalPlaces ($v, $places);
                    };
                } else {
                    $result .= $before;
                    $result .= $k;
                };
                $result .= $after;
            };
        } else {
            $number += 1;
            $min = $thisMin;
            $max = $thisMax;
            if ($values) {
                if (($thisName == $hblank)
                    || ((is_numeric ($thisName) && (($thisName < $min) || ($thisName > $max)))
                        && (!($thisName == $blank)))) {
                    $result .= $beforeHighlight;
                } else {
                    $result .= $before;
                };
                if (($thisName == $blank) || ($thisName == $hblank)) {
                    $result .= "&nbsp;";
                } else {
                    $result .= decimalPlaces ($v, $thisName);
                };
                $result .= $after;
            } else {
                $result .= $before.$name.$after;
            };
        };
    } else {
        $number += 1;
        $result .= $before."&nbsp;".$after;
    };
    if ($count > 0)
        while ($number < $count) {
            $number += 1;
            $result .= $before."&nbsp;".$after;
        };
    return $result;
}


// echo a table showing cluster status including fan speeds and temperatures
//
// Inputs:
//    showControls    true to show buttons to power nodes
function microwayWatchStatus ($showControls) {
    global $nodewatchTM;
//  echo "<h2><a name=nodewatch>$nodewatchTM</h2>\n";


    global $microwayNodewatchDataFileName;
    global $microwayNodewatchConfigFileName;
    global $microwayNodewatchData;
    global $microwayAlertLimits;

    include_once ("microwayparsenodewatch.php");

    $microwayNodewatchData = array ();

    $iterations = 0;

//reads in fan labels, etc.
    microwayparseconfig ($microwayNodewatchConfigFileName);
    //$tempconfig = $microwayNodewatchData['10.0.0.1'];

//reads in current data from file
    $result = microwayparsenodewatch ($microwayNodewatchDataFileName);
    if ($result != "") {
        echo "<b>ERROR:</b> $result<br><br>\n";
    };

//////////////////////////////////////////////////////////////
/////break here for new loop /////////////////////////////////
//////////////////////////////////////////////////////////////
    $numfans = $numvolts = $numtemps = 0;
    $fanlabels = array ("", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "");
    $templabels = array ("", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "");
    $voltlabels = array ("", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "", "");
    $onepass = 0;

    foreach ($microwayNodewatchData as $node => $v) {
        $newconfig = 0;         //assumes data is in same format unless following tests fail

//check for changes in labels, etc- denotes a new table should be made
        if (($numfans != $v['NUMFANS']) || ($numtemps != $v['NUMTEMPS']) || ($numvolts != $v['NUMVOLTS'])) {
            $newconfig = 1;     //simple test- different number of fans, etc
        } else {                //actually compare label names if # of params is same
            for ($i = 0; $i < count ($v['FANNAMES']); $i++)
                if ($fanlabels[$i] != $v['FANNAMES'][$i])
                    $newconfig = 1;
            for ($i = 0; $i < count ($v['TEMPNAMES']); $i++)
                if ($templabels[$i] != $v['TEMPNAMES'][$i])
                    $newconfig = 1;
            for ($i = 0; $i < count ($v['VOLTNAMES']); $i++)
                if ($voltlabels[$i] != $v['VOLTNAMES'][$i])
                    $newconfig = 1;
        }

        $serialconsole = $v['SERIALCONSOLE'];
        $ismaster = $v['MASTER'];

//store this pass's fan labels, etc for comparison in the next loop
        $fanlabels = $v['FANNAMES'];
        $templabels = $v['TEMPNAMES'];
        $voltlabels = $v['VOLTNAMES'];

        $numfans = $v['NUMFANS'];
        $numtemps = $v['NUMTEMPS'];
        $numvolts = $v['NUMVOLTS'];

//check if we are dealing with a new set of labels/headings  
        if ($newconfig == 1) {
            if ($onepass != 0)
                echo "</table><br>\n";  //only after the first pass will you have to first end a table    
            echo "<table border=1>\n";
            echo "<tr bgcolor=lightblue>\n";
            echo "<th bgcolor=mediumblue><font color=white>Name</font></th>\n";
            if (strcmp ($v['CHASSIS'], "fastree") != 0) {
                echo "<td>&nbsp;</td>\n";
                echo "<th bgcolor=mediumblue><font color=white>Power</font></th>\n";
            }
            if ($showControls) {
                echo "<th bgcolor=mediumblue><font color=white>Reset</font></th>\n";
                if ($_SESSION['usesSerialconsole'] == "yes")
                    echo "<th bgcolor=mediumblue><font color=white>Console</font></th>\n";
            };
            echo "<td>&nbsp;</td>\n";
//category headings
            echo "<th colspan=$numfans bgcolor=mediumblue><font color=white>Fan&nbsp;Speed&nbsp;(rpm)</font></th>\n";
            echo "<td>&nbsp;</td>\n";
            echo "<th colspan=$numvolts bgcolor=mediumblue><font color=white>Voltages&nbsp;(V)</font></th>\n";
            echo "<td>&nbsp;</td>\n";
            echo "<th colspan=$numtemps bgcolor=mediumblue><font color=white>Temperatures (&deg;C)</font></th>\n";

            if (strcmp ($v['CHASSIS'], "fastree") != 0) {
                echo "<td>&nbsp;</td>\n";
                echo "<th bgcolor=mediumblue><font color=white>OS Running</font></th>\n";
            }
            echo "</tr>\n";

            echo "<tr bgcolor=lightblue>\n";
            echo "<th>&nbsp;</th>\n";
            if (strcmp ($v['CHASSIS'], "fastree") != 0) {
                echo "<td>&nbsp;</td>\n";
                echo "<th>&nbsp;</th>"; // Power
            }
            if ($showControls) {
                echo "<th>&nbsp;</th>"; // Reset
                if ($_SESSION['usesSerialconsole'] == "yes")
                    echo "<th>&nbsp;</th>";     // Console
            };

//creates labels for parameters for this type of node
            echo "<td>&nbsp;</td>";
            echo microwayNodewatchHeaders ($node, 'FANNAMES', $numfans, "<th>", "</th>");
            echo "<td>&nbsp;</td>";
            echo microwayNodewatchHeaders ($node, 'VOLTNAMES', $numvolts, "<th>", "</th>");
            echo "<td>&nbsp;</td>";
            echo microwayNodewatchHeaders ($node, 'TEMPNAMES', $numtemps, "<th>", "</th>");
            if (strcmp ($v['CHASSIS'], "fastree") != 0) {
                echo "<td>&nbsp;</td>";
                echo "<td bgcolor=lightblue>&nbsp;</td>";
            }
            echo "</tr>\n";
        };

        echo "<tr>\n";

//display time data was acquired
        $thisNode = $microwayNodewatchData[$node];
        $sampleTime = $thisNode['TIME'];
        if (microwayStatusCurrent ($node)) {
            echo "<th bgcolor=lightblue>$node</th>\n";
        } else {
            if ($sampleTime == 0) {
                $sampleTimeString = "NO DATA";
            } else {
                $sampleTimeString = "Old data<br>".strftime ("%d%b%g %H:%M:%S", $sampleTime);
            };
            echo "<th bgcolor=lightpink>$node<br>$sampleTimeString</th>\n";
        };


//display whether machine is powered on and also offer on/off buttons in control mode
        if (strcmp ($v['CHASSIS'], "fastree") != 0) {
            echo "<td>&nbsp;</td>\n";
            $power = microwayNodePowerStatus ($node);
            switch ($power) {
            case 'off':
                echo "<td bgcolor=yellow align=center>";
                break;
            case 'on':
                echo "<td bgcolor=lightgreen align=center>";
                break;
            default:
                echo "<td bgcolor=lightpink align=center>";
                break;
            };
            echo "<table cellpadding=0 cellspacing=0 valign=middle><tr>";
            if ($power == 'on')
                echo "<td>On</td>";
            if ($power == 'off')
                echo "<td>Off</td>";
            if ($power == 'unknown' && !$showControls)
                echo "<td>Unknown</TD>";
            if ($showControls) {
                echo "<td>";
                echo microwayNodePowerButton ($node);
                echo "</td>";
            };
            echo "</tr></table>";
            echo "</td>";
        }
        if ($showControls) {
            echo "<td bgcolor=lightgreen>";
            echo microwayNodeResetButton ($node);
            echo "</td>";
        };
        if ($_SESSION['usesSerialconsole'] == "yes") {
            if ($showControls && $serialconsole) {
                echo "<td bgcolor=lightgreen>";
                echo microwaySerialConsoleButton ($node);
                echo "</td>";
            } else if ($showControls && $ismaster) {
                echo "<td bgcolor=lightgreen>";
                echo microwayLocalConsoleButton ($node);
                echo "</td>";
            } else if ($showControls) {
                echo "<td bgcolor=lightgreen>";
                echo "No serial console";
                echo "</td>";
            }
        }
        echo "<td>&nbsp;</td>\n";

//print out the individual node's elements
        echo microwayNodewatchElements ($node, 'FAN', $numfans, 0,
                                        "<td align=right bgcolor=lightgreen>", "<td align=right bgcolor=yellow>", "<td align=right bgcolor=\"#FF0505\">", "</td>", true, -1, -2);
        echo "<td>&nbsp;</td>\n";
        echo microwayNodewatchElements ($node, 'VOLT', $numvolts, 1,
                                        "<td align=right bgcolor=lightgreen>", "<td align=right bgcolor=yellow>", "<td align=right bgcolor=\"#FF0505\">", "</td>", true, 0, -1);
        echo "<td>&nbsp;</td>\n";
        echo microwayNodewatchElements ($node, 'TEMP', $numtemps, 1,
                                        "<td align=right bgcolor=lightgreen>", "<td align=right bgcolor=yellow>", "<td align=right bgcolor=\"#FF0505\">", "</td>", true, -100, -200);

        if (strcmp ($v['CHASSIS'], "fastree") != 0) {
            $hostList = mw2gHostsUpOfCluster ();
            $match = 0;
            foreach ($hostList as $k => $v) {
                $shortname = ereg_replace ("\..*", "", $k);
                if (strcmp (trim ($shortname), trim ($node)) == 0)
                    $match = 1;
            }
            echo "<td bgcolor=lightblue>&nbsp;</td>";
            if ($match == 0) {
                $temprww = array_search ($node, $hostList);
                echo "<td align=center bgcolor=yellow>no</td>";
            } else {
                echo "<td align=center bgcolor=lightgreen>yes</td>";
            }
        }

        echo "</tr>\n";
        $onepass++;
    };

    echo "</table>\n";
    return;
}
// emit a visible control that provides serial console access
//
// Inputs:
//      nodeName        name of node to reset
function microwaySerialConsoleButton ($nodeName) {
    echo "<form ACTION=serialconsole.php METHOD=get>";
    // NOTE: align=center or no alignment causes extensive cell padding in IE 5
    echo "<table cellpadding=0 cellspacing=0 align=right><tr>\n";
    echo "<td align=left>";
    echo "<input TYPE=hidden NAME=nodeName VALUE=\"$nodeName\">";
    echo "<input TYPE=submit onclick=\"launchserialconsole('serialconsole.php?nodeName=$nodeName'); return false;\"  VALUE=\"Serial\">";
    echo "&nbsp;</td>";
    echo "</tr></table>";
    echo "</form>";
    return;
}

// emit a visible control that provides local console access
//
// Inputs:
//      nodeName        name of node to reset
function microwayLocalConsoleButton ($nodeName) {
    echo "<form ACTION=/ganglia/admin/localconsole.php METHOD=get>";
    // NOTE: align=center or no alignment causes extensive cell padding in IE 5
    echo "<table cellpadding=0 cellspacing=0 align=left><tr>\n";
    echo "<td>";
    echo "<input TYPE=hidden NAME=nodeName VALUE=\"$nodeName\">";
    echo "<input TYPE=submit onclick=\"launchconsole('/ganglia/admin/localconsole.php'); return false;\"  VALUE=\"Local\">";
    echo "</td>";
    echo "</tr></table>";
    echo "</form>";
    return;
}


// emit a visible control that resets a node
//
// Inputs:
//      nodeName        name of node to reset
// ???At first this is just a button, but it may require password and user fields
function microwayNodeResetButton ($nodeName) {
    echo "<form ACTION=resetnode.php METHOD=get>";
    // NOTE: align=center or no alignment causes extensive cell padding in IE 5
    echo "<table cellpadding=0 cellspacing=0 align=left><tr>\n";
    echo "<td>";
    echo "<input TYPE=hidden NAME=nodeName VALUE=\"$nodeName\">";
    echo "<input TYPE=submit VALUE=\"Reset\">";
    echo "</td>";
    echo "</tr></table>";
    echo "</form>";
    return;
}

// emit a visible control that turns power on/off for a node
// Both buttons are emitted if power status is unknown
//
// Inputs:
//      nodeName        name of node to reset
function microwayNodePowerButton ($nodeName) {
    $power = microwayNodePowerStatus ($nodeName);
    echo "<form ACTION=powernode.php METHOD=get>";
    echo "<table cellpadding=0 cellspacing=0 align=left><tr><td>";
    if ($power != 'off') {
        echo "<input TYPE=hidden NAME=nodeName VALUE=\"$nodeName\">";
        echo "<input TYPE=hidden NAME=power VALUE=OFF>";
        echo "<input bgcolor=green TYPE=submit VALUE=Off>";
    };
    if ($power != 'on') {
        echo "<input TYPE=hidden NAME=nodeName VALUE=\"$nodeName\">";
        echo "<input TYPE=hidden NAME=power VALUE=ON>";
        echo "<input bgcolor=yellow TYPE=submit VALUE=On>";
    };
    echo "</td></tr></table></form>";
    return;
}

//emit a button to reset all nodes
function microwayAllNodesResetButton () {
    echo "<form ACTION=powernode.php METHOD=get>";
    // NOTE: align=center or no alignment causes extensive cell padding in IE 5
    echo "<table cellpadding=0 cellspacing=0 align=right><tr>\n";
    echo "<td>&nbsp;";
    echo "<input TYPE=hidden NAME=nodeName VALUE=\"all\">\n";
    echo "<input TYPE=hidden NAME=power VALUE=reset>\n";
    echo "<input TYPE=submit VALUE=\"All Slaves Reset\">\n";
    echo "&nbsp;</td>\n";
    echo "</tr></table>";
    echo "</form>";
    return;
}

// emit a visible control that turns power on/off for all nodes
//
// Inputs:
//      powerOn    true for button to turn power on
function microwayAllNodesPowerButton ($powerOn) {
    echo "<form ACTION=powernode.php METHOD=get>";
    // NOTE: align=center or no alignment causes extensive cell padding in IE 5
    echo "<table cellpadding=0 cellspacing=0 align=right><tr>\n";
    echo "<td>&nbsp;";
    echo "<input TYPE=hidden NAME=nodeName VALUE=\"all\">\n";
    if ($powerOn) {
        echo "<input TYPE=hidden NAME=power VALUE=ON>\n";
        echo "<input TYPE=submit VALUE=\"All Slaves Power ON\">\n";
    } else {
        echo "<input TYPE=hidden NAME=power VALUE=OFF>\n";
        echo "<input TYPE=submit VALUE=\"All Slaves Power OFF\">\n";
    };
    echo "&nbsp;</td>\n";
    echo "</tr></table>";
    echo "</form>";
    return;
}
// Control to collect username and password
//
// Inputs:
//      defaultUser
//      page                    page to read when clicked
function microwayGetUserAndPwd ($defaultUser, $page) {
    // When user is established and matches keep the password that was
    // passed to the page. Otherwise junk it.

    if (!isset ($_SESSION['currentUser']))
        $_SESSION['currentUser'] = "";
    if (!isset ($_SESSION['currentPassword']))
        $_SESSION['currentPassword'] = "";
    if ($defaultUser == "") {
        $defaultUser = $_SESSION['currentUser'];
    };
    echo "<form ACTION=\"$page\" METHOD=submit>";
    echo "<table cellpadding=0 cellspacing=0>\n";
    echo "<tr><th colspan=3>Change current user</th></tr>\n";
    echo "<tr>";
    echo "<td>User:&nbsp;<input type=text size=30 name=newUser></td>\n";
    echo "<td>Password:&nbsp;<input type=password size=30 name=newPassword></td>\n";
    echo "<td><input TYPE=submit VALUE=\"Change User\"></td>\n";
    echo "<td><input TYPE=reset VALUE=\"Clear\"></td>\n";
    echo "</tr>";
    echo "</table>";
    echo "</form>";
    return;
}

function microwayInitialLogin ($defaultUser, $page) {
    // When user is established and matches keep the password that was
    // passed to the page. Otherwise junk it.

    if (!isset ($_SESSION['currentUser']))
        $_SESSION['currentUser'] = "";
    if (!isset ($_SESSION['currentPassword']))
        $_SESSION['currentPassword'] = "";

    if ($defaultUser == "") {
        $defaultUser = $_SESSION['currentUser'];
    };
    echo "<form ACTION=\"$page\" METHOD=post>";
    echo "<table cellpadding=0 cellspacing=0>\n";
    echo "<tr><th colspan=3>Login</th></tr>\n";
    echo "<tr><td>&nbsp;</td></tr>";
    echo "<tr>";
    echo "<td>User:&nbsp;<input type=text size=30 name=newUser></td>\n";
    echo "<td>&nbsp;</td><td>Password:&nbsp;<input type=password size=30 name=newPassword></td>\n";
    echo "</tr><tr>";
    echo "<td align=right><input TYPE=submit VALUE=\"Login\"></td>\n";
    echo "<td align=left><input TYPE=reset VALUE=\"Clear\"></td>\n";
    echo "</tr>";
    echo "</table>";
    echo "</form>";
    return;
}

// Control to copy a file to all nodes of the cluster
//
// Control to copy a file to all nodes of the cluster
//
// Inputs:
//      defaultFrom
//      defaultDestDir
//      defaultNodes
function microwayCopyFile ($defaultFrom, $defaultDestDir, $defaultNodes) {
    echo "<form ACTION=copyfile.php METHOD=get>";
    echo "<table cellpadding=0 cellspacing=0>\n";
    echo "<tr><th colspan=3>Copy file from master to node(s)</th></tr>\n";
    echo "<tr>";
    echo "<td>From:&nbsp;<input type=text size=22 name=fromfile value=\"$defaultFrom\"></td>\n";
    echo "<td>To&nbsp;path:&nbsp;<input type=text size=22 name=todir value=\"$defaultDestDir\"></td>\n";
    echo "<td>";
    echo microwaySlaveSelect ($defaultNodes, true, "nodes");
    echo "</td>";
    echo "<td><input TYPE=submit VALUE=Copy></td>\n";
    echo "</tr>";
    echo "</table>";
    echo "</form>";
    return;
}
// Shutdown a particular node
function microwayShutdownNodes () {
    echo "<form ACTION=haltnode.php METHOD=get>";
    echo "<table cellpadding=0 cellspacing=0>\n";
    echo "<tr><td align=right>Shutdown node(s):&nbsp;&nbsp;</td>\n";
    echo "<td>";
    if (!isset ($nodes))
        $nodes = "";
    echo microwaySlaveSelect ($nodes, true, "nodes");
    echo "</td>";
    echo "<td><input TYPE=submit VALUE=Shutdown></td>\n";
    echo "</tr></table></form>";
}

// Shutdown a particular node
function microwayRebootNodes () {
    echo "<form ACTION=rebootnode.php METHOD=get>";
    echo "<table cellpadding=0 cellspacing=0>\n";
    echo "<tr><td>Reboot node(s):&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;&nbsp;</td>\n";
    echo "<td>";
    if (!isset ($nodes))
        $nodes = "";
    echo microwaySlaveSelect ($nodes, true, "nodes");
    echo "</td>";
    echo "<td><input TYPE=submit VALUE=Reboot></td>\n";
    echo "</tr></table></form>";
}

// Control to perform a command on all nodes of the cluster
// Or on one selected node.
//
// Inputs:
//      defaultCommand
function microwayCommandNodes ($defaultCommand, $nodes = "Master") {
    echo "<form ACTION=executecommand.php METHOD=get>";
    echo "<table cellpadding=0 cellspacing=0>\n";
    echo "<tr><th colspan=3>Perform a command on node(s) as root</th></tr>\n";
    echo "<tr>";
    echo "<td>Command:&nbsp;<input type=text size=40 name=command value=\"$defaultCommand\">&nbsp;</td>\n";
    echo "<td>";
    echo microwayNodeSelect ($nodes, true, true, "nodes");
    echo "</td>";
    echo "<td><input TYPE=submit VALUE=Execute></td>\n";
    echo "</tr>";
    echo "</table>";
    echo "</form>\n";
    return;
}

// Test to make sure user and password are defined
function microwayUserRequired () {
    if (($_SESSION['currentUser'] == "") || ($_SESSION['currentPassword'] == "")) {
        echo "<h1>ERROR: User and password must be defined</h1><br>";
        echo "<h2><a CLASS=menu HREF=login.php>Return&nbsp;to&nbsp;Login&nbsp;Screen</a></h2><br>\n";

        pageEnd ();
        exit ();
    };
    return;
}

// Perform a command locally (as whatever user the PHP page runs as)
//
// Inputs:
//      command
// Returns:
//      output of the command
function microwayPerformCommandLocally ($command) {
    $output = shell_exec ("$command 2>&1");
    return $output;
}

// Perform a command on the current node as the current user with
// the current password using the wrapper for root access
//
// Inputs:
//      command
// Returns:
//      output of the command
function microwayPerformCommand ($command) {
    $currentPassword = $_SESSION['currentPassword'];
    $currentUser = $_SESSION['currentUser'];
    //using proc_open
    $descriptorspec = array (0 => array ("pipe", "r"),  // stdin is a pipe that the child will read from
                             1 => array ("pipe", "w"),  // stdout is a pipe that the child will write to
                             2 => array ("pipe", "w")); // stderr is a pipe that the child will write to
    //invoke wrapper and pass it the command and password via stdin
    $process = proc_open ("PATH=\$PATH:/usr/local/bin:/usr/local/sbin /usr/local/sbin/wrapper $currentUser \"$command\" 2>&1", $descriptorspec, $pipes);
    if (is_resource ($process)) {
        // $pipes now looks like this:
        // 0 => writeable handle connected to child stdin
        // 1 => readable handle connected to child stdout
        // 2 => readable handle connected to child stderr
        $output = "";
        fwrite ($pipes[0], "$currentPassword\n");
        fclose ($pipes[0]);
        while (!feof ($pipes[1])) {
            $output .= fgets ($pipes[1], 1024);
        };
        fclose ($pipes[1]);
        fclose ($pipes[2]);
        // It is important that you close any pipes before calling
        // proc_close in order to avoid a deadlock
        $return_value = proc_close ($process);

        return $output;
    };
    return "";
}

// Perform a command on the current node as the current user with
// the current password using the wrapper for root access
//
// Inputs:
//      command
//      input
// Returns:
//      output of the command
function microwayPerformCommandInput ($command,$input) {
    $currentPassword = $_SESSION['currentPassword'];
    $currentUser = $_SESSION['currentUser'];
    //using proc_open
    $descriptorspec = array (0 => array ("pipe", "r"),  // stdin is a pipe that the child will read from
                             1 => array ("pipe", "w"),  // stdout is a pipe that the child will write to
                             2 => array ("pipe", "w")); // stderr is a pipe that the child will write to
    //invoke wrapper and pass it the command and password via stdin
    $process = proc_open ("PATH=\$PATH:/usr/local/bin:/usr/local/sbin /usr/local/sbin/wrapper $currentUser \"$command\" 2>&1", $descriptorspec, $pipes);
    if (is_resource ($process)) {
        // $pipes now looks like this:
        // 0 => writeable handle connected to child stdin
        // 1 => readable handle connected to child stdout
        // 2 => readable handle connected to child stderr
        $output = "";
        fwrite ($pipes[0], "$currentPassword\n");
        usleep(20000);
        fwrite ($pipes[0], "$input");
        fclose ($pipes[0]);
        while (!feof ($pipes[1])) {
            $output .= fgets ($pipes[1], 1024);
        };
        fclose ($pipes[1]);
        fclose ($pipes[2]);
        // It is important that you close any pipes before calling
        // proc_close in order to avoid a deadlock
        $return_value = proc_close ($process);

        return $output;
    };
    return "";
}

// Perform a command on the all nodes as the current user with
// the current password using the wrapper for root access
//
// Inputs:
//      command
//      input
// Returns:
//      output of the command
function microwayPerformCommandInputAll ($command,$input) {
    $currentPassword = $_SESSION['currentPassword'];
    $currentUser = $_SESSION['currentUser'];
    
    $output = "";
    $hostList = mw2gHostsUpOfCluster();
    if (is_array ($hostList)) {
        foreach ($hostList as $k => $v) {

            //using proc_open
            $descriptorspec = array (0 => array ("pipe", "r"),  // stdin is a pipe that the child will read from
                                     1 => array ("pipe", "w"),  // stdout is a pipe that the child will write to
                                     2 => array ("pipe", "w")); // stderr is a pipe that the child will write to
            //invoke wrapper and pass it the command and password via stdin
            $process = proc_open ("PATH=\$PATH:/usr/local/bin:/usr/local/sbin /usr/local/sbin/wrapper $currentUser rsh $k \"$command\" 2>&1", $descriptorspec, $pipes);
            if (is_resource ($process)) {
                // $pipes now looks like this:
                // 0 => writeable handle connected to child stdin
                // 1 => readable handle connected to child stdout
                // 2 => readable handle connected to child stderr
                fwrite ($pipes[0], "$currentPassword\n");
                usleep(20000);
                fwrite ($pipes[0], "$input");
                fclose ($pipes[0]);
                while (!feof ($pipes[1])) {
                    $tmpoutput .= fgets ($pipes[1], 1024);
                };
                if($tmpoutput != "") $output .= "$k: $tmpoutput\n";
                fclose ($pipes[1]);
                fclose ($pipes[2]);
                // It is important that you close any pipes before calling
                // proc_close in order to avoid a deadlock
                $return_value = proc_close ($process);
            } else {
                $output .= "Could not run command on $k\n";
            }
        }
        return $output;
    } else return "";
}


// Execute a command on all nodes
//
// Inputs:
//      command
function microwayPerformCommandAll ($command) {
    return microwayPerformCommand ("/usr/local/bin/rcom \"".$command."\"");
}


// Test if the element (for any node) is represented as an array.
//
// Inputs:
//      element         name of element to test
// Returns:
//      true if some node represents this element as an array
function microwayIsElementArray ($element) {
    global $microwayNodewatchData;

    foreach ($microwayNodewatchData as $node1 => $v) {
        $thisNode = $microwayNodewatchData[$node1];
        if ($thisNode != "") {
            $thisName = $thisNode[$element];
            if (is_array ($thisName)) {
                return true;
            };
        };
    };
    return false;
}

// Test if status for a node is current
//
// Inputs:
//    $node
// Returns:
//    true if current
function microwayStatusCurrent ($node) {
    global $microwayNodewatchData;
    $thisNode = $microwayNodewatchData[$node];
    $sampleTime = $thisNode['TIME'];
    if ($sampleTime == 0)
        return false;
    if ($sampleTime > (time () - (2 * 60))) {
        return true;
    };
    return false;
}


// Test if the element (for any node) is represented as an array.
//
// Inputs:
//      nodeName        node name to check
// Returns:
//      'on' 'off' or 'unknown'
function microwayNodePowerStatus ($nodeName) {
    global $microwayNodewatchData;

    if (!array_key_exists ($nodeName, $microwayNodewatchData)) {
        return 'unknown';
    };
    if (!microwayStatusCurrent ($nodeName))
        return 'unknown';
    if (!array_key_exists ('SWITCHON', $microwayNodewatchData[$nodeName])) {
        return 'unknown';
    };
    if ($microwayNodewatchData[$nodeName]['SWITCHON'])
        return 'on';
    return 'off';
}


// Limit a number to N decimal places.
// Do nothing if not a number.
//
// Inputs:
//      value
//      places  -1 to ignore number of places
// Returns:
//      number rounded to N decimal places
function decimalPlaces ($value, $places) {
    if ($places < 0)
        return $value;
    if (!is_numeric ($value))
        return $value;
    return number_format (round ($value, $places), $places);
}

// Display headers to allow heterogeneous cluster
//
// Inputs:
//      node    node to report
//      name    name of element to report
//      count   at least this number of elements are put out, &nbsp; for others
//      before  string to place prior to each element
//      after   string to place after each element
// Returns:
//      string in the form:
//              before value after
function microwayNodewatchHeaders ($node, $name, $count, $before, $after) {
    global $microwayNodewatchData;
    global $microwayAlertLimits;
    $result = "";
    $thisNode = $microwayNodewatchData[$node];
    $number = 0;
    if ($thisNode != "") {
        $thisName = $thisNode[$name];
        if (is_array ($thisName)) {
            foreach ($thisName as $k => $v) {
                $number += 1;
                if ($number > $count)
                    break;
                $result .= $before;
                $result .= $v;
                $result .= $after;
            };
        }
    };
    return $result;
}

// Report an error by emitting a page
//
// Inputs:
//      message         error message
function microwayErrorPage ($message) {
    pageStart ("ERROR");
    echo "<h2>$message</h2>";
    pageEnd ();
    exit ();
}

function microwayGetGroups($username)
{
    $groupfile=fopen("/etc/group","r");
    $count=0;
    $membergroups=array();

    while (!feof($groupfile)) {
        $groupline=fgets($groupfile,1024);
        if ($groupline=="") continue;
        $grouparray=preg_split("/:/",$groupline);
        $groupname=$grouparray[0];
        $groupmembers=preg_split("/,/",$grouparray[3]);
        foreach ($groupmembers as $members) {
            if(strcmp($username,str_replace(array(" ","\n"),"",$members))==0) {
                $membergroups[$count]=$groupname;
                $count++;
            }
        }
    }
    fclose($groupfile);
    return $membergroups;
}

function microwayGetDefaultGroups()
{
    $groupfile=fopen("/etc/mcms-default-groups","r");
    if(!$groupfile) return array();
    $count=0;
    $defaultgroups=array();

    while (!feof($groupfile)) {
        $groupline=trim(fgets($groupfile,1024));
        if ($groupline=="") continue;
	$defaultgroups[$count]=$groupline;
	$count++;
    }
    fclose($groupfile);
    return $defaultgroups;
}
function microwaySetDefaultGroups($groups)
{
    microwayPerformCommandInput("cat > /etc/mcms-default-groups",$groups);
}
?>
