<?php
// A library of function definitions that establish the page framework for
// the entire collection of pages.
//
// It contains functions:
//
//      pageInit                Called to initialize early. Emits no page data
//      pageStart               Starts the HTML output, emits header and a table
//                              framework with menu down the left side.
//      pageEnd                 completes the output, closes the table and ends the HTML.
//      pageSetRefresh          call prior to pageStart to set alternate refresh interval
//
// Other modules call these functions to create a page framework.
// Major modifications and rewrites by Rick Warner, from 2002-2007
//
session_start ();
include_once ("microwaylibrary.php");

$_SESSION['usesNodewatch'] = "no";
$_SESSION['usesIPMI'] = "yes";
$_SESSION['usesSerialconsole'] = "yes";

// Nodewatch with trademark symbol
$nodewatchTM = "NodeWatch<FONT SIZE=\"-1\"><SUP>TM</SUP></FONT>";
$mcmsTM = "MCMS<FONT SIZE=\"-1\"><SUP>TM</SUP></FONT>";

// Flag used within this module to identify refresh time per page
$pageRefresh = 0;

// Set time to refresh this page
//
// Inputs:
//      refreshRate
// Returns:
//      previous refresh rate
function pageSetRefresh ($refreshRate) {
    global $pageRefresh;
    $temp = $pageRefresh;
    $pageRefresh = $refreshRate;
    return $temp;
}

// Perform initial housekeeping, determine if user exits etc.
// (This page does not have users implemented, but libraries
// call these functions and expect this subroutine interface).
//
// Inputs:
//  autoLogin       if true try to login using cookie
function pageInit ($autoLogin) {
    microwayHeaderStuff ();
};

// Initialize a page for HTML output. Emit the header and start the body.
//
// Inputs:
//  pageName        name of the page for the title and header
function pageStart ($pageName) {
    global $pageRefresh;
    global $stylesheet;
    global $PHPSESSID;

    $title = ereg_replace ('<BR>', '. ', $pageName);

    if (isset ($_REQUEST['newUser']))
        $_SESSION['currentUser'] = $_REQUEST['newUser'];
    if (isset ($_REQUEST['newPassword']))
        $_SESSION['currentPassword'] = $_REQUEST['newPassword'];

    if (isset ($_SESSION['currentUser']))
        $currentUser = $_SESSION['currentUser'];
    if (isset ($_SESSION['currentPassword']))
        $currentPassword = $_SESSION['currentPassword'];

    echo "<!DOCTYPE HTML PUBLIC \"-//W3C//DTD HTML 4.0 Transitional//EN\">";
    echo "<html>\n<head>\n<title>$title</title>\n";
    if ($stylesheet != "") {
        echo "<link rel=stylesheet type=\"text/css\" href=\"$stylesheet\">\n";
    };
    if (isset ($currentUser) && $currentUser != "SearchableUserAccountEntry123") {
        $logintest = microwayPerformCommand ("/bin/true");
        if ($logintest != "") { //any output besides nothing indicates error
            echo "<META http-equiv=\"refresh\" content=\"0;URL=login.php?failed=yes\">\n";
            echo "</HEAD></HTML>";
            exit ();
        }
    }
    if (isset ($currentUser) && $currentUser == "SearchableUserAccountEntry123") {
        $self = $_SERVER['PHP_SELF'];
        if (substr_count ($self, "login.php") != 1) {
            echo "<META http-equiv=\"refresh\" content=\"0;URL=login.php?failed=yes\">\n";
            echo "</HEAD></HTML>";
            exit ();
        }
    }

    if (isset ($_SESSION['refreshTo']))
        $refreshTo = $_SESSION['refreshTo'];
    if (isset ($_SESSION['refreshTime']))
        $refreshTime = $_SESSION['refreshTime'];
    if (isset ($refreshTo) && $refreshTo != "") {
        echo "<META http-equiv=\"refresh\" content=\"$refreshTime;URL=$refreshTo\">\n";
    } else if ($pageRefresh > 0) {
        $self = $_SERVER['PHP_SELF'];
        echo "<META http-equiv=\"refresh\" content=\"$pageRefresh; URL=$self?refresh=$pageRefresh\">\n";
    };
    ?>
<SCRIPT LANGUAGE = "JavaScript" type = "text/javascript">
    var timeoutfunc;
    var serialconsole;

    function launchserialconsole (url) {
        self.name = "opener";
        var newwidth = 0;
        var newheight = 0;
        if (document.getElementById) {
            var rulerSpan = document.getElementById ('ruler');
            rulerSpan.innerHTML = '12345678901234567890123456789012345678901234567890123456789012345678901234567890123';
            newwidth = rulerSpan.offsetWidth;
            rulerSpan.innerHTML = 'a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>';
            newheight = rulerSpan.offsetHeight;
            rulerSpan.innerHTML = '';
        }
        serialconsole = open (url, "serialconsole", "resizable,scrollbars=no,width=" + newwidth + ",height=" + newheight + ",left=300,top=100");
        timeoutfunc = window.setInterval ("isserialclosed()", 250);
    }

    function isserialclosed () {
        if (window.serialconsole.closed) {
            clearserial ();
            clearInterval (timeoutfunc);
        }
    }

    function launchconsole (url) {
        self.name = "opener";
        var newwidth = 0;
        var newheight = 0;
        if (document.getElementById) {
            var rulerSpan = document.getElementById ('ruler');
            rulerSpan.innerHTML = '12345678901234567890123456789012345678901234567890123456789012345678901234567890123';
            newwidth = rulerSpan.offsetWidth;
            rulerSpan.innerHTML = 'a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>a<BR>';
            newheight = rulerSpan.offsetHeight;
            rulerSpan.innerHTML = '';
        }
        remote = open (url, "", "resizable,scrollbars=no,width=" + newwidth + ",height=" + newheight + ",left=300,top=100");
    }
    function clearserial () {
        var ua = navigator.userAgent.toLowerCase ();
        if (!window.ActiveXObject)
            clearrequest = new XMLHttpRequest ();
        else if (ua.indexOf ('msie 5') == -1)
            clearrequest = new ActiveXObject ("Msxml2.XMLHTTP");
        else
            clearrequest = new ActiveXObject ("Microsoft.XMLHTTP");

        clearrequest.open ("POST", "/ganglia/admin/clearseriallockdumb.php");
        clearrequest.send ("");
    }
</script>
<?php
    echo "</head>\n";
    echo "<body>\n";
    echo "<span id=\"ruler\" style=\"visibility:hidden;\"></span>\n";
    echo "<TABLE width=100% cellspacing=2 border=0>\n";
    echo "<tr><td colspan=2>";
    microwayBanner ($pageName);
    echo "</td>";               //<td>&nbsp;</TD>\n";
    echo "</tr>";
    // This second row contains the selections and the content
    echo "<tr>\n";
    echo "<td WIDTH=\"20%\" ALIGN=left VALIGN=top bgcolor=lightblue>\n";
    microwayClusterControlMenu ();
    echo "</td>\n";
    echo "<td WIDTH=\"80%\" ALIGN=left VALIGN=top>";
    return;
};

// Complete the HTML output. Closes the table, emits a footer and closes the page
//
function pageEnd () {
    global $webmaster;
    global $mcmsTM;
    echo "</td>\n";
    echo "</tr>\n";
    // This row of spaces fakes out the table layout system and forces the left
    // column to be narrow and the right column to take up the rest of the page
    echo "<tr><td>&nbsp;</td><td>\n";
    echo "&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n";
    echo "&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n";
    echo "&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n";
    echo "&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n";
    echo "&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n";
    echo "&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n";
    echo "&nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n";
    echo "&nbsp; &nbsp; &nbsp; &nbsp; &nbsp;\n";
    echo "</td></tr>\n";

    // Footer for the page
    echo "<tr><td align=center colspan=2>\n";
    echo "<address>\n";
    echo "$mcmsTM\n";
    echo " - <a CLASS=menu HREF=\"http://www.microway.com/\">Microway</a>\n";
    echo "Cluster Management Software<BR>Send comments and corrections to: \n";
    echo "<a href=\"mailto:$webmaster\">$webmaster</a>\n";
    echo "</address>\n";
    echo "</td></tr>\n";
    echo "</table>\n";
    echo "</body>\n";
    echo "</html>";
    return;
};


$admin_mailto = "support@microway.com";
$database = "";
$workDatabase = "Microway";
$workName = "Microway";
$dbuser = "";
$dbpassword = "";
$groupname = "Microway&nbsp;Cluster&nbsp;Management";
$keywords = "Microway, Cluster Management";
$pagename = "";
$stylesheet = "microwaystyle.css";
$webmaster = "support@microway.com";

?>
