#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <list>
using namespace std;

alert::alert(int mytype, char mynodename[], int myindex, double myvalue, double mylimits[]) {
  type=mytype;
  index=myindex;
  value=myvalue;
  strcpy(nodename,mynodename);
  limits[0]=mylimits[0]; 
  limits[1]=mylimits[1]; 
  limits[2]=mylimits[2]; 
  limits[3]=mylimits[3];
  timestamp=time(NULL);
  occurances=0;
  reported=0;
}

bool alert::operator<(const alert& T) {
  time_t myowndifference, comparedifference;
  myowndifference=timestamp-reported;
  comparedifference=T.timestamp-T.reported;
  return (myowndifference<comparedifference);
}

/* this function might not be needed - moved into add function
int alerts::find(int type, char nodename[], int index){ //returns index to myalerts with the match, or -1 if no match
  list<alert>::iterator i;
  for (i=begin();i!=end();++i) {
    if(i.matches(type,nodename,index)) return i;
    }
  }
  return -1;
}*/


bool alert::matches(int comptype, char compnodename[], int compindex ) {
  return ( (type==comptype) && (index==compindex) && (strcmp(nodename,compnodename)==0) );
}
  
bool alert::matches(char compnodename[]) {
  return (strcmp(nodename,compnodename)==0);
}
  
void alerts::add(int type, char nodename[], int index, double value, double limits[]) {
  //fprintf(ERROROUTPUT,"Attempting to add alert of type: %d nodename: %s index: %d, value: %0.2f, limits: %0.2f %0.2f %0.2f %0.2f\n",
  //        type, nodename, index, value, limits[0], limits[1], limits[2], limits[3]); 
  list<alert>::iterator i;
  
  for (i=myalerts.begin();i!=myalerts.end();++i) {
    if(i->matches(type,nodename,index)) {
      i->value=value;
      i->timestamp=time(NULL);
      i->occurances++;
      //fprintf(ERROROUTPUT, "Found existing entry and modified it\n");
      return;
     }
  }
  alert newalert(type, nodename, index, value, limits);
  myalerts.push_back(newalert);
  //fprintf(ERROROUTPUT, "Did not find existing entry, added new one\n");
  return;
}

void alerts::prune(int type, char nodename[], int index) {
  list<alert>::iterator i;
  while(alertlock==true) usleep(5000);
  alertlock=true;

  if(myalerts.empty()) {
    alertlock=false;
    return;
  }
  
  for (i=myalerts.begin();i!=myalerts.end();++i) {
    if(i->matches(type,nodename,index)) {
      if((time(NULL) - i->timestamp) > (ALERTSOFTFREQUENCY-30)) {   //only remove alerts that are at least ALERTSOFTFREQUENCY (7m currently) old
        myalerts.erase(i);
        break;
      }
    }
  }
  alertlock=false;
}

void alerts::remove(char nodename[]) {
  list<alert>::iterator i;
  bool erasedsomething=false;   //could be more than 1 alert present for this node, make sure to del all of them

  while(alertlock==true) usleep(5000);
  alertlock=true;
  if(myalerts.empty()) {
    alertlock=false;
    return;
  }
  
  for (i=myalerts.begin();i!=myalerts.end();++i) {
    if(i->matches(nodename)) {
      myalerts.erase(i);
      erasedsomething=true;
      break;
    }
  }
  alertlock=false;
  if(erasedsomething) remove(nodename);  //recurse until you removed nothing
} 

int alert::delaytime(bool sendingalready) {
  if(!sendingalready) return ALERTHARDFREQUENCY;
  else return ALERTSOFTFREQUENCY;
}



void *LaunchSendFunction(void *obj)
{
    email *myLauncher = reinterpret_cast<email *>(obj);
    return myLauncher->send();
}

void alerts::process() {
  while(alertlock==true) usleep(5000);
  alertlock=true;
  if (myalerts.empty()) {
    alertlock=false;
    return;
  }

  myalerts.sort();  // sort by (timestamp-reported)
  myalerts.reverse();
  
  list<alert>:: iterator i;
  bool pending=false;
  time_t currenttime=time(NULL);
  
  for (i=myalerts.begin(); i!=myalerts.end();++i) {
    //critical alerts ignore the pending time
    if(i->occurances>1) {  //  alert must trigger a few times before shutdown - this is to prevent incorrect detection of a crit system as it shuts off or ipmi that momentarily flakes out
      if( (abs(comparelimits(i->value, i->limits))==2 ) && (i->timestamp - i->reported) > CRITFREQUENCY ){   //CRITFREQUENCY defined in defines.h, currently 45
        if(pending==false) myemail.clear();
        pending=true;
        i->reported=currenttime;
        globalreporttime=currenttime;
        myemail.add(*i);
      }
    }
  } 
  if ( ((currenttime-globalreporttime) > GLOBALFREQUENCY) || pending==true ) {
    for (i=myalerts.begin(); i!=myalerts.end();++i) {
      if(i->occurances>1) {  //  alert must trigger a few times before being handled - avoids incorrect reports for systems shutting down or ipmi that momentarily flakes out
        if((i->timestamp - i->reported) > i->delaytime(pending)) {
          if(pending==false) myemail.clear();
          pending=true;
          i->reported=currenttime;
          globalreporttime=currenttime;
          myemail.add(*i);
        }
      }
    }
  }
  alertlock=false;
  if(pending) {
    if ( (myemail.threadstatus==ALERT_STATUS_IDLE) || (myemail.threadstatus==ALERT_STATUS_DONE)){
      myemail.threadstatus=ALERT_STATUS_SPUN;
      if (pthread_create(&emailthread, NULL, LaunchSendFunction, &myemail)) {
        pthread_detach(emailthread);
      }
    } else { 
      pthread_cancel(emailthread);
      error_msg((char *)"Last warning email did not complete sending before next mail was ready.  Terminated previous one and sending new.\n");
      if (pthread_create(&emailthread, NULL, LaunchSendFunction, &myemail)) {
        pthread_detach(emailthread);
      }
    }
  }
     
  return;
}

 
