#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <list>
#include <sys/types.h>
#include <sys/socket.h>
#include <unistd.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <pthread.h>
using namespace std;

bool email::add(alert i){
  char temp[8000];
  
  switch(comparelimits(i.value, i.limits)) {
    case -2:    //crit low
      sprintf(temp,"Critical low alert generated for node %s:\n", i.nodename);
      break;
    case -1:    //alert low
      sprintf(temp,"Warning low alert generated for node %s:\n", i.nodename);
      break;
    case 1:     //alert high
      sprintf(temp,"Warning high alert generated for node %s:\n", i.nodename);
      break;
    case 2:     //crit high
      sprintf(temp,"Critical high alert generated for node %s:\n", i.nodename);
      break;
  }

  switch(i.type) {
    case FAN:
      sprintf(temp,"%s"
                   "  Fan number %d reading %0.2f\n", temp, i.index+1, i.value);
      break;
    case TEMP:
      sprintf(temp,"%s"
                   "  Temperature number %d reading %0.2f\n", temp, i.index+1, i.value);
      break;
    case VOLT:
      sprintf(temp,"%s"
                   "  Voltage number %d reading %0.2f\n", temp, i.index+1, i.value);
      break;
  }
  
  sprintf(temp,"%s"
               "  Configured limits for this sensor are crit low: %0.2f, alert low: %0.2f,\n"
               "  alert high: %0.2f, and crit high: %0.2f\n",
               temp, i.limits[0],i.limits[1],i.limits[2],i.limits[3]);

  if (abs(comparelimits(i.value, i.limits)) > 1) { // this is critical - initiate shutdown
    sprintf(temp, "%s"
                  "  As this is a critical value, the node is being powered off now.\n\n", temp);
    mycluster.poweroffnode(i.nodename);
  } else {
    sprintf(temp, "%s"
                  "  This is just a warning.  No automatic shutdowns are occuring for this event.\n\n", temp);
  }

  sprintf(bodyindex,temp);
  bodyindex+=strlen(temp);
  return true;
}

bool emailverbose=false;
//MOST of this code reused from nodewatch server code, which I (Rick) wrote originally anyway
void * email::send() {
  struct hostent *hostinfo;
  struct sockaddr_in name;
  int sock;
  char temp[MAXEMAILSIZE], temp2[401];
  char dests[MAXEMAILRECIPIENTS][MAXEMAILADDRESS];
  int count, count2 = 0, numdests = 0;

  pthread_setcancelstate ( PTHREAD_CANCEL_ENABLE, NULL );
  pthread_setcanceltype ( PTHREAD_CANCEL_ASYNCHRONOUS, NULL );

  for ( count = 0; to[count] != '\0'; count++ ) {
    if ( to[count] == ',' ) {
      numdests++;
      count2 = 0;
    } else {
      dests[numdests][count2] = to[count];
      count2++;
    }
  }
  threadstatus = ALERT_STATUS_BUSY;

  sock = socket ( PF_INET, SOCK_STREAM, IPPROTO_TCP );
  if ( sock < 0 ) {
    error_msg ( (char *) "Email transmission socket error\n" );
    close ( sock );
    return NULL;
  }
  name.sin_family = AF_INET;
  name.sin_port = htons ( 25 ); //mail server runs on port 25

//resolve hostname
  hostinfo = gethostbyname ( server );
  if ( hostinfo == NULL ) {
    hostinfo =
      gethostbyaddr ( server, strlen ( server ), AF_INET );
  }
  if ( hostinfo == NULL ) {
    sprintf ( temp,
              "Could not resolve %s.  The following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, to, subject, body );
    error_msg ( temp );
    threadstatus = ALERT_STATUS_ERROR;
    close ( sock );
    return NULL;
  }
  memcpy ( &name.sin_addr, hostinfo->h_addr, hostinfo->h_length );

#ifdef EMAIL_DEBUG
  fprintf ( stderr, "ip=%d\n", hostinfo->h_addr );

#endif /*  */
//open socket      
  if ( connect ( sock, ( struct sockaddr * ) &name, sizeof ( name ) ) < 0 ) {
    sprintf ( temp,
              "Could not connect to %s (IP %s) "
              "The following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, hostinfo->h_addr, to, subject, body );
    error_msg ( temp );
    threadstatus = ALERT_STATUS_ERROR;
    close ( sock );
    return NULL;
  }

//////////////////////////////////////////////////   
//actually send message
//////////////////////////////////////////////
  //read in all the stuff a mail server pronounces about itself
  if ( read ( sock, temp, 1999 ) == 0 ) {
    threadstatus = ALERT_STATUS_ERROR;
    sprintf ( temp,
              "Lost communication with %s (IP %s) "
              "The following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, hostinfo->h_addr, to, subject, body );
    error_msg ( temp );
    close ( sock );
    return NULL;
  }
  if ( emailverbose )
    error_msg ( temp );

  //send the initial HELO
  sprintf ( temp, "HELO %s\n", server );
  if ( 0 == write ( sock, temp, strlen ( temp ) ) ) {   //verfity transmit
    threadstatus = ALERT_STATUS_ERROR;
    sprintf ( temp,
              "Lost communication with %s (IP %s) "
              "The following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, hostinfo->h_addr, to, subject, body );
    error_msg ( temp );
    close ( sock );
    return NULL;
  }
  if ( read ( sock, temp2, 400 ) == 0 ) {       //verify a response
    threadstatus = ALERT_STATUS_ERROR;
    sprintf ( temp,
              "Lost communication with %s (IP %s) "
              "The following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, hostinfo->h_addr, to, subject, body );
    error_msg ( temp );
    close ( sock );
    return NULL;
  }
  if ( emailverbose )
    error_msg ( temp2 );
  if ( 0 != strncmp ( temp2, "250", 3 ) ) {
    threadstatus = ALERT_STATUS_ERROR;
    sprintf ( temp,
              "Error transmitting alert to %s (IP %s) -- "
              "HELO message failed\nThe following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, hostinfo->h_addr, to, subject, body );
    error_msg ( temp );
    close ( sock );
    return NULL;
  }
  
  //send the from:      
  sprintf ( temp, "mail from: %s\n", from );
  if ( 0 == write ( sock, temp, strlen ( temp ) ) ) {   //verfity transmit
    threadstatus = ALERT_STATUS_ERROR;
    sprintf ( temp,
              "Lost communication with %s (IP %s) "
              "The following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, hostinfo->h_addr, to, subject, body );
    error_msg ( temp );
    close ( sock );
    return NULL;
  }
  if ( read ( sock, temp2, 400 ) == 0 ) {       //verify a response
    threadstatus = ALERT_STATUS_ERROR;
    sprintf ( temp,
              "Lost communication with %s (IP %s) "
              "The following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, hostinfo->h_addr, to, subject, body );
    error_msg ( temp );
    close ( sock );
    return NULL;
  }
  if ( emailverbose )
    error_msg ( temp2 );
  if ( 0 != strncmp ( temp2, "250", 3 ) ) {
    threadstatus = ALERT_STATUS_ERROR;
    sprintf ( temp,
              "Error transmitting alert to %s (IP %s) "
              "-- mail from: %s failed\nThe following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, hostinfo->h_addr, from, to, subject, body );
    error_msg ( temp );
    close ( sock );
    return NULL;
  }
  
  //send the to:
  for ( count = 0; count <= numdests; count++ ) {
    sprintf ( temp, "rcpt to: %s\n", dests[count] );
    if ( 0 == write ( sock, temp, strlen ( temp ) ) ) { //verfity transmit
      threadstatus = ALERT_STATUS_ERROR;
      error_msg ( temp );
      sprintf ( temp,
                "Lost communication with %s (IP %s) "
                "The following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
                server, hostinfo->h_addr, to, subject, body );
      close ( sock );
      return NULL;
    }
    if ( read ( sock, temp2, 400 ) == 0 ) {     //verify a response
      threadstatus = ALERT_STATUS_ERROR;
      sprintf ( temp,
                "Lost communication with %s (IP %s) "
                "The following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
                server, hostinfo->h_addr, to, subject, body );
      close ( sock );
      return NULL;
    }
    if ( emailverbose )
      error_msg ( temp2 );
    if ( 0 != strncmp ( temp2, "250", 3 ) ) {
      threadstatus = ALERT_STATUS_ERROR;
      sprintf ( temp,
                "Error transmitting alert to %s (IP %s) "
                "-- rcpt to: %s failed\nThe following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
                server, hostinfo->h_addr, to, to, subject, body );
      error_msg ( temp );
      close ( sock );
      return NULL;
    }
  }

  //send DATA
  sprintf ( temp, "DATA\n" );
  if ( 0 == write ( sock, temp, strlen ( temp ) ) ) {   //verfity transmit
    threadstatus = ALERT_STATUS_ERROR;
    sprintf ( temp,
              "Lost communication with %s (IP %s) "
              "The following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, hostinfo->h_addr, to, subject, body );
    error_msg ( temp );
    close ( sock );
    return NULL;
  }
  if ( read ( sock, temp2, 400 ) == 0 ) {       //verify a response
    threadstatus = ALERT_STATUS_ERROR;
    sprintf ( temp,
              "Lost communication with %s (IP %s) "
              "The following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, hostinfo->h_addr, to, subject, body );
    error_msg ( temp );
    close ( sock );
    return NULL;
  }
  if ( emailverbose )
    error_msg ( temp2 );
  if ( 0 != strncmp ( temp2, "354", 3 ) ) {
    threadstatus = ALERT_STATUS_ERROR;
    sprintf ( temp,
              "Error transmitting alert to %s (IP %s) "
              "-- DATA command failed\nThe following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, hostinfo->h_addr, to, subject, body );
    error_msg ( temp );
    close ( sock );
    return NULL;
  }

  //send the subject and the From: and the to:
  sprintf ( temp, "Subject: %s\nFrom: Cluster IPMI monitoring <%s>\nTo: %s\n", subject, from, to );
  if ( 0 == write ( sock, temp, strlen ( temp ) ) ) {   //verfity transmit
    threadstatus = ALERT_STATUS_ERROR;
    sprintf ( temp,
              "Lost communication with %s (IP %s) "
              "The following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, hostinfo->h_addr, to, subject, body );
    error_msg ( temp );
    close ( sock );
    return NULL;
  }

  //send the body
  sprintf ( temp, "\n%s\n.\n", body );
  if ( (write ( sock, temp, strlen(temp))) < (signed int)strlen(temp) ) {   //verify transmit
    threadstatus = ALERT_STATUS_ERROR;
    sprintf ( temp,
              "Lost communication with %s (IP %s) "
              "The following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, hostinfo->h_addr, to, subject, body );
    
    error_msg ( temp );
    close ( sock );
    return NULL;
  }
  if ( read ( sock, temp2, 400 ) == 0 ) {       //verify a response
    threadstatus = ALERT_STATUS_ERROR;
    sprintf ( temp,
              "Lost communication with %s (IP %s) "
              "The following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, hostinfo->h_addr, to, subject, body );
    error_msg ( temp );
    close ( sock );
    return NULL;
  }
  if ( emailverbose )
    error_msg ( temp2 );
  if ( 0 != strncmp ( temp2, "250", 3 ) ) {
    threadstatus = ALERT_STATUS_ERROR;
    sprintf ( temp,
              "Error transmitting alert to %s (IP %s) "
              "-- BODY not accepted\nThe following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, hostinfo->h_addr, to, subject, body );
    error_msg ( temp );
    close ( sock );
    return NULL;
  }

  //send the QUIT
  sprintf ( temp, "QUIT\n" );
  if ( 0 == write ( sock, temp, strlen ( temp ) ) ) {   //verfity transmit
    threadstatus = ALERT_STATUS_ERROR;
    sprintf ( temp,
              "Lost communication with %s (IP %s) "
              "The following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, hostinfo->h_addr, to, subject, body );
    error_msg ( temp );
    close ( sock );
    return NULL;
  }

  //read in all the stuff a mail server pronounces about itself
  if ( read ( sock, temp, 1999 ) == 0 ) {
    threadstatus = ALERT_STATUS_ERROR;
    sprintf ( temp,
              "Lost communication with %s (IP %s) "
              "The following alert could not be delivered:\nTo: %s\nSubject: %s\n%s\n",
              server, hostinfo->h_addr, to, subject, body );
    error_msg ( temp );
    close ( sock );
    return NULL;
  }
  if ( emailverbose )
    error_msg ( temp );

  close ( sock );
  threadstatus = ALERT_STATUS_DONE;
  return NULL;
}

