#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <list>
#include <pthread.h>
#include <OpenIPMI/ipmiif.h>
#include <OpenIPMI/ipmi_smi.h>
#include <OpenIPMI/ipmi_err.h>
#include <OpenIPMI/ipmi_auth.h>
#include <OpenIPMI/ipmi_lan.h>
#include <OpenIPMI/ipmi_posix.h>

using namespace std;

node::node(){
  inShutdown=false;
  inAlert=false;
  lastupdate=0;
  lastquery=0;
  lastcheck=0;
  power=false;
  sensorcount=0;
  firsttime=0;
  wasdead=false;
}

node::node(chassis *chassisptr){
  thischassis=chassisptr;
  inShutdown=false;
  inAlert=false;
  lastupdate=0;
  lastquery=0;
  lastcheck=0;
  power=false;
  sensorcount=0;
  firsttime=0;
  wasdead=false;
}

bool node::getdata(){
  int count=0;
  time_t ctime;
  ctime=time(NULL);
  
  if (dead==1) {
    wasdead=true;
    return false;    //node is not reachable, something happened?
  }
  if (wasdead) {
    wasdead=false;
    firsttime=ctime;   //start fresh with regards to alerts since it came back to life
  }
  
  if (firsttime==0) firsttime=ctime;  //first reading time, do not process alerts until ALERT_DELAY seconds after this

  if ((ctime-lastquery) > QUERY_DELAY) {
    lastquery=ctime;
    while (count<sensorcount) {
      use_sdata(mysdatas[count]);
      ipmi_sensor_get_reading(mysensors[count], got_thresh_reading, mysdatas[count]);
      //release_sdata(mysdatas[count]);
      count++;
    }
    usleep(5000);  //give a little time for the replies to come in
    return true;
  } else return false;
}

bool node::command(int command) {    //this still uses ipmitool for now
  FILE *ipmiproc;
  char ipmicommand[150];
  char *outputline;
  outputline=(char *)malloc(300);
  
  switch (command) {
    case POWEROFF:
      snprintf(ipmicommand,150,"%s -I lan -H %s-ipmi -U %s -P %s power off", ipmitool, name, username, password);
      ipmiproc=popen(ipmicommand,"r");
      while (!feof(ipmiproc)) {
        fgets(outputline,300,ipmiproc);
        //FIXME - CHECK IF IT ACTUALLY WORKED
      }
      pclose(ipmiproc);
      break;
    case POWERON:
      snprintf(ipmicommand,150,"%s -I lan -H %s-ipmi -U %s -P %s power on", ipmitool, name, username, password);
      ipmiproc=popen(ipmicommand,"r");
      while (!feof(ipmiproc)) {
        fgets(outputline,300,ipmiproc);
        //FIXME - CHECK IF IT ACTUALLY WORKED
      }
      pclose(ipmiproc);
      break;
    case RESET:
      snprintf(ipmicommand,150,"%s -I lan -H %s-ipmi -U %s -P %s power reset", ipmitool, name, username, password);
      ipmiproc=popen(ipmicommand,"r");
      while (!feof(ipmiproc)) {
        fgets(outputline,300,ipmiproc);
        //FIXME - CHECK IF IT ACTUALLY WORKED
      }
      pclose(ipmiproc);
      break;
    default:
      //log.append("Unknown command code %i issued", command);
      fprintf(ERROROUTPUT,"Unknown command code %i issued", command);
  }
  free(outputline); 
  return true;
}


int node::checklimits() {
  int count, retval=true;
  time_t ctime=time(NULL);

  if((ctime-firsttime)<ALERT_DELAY) return true;   //return an OK until ALERT_DELAY seconds have passed since 1st reading
                                                   //this ensures enough time for good readings
  if(inShutdown==true) return true;      //already shutting down, stop checking

  if (power==false) return true;  //node is not powered, or is otherwise not readable
  
  if ((ctime-lastcheck) > QUERY_DELAY) {   //in addition to querying every 5s,limit checks to every 5s too
    lastcheck=ctime;
  } else return true;

  for (count=0 ; count<thischassis->numfans; count++) {
    if (comparelimits(fan[count],thischassis->fanlimit[count]) != 0 ) {
      retval=false;
      pendingalerts.add(FAN, name, count, fan[count], thischassis->fanlimit[count]);
    } else pendingalerts.prune(FAN,name,count);
   
  }
  for (count=0 ; count<thischassis->numtemps; count++) {
    if (comparelimits(temp[count],thischassis->templimit[count]) != 0 ) {
      retval=false;
      pendingalerts.add(TEMP, name, count, temp[count], thischassis->templimit[count]);
    } else pendingalerts.prune(TEMP,name,count);
  }
  for (count=0 ; count<thischassis->numvolts; count++) {
    if (comparelimits(volt[count],thischassis->voltlimit[count]) != 0 ) {
      retval=false;
      pendingalerts.add(VOLT, name, count, volt[count], thischassis->voltlimit[count]);
    } else pendingalerts.prune(VOLT,name,count);
  }
  return(retval);
}

bool node::registersensor(ipmi_sensor_t *sensor){
  sdata_t *sdata;
  char sensorname[MAXSENSORNAME];
  int count=0;
  unsigned int linelen;

  sdata = alloc_sdata(sensor);
  if (!sdata) {
    printf("Unable to allocate sensor name memory\n");
    return false;
  }
  sensor_nameonly(sensorname,sdata->name);

  for (count=0 ; count < thischassis->numfans ; count++) {
    linelen=strlen(thischassis->fanname[count]);
    if (strlen(sensorname)==linelen && strncmp(sensorname,thischassis->fanname[count],linelen)==0) {
      addsensor(sensor);
    }
  }
    
  for (count=0 ; count < thischassis->numvolts ; count++) {
    linelen=strlen(thischassis->voltname[count]);
    if (strlen(sensorname)==linelen && strncmp(sensorname,thischassis->voltname[count],linelen)==0) {
      addsensor(sensor);
    }
  }

  for (count=0 ; count < thischassis->numtemps ; count++) {
    linelen=strlen(thischassis->tempname[count]);
    if (strlen(sensorname)==linelen && strncmp(sensorname,thischassis->tempname[count],linelen)==0) {
      addsensor(sensor);
    }
  }
  return true;
}

bool node::reportsensor(char *sensorname, double value) {
  int count=0;
  unsigned int linelen;
  time_t ctime=time(NULL);

  for (count=0 ; count < thischassis->numfans ; count++) {
    linelen=strlen(thischassis->fanname[count]);
    if (strlen(sensorname)==linelen && strncmp(sensorname,thischassis->fanname[count],linelen)==0) {
      fan[count]=value;
      lastupdate=ctime;
      if( (fan[count]>0) && (inShutdown==false)) {
        if(power==false) { 
          firsttime=time(NULL);   //node was just powered back on, give it time to settle
          power=true;
        }
      }
      return true;
    }
  }
  for (count=0 ; count < thischassis->numvolts ; count++) {
    linelen=strlen(thischassis->voltname[count]);
    if (strlen(sensorname)==linelen && strncmp(sensorname,thischassis->voltname[count],linelen)==0) {
      volt[count]=value;
      lastupdate=ctime;
      if( (volt[count]>0) && (inShutdown==false)) {
        if(power==false) { 
          firsttime=time(NULL);   //node was just powered back on, give it time to settle
          power=true;
        }
      }
      if(volt[count]==0) power=false;
      return true;
    }
  }
  for (count=0 ; count < thischassis->numtemps ; count++) {
    linelen=strlen(thischassis->tempname[count]);
    if (strlen(sensorname)==linelen && strncmp(sensorname,thischassis->tempname[count],linelen)==0) {
      temp[count]=value;
      lastupdate=ctime;
      if( (temp[count]>0) && (inShutdown==false)) {
        if(power==false) { 
          firsttime=time(NULL);   //node was just powered back on, give it time to settle
          power=true;
        }
      }
      return true;
    }
  }
  return true;
}

bool node::reportdomain(int err) {
  if(err!=0) dead=1;
  else dead=0;
  return true;
}

bool node::zerosensors() {
  int count;
  lastupdate=time(NULL);
  power=false;
  
  for (count=0 ; count < thischassis->numfans ; count++) fan[count]=0;
  for (count=0 ; count < thischassis->numvolts ; count++) volt[count]=0;
  for (count=0 ; count < thischassis->numtemps ; count++) temp[count]=0;
  return true;
}

 
bool node::addsensor(ipmi_sensor_t *sensor){
  if(sensorcount>=MAXSENSORS) {
    sdata_t *sdata;
    char sensorname[MAXSENSORNAME];

    sdata = alloc_sdata(sensor);
    if (!sdata) {
      printf("Unable to allocate sensor name memory\n");
      exit(1);
    }
    sensor_nameonly(sensorname,sdata->name);

    char errmsg[1024];
    sprintf(errmsg,"Unable to add sensor %s to %s.  Max sensor count of %d reached for this node.\n",sensorname,name,MAXSENSORS);
    error_msg((char *)errmsg);
    return false;
  }
  mysensors[sensorcount]=sensor;
  mysdatas[sensorcount]=alloc_sdata(sensor);
  sensorcount++;
  return true;
}
 
///start shutdown()
void node::shutdown(){  
  pthread_t shutdown_monitor;
  shutdown_node *myself;
  myself=(shutdown_node *)malloc(sizeof(shutdown_node));
  strcpy(myself->nodename,name);
  strcpy(myself->password,password);
  strcpy(myself->username,username);
  myself->inShutdown=&inShutdown;

  if (inShutdown==true) { //we are already in a shutdown.  let that one go instead
    return;
  }
  inShutdown=true;
  

  //RSH TO SYSTEM TO TRY A SAFE SHUTDOWN
  error_msg((char *)"Spawning shutdown thread\n");
 
  if (pthread_create(&myself->thread, NULL, safeShutdown, myself)==0){
    pthread_detach(myself->thread);
  }  
  
  //IF after a few mins it still isn't down, use command(POWEROFF) to send an IPMI poweroff
  if (pthread_create(&shutdown_monitor, NULL, ensureShutdown, myself)){
    pthread_detach(shutdown_monitor);
  }
 
  //inShutdown=false;  only when shutdown is over should this toggle back, and the thread does that
}
///end shutdown


