// cardconfig.h

// specifies global parameters describing the kind of card installed
// in the system (hardware configuration) and cluster info:

// Also specifies some conventions for communication in the file.

#define CLUSTER_NODES 32        // ACTUAL NUMBER OF NODES IN THIS CLUSTER

#define CYGNAL_DVREF 2.4        // voltage reference on the Cygnal chip.
#define CYGNAL_ANALOG_GAIN 1    // fixed in Cygnal software.

// Error values returned to file for problematic thermistors:
#define UNPLUG_THERMO_OUT -100. // -100 for unplugged thermistor
#define SHORT_THERMO_OUT -200.  // -200 for shorted thermistor
// Note that actual outputs will be multiplied by scale factor.
#define UNPLUG_THERMO_LIMIT 4090        // above this raw value, unplugged.
#define SHORT_THERMO_LIMIT 10   // below this raw value, short circuit.

// Please note:  Although this software does its best to identify
// shorted or unplugged thermistors, if there are otherwise malfunctioning
// thermistors, there is no way to detect that, unless we define
// a range of impossible temperatures, and even then we can't be sure.

//   the assignment of the analog lines.
//   reference voltage information
//   

// For this version, only one kind of card hardware configuration is
// allowed (homogenius cluster)

// In future versions, more and more of the items contained here 
// will be migrated to run-time arguments to nodewserver,
// and eventually to personalized data stored in the flash of each card.

// ALLOCATION OF THE 8 AVAILABLE ANALOG CHANNELS:

#define MAX_TACK    13          // changed from 5 to 11 when T2 tacks were added. - then changed to 13 with new tricom

#define MAX_TEMP 5
#define MAX_VOLT 3

// the following are physical analog channel numbers.  
// Our card always has 8 total physical analog channels.

#define THERMO_CYGNAL 8         // This is the NW card chip temp.
#define THERMO_ARRAY {3, 4, 5, 6, 7}    // 5 thermistors hooked up like so
#define VOLT_ARRAY {0, 1, 2}    // first three are 12, 5 and 3.3 volts
#define TACK_ARRAY {0, 1, 2, 3, 4, 8, 5, 6, 7, 9, 10}
// output sequence for fans: first 5 original fans, one addl fan,
// 2 power supply fans, one spare PS fan, and 2 spare case fans.


// VOLTAGE MEASUREMENT CALIBRATIONS

// The following describe the resistors in the volt circuits.
// indexes into following arrays correspond with those of VOLT_ARRAY

// It is OK to scale everything up in the following.
#define VOLT_R1_ARRAY {825., 681., 562.}        // preceding multipled by 10.
#define VOLT_R2_ARRAY {150., 392., 562.}        // preceding multipled by 10.

#define VOLT_NOMINAL_ARRAY {1200., 500., 330.}  // What we nominally expect.
#define VOLT_MULTIPLIER 100.    // voltages are reported x 100 volts.
#define THERMO_MULTIPLIER 100.  // voltages are reported x 100 degrees C.

// I'm sorry, but the above had to be floating doubles rather than ints,
// to preserve the integrity of intermediate calculations.
// the final output of the calculations are cast to ints for printing.


// THERMISTOR MEASUREMENT CALIBRATIONS

// The following arrays describe how to convert thermocouple 
// measurements to degrees C:

/*  The following info for flat disc thermistors is not being used.
// Material Type D10.3
// ratio range               72.5 - 3.363   3.363 - .3507  .3507 - .06370  .06370 - .0169

#define TEMA_ARRAY { 3.3564846E-03,  3.3540172E-03,  3.3530496E-03,  3.3467185E-03}
#define TEMB_ARRAY {2.4622340E-04,  2.5027462E-04,  2.4792929E-04,  2.4342924E-04} //MLJ corrected typo
#define TEMC_ARRAY { 4.0776512E-06,  2.4300527E-06,  1.0822040E-06,  3.3465275E-07} // MLJ corrected typo
#define TEMD_ARRAY {-6.0295285E-08, -7.2909526E-08, -6.4194339E-08, -4.5415459E-08}

#define TEMSEL_ARRAY {3.363, 0.3507, 0.06370, 0.0169}
*/

// Material Type D9.7A
// Bulb-style thermistors, we are using these:
// ratio range               69.26 - 3.277   3.277 - .3599  .3599 - .06816  .06816 - .0187

#define TEMA_ARRAY { 3.3570420E-03,  3.3540170E-03,  3.3530481E-03,  3.3536166E-03}
#define TEMB_ARRAY {2.5214848E-04,  2.5617244E-04,  2.5420230E-04,  2.5377200E-04}      //MLJ corrected typo
#define TEMC_ARRAY { 3.3743283E-06,  2.1400943E-06,  1.1431163E-06,  8.5433271E-07}     // MLJ corrected typo
#define TEMD_ARRAY {-6.4957311E-08, -7.2405219E-08, -6.9383563E-08, -8.7912262E-08}

#define TEMSEL_ARRAY {3.277, 0.3599, 0.06816, 0.0187}
