// This file describes the communication protocol between
// the card and the PC in terms of what and how much data
// to expect in packets.

// By comparison, cardconfig.h describes how the card is physically
// set up and what this data really means.

#define MAX_NODES 256           // changed from 128 to 256 MLJ to avoid buffer overruns.
#define MAX_CHAN    9
// Max A2D channels.
//MAX_TACK is in cardconfig.h now

#define DEAD_DELAY 4
#define ALIVESTATE 0

typedef struct tagMYSHORTINT {  // big endian two byte integer.
  // the endian function converts this to a real short int. 
  unsigned char Highval;
  unsigned char Lowval;
} MYSHORTINT;

typedef struct tagSTATUS {
  unsigned char bAlarm:1;
  unsigned char bStatus1:1;
  unsigned char bStatus2:1;
  unsigned char bStatus3:1;
  unsigned char bStatus4:1;
  unsigned char bStatus5:1;
  unsigned char bStatus6:1;
  unsigned char bStatus7:1;
} STATUS;

typedef struct tagVERSION {
  unsigned char ucMajor;
  unsigned char ucMinor;
  //  unsigned short pad;
} VERSION;

typedef struct tagNODE {
  STATUS status;
  VERSION version;              // software version of the CARD
  unsigned char alive;          // Mike added to keep track of non-responsive cards.
  unsigned long ulSerial;
  unsigned char bLower[MAX_CHAN];
  unsigned char bUpper[MAX_CHAN];
  signed short int uaA2D[MAX_CHAN];     // AFTER endian converts
  // changed to SIGNED
  unsigned short int uaRpm[MAX_TACK];   // AFTER endian converts
  unsigned long timestamp;      // when latest data received.
  unsigned char powerstat;      // indicates whether the system is on or off.
  // above updated by execnode.c

  unsigned char power_status;   // is it on or off or changing right now?

  unsigned int shutdowncount;   // how many measurements has it been in
  // shutdown range?
  unsigned int alertcount;      // how many measurements has it been in
  // email alert range?

} NODE;

typedef struct tagSTAPACKET {
  unsigned char ucLength;
  unsigned char ucUnit;
  unsigned char ucCode;
  STATUS status;
  // unsigned char ucStatus;  // done in WindowsNT...
} STAPACKET;

typedef struct tagVERPACKET {
  unsigned char ucLength;
  unsigned char ucUnit;
  unsigned char ucCode;
  VERSION version;
} VERPACKET;

typedef struct tagADRPACKET {
  unsigned char ucLength;
  unsigned char ucUnit;
  unsigned char ucCode;
  unsigned char ucAddress;
} ADRPACKET;

typedef struct tagA2DPACKET {
  unsigned char ucLength;
  unsigned char ucUnit;
  unsigned char ucCode;
  unsigned char ucChannel;
  MYSHORTINT uA2D;              // MLJ ADDED SHORT
} A2DPACKET;

typedef struct tagANAPACKET {
  unsigned char ucLength;
  unsigned char ucUnit;
  unsigned char ucCode;
  MYSHORTINT uA2D[MAX_CHAN];    // MLJ ADDED SHORT
} ANAPACKET;

typedef struct tagTACPACKET {
  unsigned char ucLength;
  unsigned char unUnit;
  unsigned char ucCode;
  unsigned char ucChannel;
  MYSHORTINT uRpm;
} TACPACKET;

typedef struct tagTCAPACKET {
  unsigned char ucLength;
  unsigned char unUnit;
  unsigned char ucCode;
  MYSHORTINT uaRpm[MAX_TACK];
} TCAPACKET;


typedef struct tagSERPACKET {
  unsigned char ucLength;
  unsigned char ucUnit;
  unsigned char ucCode;
  unsigned long ulSerial;
} SERPACKET;

bit execNode ( PACKET * );

void setclusterAlive (  );

void requestProcess (  );
// SYSTEMATICALLY  Enqueues requests for data.
