// process received slave node packets

#include <string.h>
#include <stdio.h>
#include <stdlib.h>


#include "cardconfig.h"
#include "net.h"
#include "codes.h"
#include "exec.h"
#include "flash.h"
#include "limits.h"             // for the CLUSTERNODE typedefs etc.
#include "syslogging.h"
#include "timer.h"

// NODE Node[MAX_NODES];  // This is now stored within ClusterList.

// all from Limits.c:

extern CHASSIS *ChassisList;    // pointer to the start of the chassis 
// linked list database.

extern CLUSTERNODE *ClusterList;        // pointer to the start of the chassis
// linked list database.

extern CLUSTERNODE *NWMaster;   // pointer to this master (running
// this software) as stored in the chassis list.

extern unsigned char flash_override;
extern unsigned int NodeCount;  // number of nodes in linked list
extern unsigned int ChassisCount;       // number of chasses defined
extern unsigned int TempavCount;        // number of nodes containing defined tempaverage probes.
// (if tempaverage = 0 for a node, that node is not included in tempaveragecount)


// unsigned char cluster_nodes = CLUSTER_NODES;    // the number of nodes actually in cluster.  

extern QUEUE QueueMessage;

int stacount = 0;

// static unsigned char nodereq;   // keep track of which node we are requesting. 

CLUSTERNODE *nodereq = NULL;    // keep track of which node we are requesting. 

CLUSTERNODE *addrNode = NULL;   // to save time searching for nodes by
// card address, keep track of the last one and try it first, then
// try the next one, then try them all.
// (see new function searchCardNode)


// the following packets are FROM the PC TO the CARD
// to REQUEST the corresponding information.
PACKET ana = { 2, U_NODE, ANA };
PACKET tca = { 2, U_NODE, TCA };

PACKET ver = { 2, U_NODE, VER };

CLUSTERNODE *searchCardNode ( unsigned char nodeaddress )
     // Locate a node based on the card address
     // uses addrNode to try to save time,
     // by starting the search at that point.
     // returns NULL if cannot locate the address at all.
{
  char searchpass = 0;          // number of times we've passed the end of the list.
  while ( searchpass < 2 )      // we are willing to wrap around only once.
  {
    if ( addrNode != NULL ) {
      if ( addrNode->cardaddress == nodeaddress )
        return ( addrNode );
      else {
        addrNode = addrNode->nextnode;
      }
    } else {
      addrNode = ClusterList;
      searchpass++;
    }
  }                             // end while loop.
  addrNode = NULL;
  return ( NULL );
}


signed short int endianold ( unsigned short int value )
{
  // convert a two-byte big endian int
  // to a SIGNED two-byte little endian int

  signed short int rval;

  ( ( unsigned char * ) &rval )[0] = ( ( unsigned char * ) &value )[1];
  ( ( unsigned char * ) &rval )[1] = ( ( unsigned char * ) &value )[0];
  return ( rval );
}

signed short int endian ( MYSHORTINT value )
{
  // convert a two-byte big endian MYSHORTINT
  // to a two-byte little endian int

  signed short int rval;

  ( ( unsigned char * ) &rval )[0] = value.Lowval;
  ( ( unsigned char * ) &rval )[1] = value.Highval;
  //  if (rval > 0x7fff)
  //     rval -= 0x8000; // SIGN REPAIR
  return ( rval );
}


// CHANGED TO BE CALLED DIRECTLY rather than via netexecu
bit execNode ( PACKET * pPacket )       // responses from slave nodes
{
  char tempstring[200];         // for errors
  if ( pPacket ) {
    // First, we will (inefficiently) create multiple
    // copies of pointers to our packet, each one 
    // assuming/expecting a structure of a particular kind of
    // packet...

    STAPACKET *pSTAPacket = ( STAPACKET * ) pPacket;
    VERPACKET *pVERPacket = ( VERPACKET * ) pPacket;
    A2DPACKET *pA2DPacket = ( A2DPACKET * ) pPacket;
    ANAPACKET *pANAPacket = ( ANAPACKET * ) pPacket;
    SERPACKET *pSERPacket = ( SERPACKET * ) pPacket;

    TCAPACKET *pTCAPacket = ( TCAPACKET * ) pPacket;
    TACPACKET *pTACPacket = ( TACPACKET * ) pPacket;

    /*  Following replaced with linked list approach

       // Below is the TRICKY CONVERSION from PACKET UNIT INDICES
       // to NODE ARRAY INDICES... 

       // WOW!!  if a packet happens to come in with a node number
       // greater than MAX_NODES, then it will cause this software
       // to go out of array bounds- -- segmentation fault!

       int node = pPacket->ucUnit - U_NODE;
     */

    int channel = pA2DPacket->ucChannel;
    // OK but when we add TAC, more correct to seek with correct packet

    // ...Then, once we determine what kind of packet 
    // we actually have, we will utilize the appropriate
    // copy of the pointer to the packet which has the
    // correct structure.

    // We will then transfer (copy) the data from the 
    // packet to the node (card data) structure. 

    if ( searchCardNode ( pPacket->ucUnit ) != NULL ) {
      // at this point, the current node is 
      // in the global addrNode

      switch ( pPacket->ucCode ) {
         case STA:             // status response
           if ( stacount++ > 100 ) {
             stacount = 0;
           }
           addrNode->nwdata.status = pSTAPacket->status;
           break;

         case VER:             // version response
           addrNode->nwdata.version = pVERPacket->version;
           break;

         case SER:             // serial number response
           addrNode->nwdata.ulSerial = pSERPacket->ulSerial;
           break;

         case ANA:             // all analog channels
           addrNode->nwdata.alive = ALIVESTATE; // Mark that the data was actually received.
           addrNode->nwdata.timestamp = mygetcurtime (  );      // mark time as current
           for ( channel = 0; channel < MAX_CHAN; channel++ ) {
             addrNode->nwdata.uaA2D[channel] =
               endian ( pANAPacket->uA2D[channel] );
             //         uSwap(&addrNode->nwdata.uaA2D[channel]);
           }

           checkThermoLimits ( addrNode );
           checkVoltLimits ( addrNode );

           // both Thermo and Volt information are provided by 
           // an analog channel message.
           // The voltarray and thermoarray contained
           // within the chassis spec identifies which analog
           // channel is hooked up to what kind of sensor.

           break;

         case A2D:             // single analog channel

           addrNode->nwdata.uaA2D[channel] = endian ( pA2DPacket->uA2D );
           // uSwap(&addrNode->nwdata.uaA2D[channel]);
           break;


           // This kind of thing needs to be 
           // adjusted to take a variable number of fans, and it should
           // be based on the length of the packet!!

         case TCA:             // all tachometer channels
           addrNode->nwdata.alive = ALIVESTATE; // Mark that the data was actually received.
           addrNode->nwdata.timestamp = mygetcurtime (  );      // mark time as current
           for ( channel = 0; channel < MAX_TACK; channel++ ) {
             addrNode->nwdata.uaRpm[channel] =
               endian ( pTCAPacket->uaRpm[channel] );
           }

           checkTackLimits ( addrNode );

           break;

         case TAC:             // single tachometer channels

           channel = ( int ) pTACPacket->ucChannel;
           addrNode->nwdata.uaRpm[channel] = endian ( pTACPacket->uRpm );

           break;



           // The LLM, ULM and NOM packets STILL transfer
           // the analog data exactly as transfered
           // by A2D.  However, by having a different
           // packet name, the concept of an out of limit
           // situation is communicated to the host.

         case LLM:             // alarm channel lower analog limit
           addrNode->nwdata.uaA2D[channel] = endian ( pA2DPacket->uA2D );
           // uSwap(&addrNode->nwdata.uaA2D[channel]);
           addrNode->nwdata.bLower[channel] = TRUE;
           break;

         case ULM:             // alarm channel upper analog limit
           addrNode->nwdata.uaA2D[channel] = endian ( pA2DPacket->uA2D );
           // uSwap(&addrNode->nwdata.uaA2D[channel]);
           addrNode->nwdata.bUpper[channel] = TRUE;
           break;

         case NOM:             // out of alarm channel vaule
           addrNode->nwdata.uaA2D[channel] = endian ( pA2DPacket->uA2D );
           // uSwap(&addrNode->nwdata.uaA2D[channel]);
           addrNode->nwdata.bUpper[channel] = FALSE;
           addrNode->nwdata.bLower[channel] = FALSE;

           // also see special case below


         case FLASH_READ:      // offering data in response to read request.
         case FLASH_WRITE:     // positive acknowledgement AFTER write physically completes.
         case FLASH_ERASE:     // positive acknowledgement AFTER erase physically completes.
           fl_copyresponse ( pPacket ); // update records appropriately
           break;
      }                         // end of Switch
      return ( TRUE );
    }                           // end if search !=NULL
    else {
      // Special case for unnamed nodes being flashed
      // MJ august 2003
      switch ( pPacket->ucCode ) {
         case FLASH_READ:      // offering data in response to read request.
         case FLASH_WRITE:     // positive acknowledgement AFTER write physically completes.
         case FLASH_ERASE:     // positive acknowledgement AFTER erase physically completes.
           fl_copyresponse ( pPacket ); // update records appropriately

           return ( TRUE );     // do not give error below.
           break;
      }

      // If it gets to this point, then it's a malformed packet
      // and also not being used for flashing.
      if ( flash_override != 10 ) {
        sprintf ( tempstring,
                  "ERROR:  execnode -- packet from unknown node %d\n",
                  pPacket->ucUnit );
        error_msg ( tempstring );
      }
      return ( TRUE );          // Just because this particular packet 
      // is malformed doesn't mean the rest are not.  Keep going.
    }
  }                             // end if(pPacket)
  else

    return ( FALSE );           // added to support while() loop in hostl.c.
  // states that the packet was false (empty)
  // Have to check how the return value is used.
}

void setclusterAlive (  )       // Called at the beginning to mark all nodes as alive.
{
  /*  The following loop is being redone below:
     int i;
     for ( i = 0; i < MAX_NODES; i++ )
     Node[i].alive = ALIVESTATE;
     nodereq = 0;                  // start at first node.
   */

  CLUSTERNODE *loopnode;

  loopnode = ClusterList;
  while ( loopnode != NULL ) {
    loopnode->nwdata.alive = ALIVESTATE;
    loopnode = loopnode->nextnode;
  }
  nodereq = ClusterList;        // start the main loop off at the beginning.
}

void requestProcess (  )        // systematically enqueues requests for data.
{
  // the two main changes I've made:
  // 1.  keep extra messages on queue (needed to keep queue from going into
  //                                   polling mode which messes things up)
  // 2.  always queue up something here.
  // 3.  change to net timeout not to requeue the packet at max priority!
  // 4.  Most recently, change to use the linked list system.
  if ( QueueMessage.ucCount < 5 )       // few queued up messages, OK to go.
  {
    if ( nodereq != NULL )      // probably unnecessary check...
      // This would only be an issue if ClusterList were null.
      // and if ClusterList were null, the program should really be halted.
    {
      if ( nodereq->nwdata.alive > 0 )
        nodereq->nwdata.alive--;        // decrement dead delay.
      else {
        // specify the node to which requests are directed:
        tca.ucUnit = nodereq->cardaddress;
        ana.ucUnit = nodereq->cardaddress;
        ver.ucUnit = nodereq->cardaddress;

        enqueue ( &QueueMessage, &ana, MIN_PRIORITY + 1 );
        enqueue ( &QueueMessage, &tca, MIN_PRIORITY + 1 );
        enqueue ( &QueueMessage, &ver, MIN_PRIORITY + 1 );

        nodereq->nwdata.alive = DEAD_DELAY;     // mark node as dead till it responds.

      }
    }
    nodereq = nodereq->nextnode;
    if ( nodereq == NULL )
      nodereq = ClusterList;
  }
}
