/* #include <graph.h> */
#include <stdio.h>
#include <getopt.h>
#include <stdlib.h>
#include <string.h>
#include "cardconfig.h"
#include "net.h"
#include "timer.h"
#include "codes.h"
#include "exec.h"
#include "flash.h"
#include "host_l.h"
#include "limits.h"
#include "datafiles.h"
#include "syslogging.h"

#define VERSION_MAJOR 0
#define VERSION_MINOR 9

extern char *my_null_string;    // declared in limits.c

const char *program_name;
// where is this variable actually stored?  it's extern everywhere it's declared!
int dumpmetrics = 0;
int nodaemon = 0;
int emailverbose = 0;
int noalertshutdown = 0;        // blocks action_limits from running.
int noshutdown = 0;             // blocks auto_shutdown_node from running.
int mydbugverbose = 0;          // enables pt_dbug and friends.  for debugging only.

char MODEMDEVICE[]= "/dev/ttyS0";

char *default_path_p = DEFAULT_PATH;



char *path_p = NULL;


// FILENAME VARIABLES:
char *cmdfilename;              // used in limits.c
//  = "/var/www/html/beostate/data/nodecommands.dat";

char *movedfilename;
//  "/var/www/html/beostate/data/nodecommands-process.dat";

char *clusterfilename;          // used in limits.c
// = "/var/www/html/beostate/data/clusterconf.dat";

char *chassisfilename;          // used in limits.c
// = "/var/www/html/beostate/data/chassisconf.dat";

char *autonodefilename;         // used in limits.c
// = "/var/www/html/beostate/data/autonodeconf.dat";

char *outfilename;

char *usage_errors[] = { "",    //0
  "Undefined option: ",         //1
  "Missing path argument",      //2
  "Missing/null path argument", //3
  "Path does not exist/is not readable: ",      //4
  "Path is not absolute: ",     //5
  "Path too long: ",            //6
  "Path too short: ",			  //7
  "Invalid serial port"            //8
};                             



unsigned char flash_override = FLASH_OVERRIDE_NOTSET;   // set up in flash.c. can be used to block normal card queries.
// to allow flash reads and writes to go through.

// Note that whenever flash_override is non-zero, normal queries for measurements are suspended.

// The initial value of flash_override indicates what this program should do first:
// eventually, this will be controlled by a command line option to nodeserver.
//   0:  normal monitoring operation, no flash intervention.
//   1:  check values of poweroff limits at startup, then revert to normal operation.
//       (this is done b/c customer can change poweroff limits file; this should be
//        the default behavior for the released software.
//   2:  diagnosis mode:  continually monitor and print complete flash contents
//                        and other information for each card seen.
//   3:  manufacturing config mode:  store and verify serial num, fudge factors, etc.
//                  (causes leds to flash under host control to indicate success;
//                   eventually will also automatically test card and calculate fudge
//                   factors)
//   4.  integration config mode:    store node numbers, names etc.
//   5:  networking config mode (unusual) to change timeouts, baud rate, etc.
//       continually on all cards seen.
//   6:  networking limits test mode:  Commands the cards to take on new
//       networking parameters in RAM, for the purpose of stress-testing the network
//     (finding the fastest possible baud rate and shortest timeouts). 
//      This is changed in card RAM only
//     so when it fails, powercycling the cluster will return the cards to 
//     a known working state.
//   7:  possibly an integration test?  ie power all on and power all off?


// all from Limits.c:

extern CHASSIS *ChassisList;    // pointer to the start of the chassis 
// linked list database.

extern CLUSTERNODE *ClusterList;        // pointer to the start of the chassis
// linked list database.

extern CLUSTERNODE *NWMaster;   // pointer to this master (running
// this software) as stored in the chassis list.


extern unsigned int NodeCount;  // number of nodes in linked list
extern unsigned int ChassisCount;       // number of chasses defined
extern unsigned int TempavCount;        // number of nodes containing defined tempaverage probes.
// (if tempaverage = 0 for a node, that node is not included in tempaveragecount)


extern QUEUE QueueResponse;

extern int my_ring_size;        // from netrecei.c, 
//  count of stored incoming serial data

void analogProcess ( void );

// struct _videoconfig videoConfig;

void print_usage ( FILE * stream, int exit_code, int error_msg_code,
                   char *str_argument )
{
  fprintf ( stream,
            "%s%s\nUsage: %s <Options>\n",
            usage_errors[error_msg_code], str_argument, program_name );

  fprintf ( stream,
            "  -h  --help\t\tDisplay this usage information\n"
            "  -d  --debug\t\tRun process in foreground, displaying limited debug info.\n"
            "  -D  --dump\t\tRun process in foreground, displaying limited debug info and nodewatch data.\n"
            "  -f  --flashid\t\tAssign card addresses interactively.  No other copy of %s may be running\n"
            "  -p  --path <path>\tSpecify the absolute path to the data and command files.\n"
            "  -s  --serial <full path>\tSpecify the serial port that nodewatch should run on.\n"
            "  -e  --expose\t\tExpose Email Errors.  Prints the email server responses.\n"
//   "  -m  --mail\t\tSends a test email and exits immediately.\n"
            "  -n  --noshutdown\tDisables automatic shutdowns\n"
            "  -N  --noaction\tDisables automatic emails and shutdowns\n"
            "  -v  --verbose\t\tVery Verbose.  (For Microway use/debugging only) (implies -d)\n",
            program_name );
  exit ( exit_code );
}


// copy path info to make all full filename constants
void configure_path (  )
{

  int path_lngth;
  int filename_lngth;
  //  info_msg("newpath ");
  //  info_msg(path_p);  // double check that the argv[] pointer was not permuted.

  // now set up all the filenames to be full ones:
  // check for valid directories, and for possible ending / in directory spec.
  // dataout file, command file, old cmd file, chassis config, cluster config, autonodeconf.
  // then merge the directory info to the front, making new filename strings.

  // either sprintf or strcpy can be used to compose this.
  // concerned that argv permutations may clobber the information.

  // therefore, will 1. check length limits on the arg and 
  //   2. copy the information.

  path_lngth = strlen ( path_p );
  if ( path_lngth < 1 )
    print_usage ( stderr, 1, 7, my_null_string );

  if ( path_lngth > PATH_LNGTH_LIM )
    print_usage ( stderr, 1, 6, path_p );

  if ( *( path_p + path_lngth - 1 ) == '/' )
    *( path_p + path_lngth - 1 ) = 0;   // clobber ending / if there.

  filename_lngth = path_lngth + 50;

  cmdfilename = malloc ( sizeof ( char ) * filename_lngth );
  movedfilename = malloc ( sizeof ( char ) * filename_lngth );
  clusterfilename = malloc ( sizeof ( char ) * filename_lngth );
  chassisfilename = malloc ( sizeof ( char ) * filename_lngth );
  autonodefilename = malloc ( sizeof ( char ) * filename_lngth );
  outfilename = malloc ( sizeof ( char ) * filename_lngth );

  sprintf ( cmdfilename, "%s/nodecommands.dat", path_p );
  sprintf ( movedfilename, "%s/nodecommands-process.dat", path_p );
  sprintf ( clusterfilename, "%s/clusterconf.dat", path_p );
  sprintf ( chassisfilename, "%s/chassisconf.dat", path_p );
  sprintf ( autonodefilename, "%s/autonodeconf.dat", path_p );
  sprintf ( outfilename, "%s/nodewatch.dat", path_p );

}


void displayOpen ( void )
{
  char tmp_error[400];
  info_msg ( "Welcome to Nodewatch\n  Copyright (c) 2003 Microway Inc. " );
  sprintf ( tmp_error, "Version %d.%d \n", VERSION_MAJOR, VERSION_MINOR );
  info_msg ( tmp_error );

  /*   _getvideoconfig(&videoConfig);
     _clearscreen(_GCLEARSCREEN);
     _settextcursor(0x2000);
     _settextposition(10, 6);
     _settextcolor(BRIGHTWHITE);
     _outtext("Slave Node 1"); */
}

void displayClose ( void )
{
  info_msg ( "Exiting.\n" );
  /*   _settextcolor(_WHITE);
     _settextcursor(0x0707);
     _clearscreen(_GCLEARSCREEN);
     _settextwindow(1, 1, videoConfig.numtextrows, videoConfig.numtextcols);
   */
}

int /* __cdecl */ main ( int argc, char *argv[] )
{
  unsigned char ucPriority;     // dummy char doesn't get used
  int next_option;
  //char *const short_options = "dhap:emvnN"; // p: means p takes an argument.
  char *const short_options = "dDhfp:evnNs:";     // p: means p takes an argument.
  struct option long_options[] = {
    {"debug", 0, NULL, 'd'},
    {"fulldebug", 0, NULL, 'D'},
    {"flashid", 0, NULL, 'f'},
    {"help", 0, NULL, 'h'},
    {"path", 1, NULL, 'p'},
    {"expose", 0, NULL, 'e'},
//    {"mail",       0, NULL, 'm'},
    {"verbose", 0, NULL, 'v'},
    {"noshutdown", 0, NULL, 'n'},
    {"noaction", 0, NULL, 'N'},
    {"serial", 1, NULL, 's'},
    {NULL, 0, NULL, 0}
  };

  path_p = default_path_p;
  program_name = argv[0];       //save program name for error msgs, etc

  do {
    next_option =
      getopt_long ( argc, argv, short_options, long_options, NULL );
    switch ( next_option ) {
       case 'D':               //options requests that process runs in
         nodaemon++;            //foreground with output
         dumpmetrics++;
         break;
       case 'd':
         nodaemon++;
         break;
       case 'e':               //options requests email output
         emailverbose++;
         break;
       case 'f':               //address assign
         nodaemon++;            //also, foreground with output
         flash_override = 10;
         break;
       case 'h':               //display usage and exit cleanly
         print_usage ( stdout, 0, 0, my_null_string );
         break;
		 case 's':				    //specify serial port
			if ( optarg == NULL )
				print_usage (stderr,1,3,my_null_string);
			if ( optarg[0] != '/' ) //must be absolute path
				print_usage ( stderr, 1, 5, optarg );
			if ( strlen(optarg) > MODEMDEVICELENGTH )
				print_usage ( stderr, 1, 8, optarg );
		
			strncpy(MODEMDEVICE,optarg,MODEMDEVICELENGTH);  //set MODEMDEVICE
			break;
       case 'p':               //change directory
         path_p = optarg;
         //      info_msg("argpath ");

         if ( optarg == NULL )  // must be absolute path
         {
           print_usage ( stderr, 1, 3, my_null_string );
         }
         if ( optarg[0] != '/' )        // must be absolute path
         {
           print_usage ( stderr, 1, 5, path_p );
         }
         info_msg ( path_p );

         break;

       case 'v':               //enable pt_dbug
         nodaemon++;
         mydbugverbose++;
         break;


       case 'n':               // disable auto_shutdown_node
         noshutdown++;
         break;


       case 'N':               //disable action_limits.
         noalertshutdown++;     //foreground with output
         break;


//    case 'm':               //send test email
//      send_test_email();
//      break;







       case '?':               //unknown option --display usage and exit with error
         print_usage ( stderr, 1, 1, my_null_string );
         break;
       case ':':               // missing argument--display usage and exit with error
         print_usage ( stderr, 1, 2, my_null_string );

       case -1:
         break;                 //all done parsing options
       default:                //not expected to occur
         abort (  );
    };
  } while ( next_option != -1 );

  configure_path (  );          // copy path info to make all full filename constants

  initmytimer (  );

  displayOpen (  );
  atexit ( displayClose );

  lineInitial (  );             // set conversion factors, set delays.
  netOpen (  );                 // This includes netCallBackInit()  // Spaghetti

  command_open (  );            // in limits.c.  Also reads in limits files.
  //  Above step reads in cluster database!
  if ( ClusterList == NULL ) {
    error_msg
      ( "NodeWatch:  Cluster List is empty. Check configuration files. Fatal error. Aborting.\n" );
    return ( -1 );
  }
  // The correct way to do this would have been for
  // command_open to return a return value with one of:
  //  OK(success), warning, error, fatal error.

  atexit ( netClose );

  setclusterAlive (  );         // Clear out alive data.

  init_flashmachine (  );       // sets up initial states for flash write/read state mchn

  //   netUnitAllocate(KEYBOARD, execKey,  DISABLE);


  /*   netUnitAllocate(U_NODE,   execNode, ENABLE);
     netUnitAllocate(U_NODE+1,   execNode, ENABLE);
     netUnitAllocate(U_NODE+2,   execNode, ENABLE);
     netUnitAllocate(U_NODE+3,   execNode, ENABLE);  */
  // Instead of netunitallocates, we'll just have a simple loop.

  // WE WILL NEED A LOOP HERE EVENTUALLY
  if ( !nodaemon ) {
    switch ( fork (  ) ) {
       case -1:
         perror ( "fork" );
         exit ( 1 );
       case 0:
         break;
       default:
         exit ( 0 );
    }
    setsid (  );
  }


  while ( TRUE ) {

    // Netprocess alternately sends out and receives packets.
    // Packets are sent either from the Message Queue,
    // if there are any pending packets there,
    // or by polling units for general status.
    // Once a non-poll packet is sent, the Netprocess
    // reverts to Receive mode, waiting up to 2 seconds
    // for a response.  While waiting/stuck in Receive mode,
    // no other polls are sent. ?? This is why when legitimate
    // packets are being transacted, the network might appear
    // slower than otherwise.

    netProcess (  );            // network layers

    // The following line dispatches incoming packets
    // to the appropriate incoming packet handler:

    // CAUTION: By eliminating pollunit(), I've 
    // prevented any possible extra message from the slaves from making
    // their way back to the host.  This is because, as we are now well aware,
    // the host state machine will not send a message from its transmit
    // queue unless it receives a poll from the master.

    flash_process (  );         // handle any flash read/write/configure issues based on state mchn.

    if ( flash_override == 0 ) {
      requestProcess (  );      // queue up a request for data.  Aliveness/deadness honored.
      command_process (  );     // in limits.c
    }
    // Queued cards marked as dead (adjustable value of 5)
    // requestProcess() is in execnode.c



    //while (netExecute(dequeue(&QueueResponse, &ucPriority)));

    while ( execNode ( dequeue ( &QueueResponse, &ucPriority ) ) );     // Changed to avoid netunits.
    // This above line will process all incoming data packets, setting aliveness.


    // keyProcess();

    if ( flash_override == 0 )
      analogProcess (  );       // In the new approach,
    // Analogprocess simply writes the file now and then.
    // At this point in the code, it is OK to report dead cards.  
    // They should have responded by now!

    if ( my_ring_size == 0 )    // if there are no pending serial port characters
      usleep ( DEFERTIME );     // defer and allow CPU time to others.
    // this will generally translate into 10000 microsecnds (1/100 sec)
    // because of the timescale of the UNIX processing scheduler interrupt.
  }
  return ( 0 );                 /* tell Linux no errors */
}

void minimal_net (  )           // this callback function allows flash or other systems to block and still
     // have the benefit of a basic network.  (Blocking functions must call this frequently)
{
  unsigned char ucPriority;     // dummy doesn't get used.

  netProcess (  );              // network layers
  while ( execNode ( dequeue ( &QueueResponse, &ucPriority ) ) );       // Changed to avoid netunits.

  if ( my_ring_size == 0 )      // if there are no pending serial port characters
    usleep ( DEFERTIME );       // defer and allow CPU time to others.
}

void minimal_net_purge (  )     // prepare to pause the network.
{
  int i;
  for ( i = 0; i < 40; i++ )
    netProcess (  );
  // Got to be a better way to do this...
}
