// Read a file parsing limits, node status and commands.
//


#include <string.h>
#include <assert.h>
#include <stdio.h>
#include <stdlib.h>
#include <ctype.h>
#include "cardconfig.h"
#include <time.h>
#include <pthread.h>
#include <math.h>
// Have to include in this order:
// cardconfig.h, net.h, exec.h, limits.h


#include "net.h"                // if you want to include exec.h, you have to
// include net.h
#include "exec.h"
#include "limits.h"
#include "power-ctl.h"
#include "timer.h"
#include "alert.h"
#include "codes.h"
#include "syslogging.h"

FILE *cmdfile;                  // file descriptor
extern char *cmdfilename;       // in host_l.c 
extern char *movedfilename;     // in host_l.c

extern int noalertshutdown;     // in host_l.c
extern int noshutdown;          // in host_l.c
extern int mydbugverbose;       // in host_l.c



static char *inFileName;
// OBSOLETE static char *limitsfilename = "/var/www/html/beostate/nodewatch/limits.conf";
// SHOULD BE CHANGED TO DATA DIRECTORY EVENTUALLY


// HERE ARE NEW FILENAMES:

extern char *clusterfilename;   // in host_l.c
// = "/var/www/html/beostate/data/clusterconf.dat";

extern char *chassisfilename;   // in host_l.c
// = "/var/www/html/beostate/data/chassisconf.dat";

extern char *autonodefilename;  // in host_l.c
// = "/var/www/html/beostate/data/autonodeconf.dat";


//
// The following text fragments are used to assemble email messages:

static char *limit_alert_shutdown[2] = { "Warning!\n",
  "URGENT MESSAGE! SYSTEM AUTOMATICALLY BEING SHUT DOWN due to following:\n"
};

static char *limit_range_id[2] = { "Normal Range", "Extended Normal Range" };

static char *limit_direction[2] = { "low", "high" };
static char *limit_measurement_type[3] = { "fan", "temperature", "voltage" };
static char *limit_units[3] = { "RPMs", "degrees C", "volts" };
static float limit_val_div[3] = { 1., 100., 100. };     // what to divide vals by

//static int   limit_val_decimal[3] = {0,2,2};  
// how many decimal places should be printed for each measurement?

int global_email_hash = -1;     // This value is used to describe previously
// sent emails, to prevent duplicate identical emails from being sent
// too rapidly.

long global_email_time = 0;     // This describes when the last email
// was sent, so reminder emails are sent on a controlled basis.

char mytimestring[50];

// We may need #defines or enums of direction standards and 
// measurement assignments (0=fan, 1=thermo, 2=volt)

// HOW WILL EUROPEAN CUSTOMERS RECEIVE CELSIUS MEASUREMENTS?
//  (both in web and in emails...)

//
extern int nodaemon;

extern long mytimestart;        // this is set up in timer.c

extern const char *program_name;        // pulled up in host_l.c

static int pausecount = 0;      // This causes a delay every 5 minutes, 


// Allocate the space for valid limit value information.
// one of the first things the program will do is read the limits
// file into this linked list:

CHASSIS *ChassisList = NULL;    // pointer to the start of the chassis 
// linked list database.

CLUSTERNODE *ClusterList = NULL;        // pointer to the start of the cluster
// linked list database.

CLUSTERNODE *NWMaster = NULL;   // pointer to this master (running
// this software) as stored in the chassis list.

CLUSTERNODE *FailingNode = NULL;        // Points to the node most recently 
// seen as having issues.  This node gets partial higher priority treatment
// for requesting additional measurements. 

// global flags to signal need for further limits checks:
int global_alert_flag = 0;
int global_shutdown_flag = 0;


//
///// EMAIL ASSEMBLY STRINGS and related variables:

/// Content for emails (server, addresses, body)

/// Control structure for email:

pthread_t email_thread;

long email_transmit_time;       // when the thread spun off would timeout.


struct nw_email the_emailmsg = { NULL,  // (server)
  NULL,                         // (source_email)
  NULL,                         // (addresses)  email_addresses is runtime allocated.
  NULL,                         // (subject)
  NULL,                         // (body) have to add in the actual body 
  ALERT_STATUS_IDLE
};

//??? will have to load in the source, destination address, subject, and body pointers.


/// Buffers for constructing together future emails, even while email above is being sent:

char *email_buf_start = NULL;   // the section where the new message is assembled.
char *email_buf = NULL;         // pointer to the end of the new message area, 
// where more text may be added.

int email_mem_total = 0;        // the total amount of memory (bytes) allocated to new_email_segment
int email_mem = 0;              // The amount of memory remaining for (email_buffer_start + email_alloc_total - email_buf)


/// Email alert constants:

char *email_addresses = NULL;   // complete email address(es) to send alerts, space separated.

// char *default_server = "64.80.227.9";
char *email_server = NULL;      // NOW YOU HAVE TO PUT THE INFO IN THE FILE.
// char *default_sender = "nobody-nodewatch@microway.com";
char *email_sender = NULL;


const char *email_default_subject =
  "AUTOMATIC WARNING from NodeWatch(tm) System";
const char *email_shutdown_subject =
  "URGENT:  AUTOMATIC SHUTDOWN FROM NODEWATCH(TM) SYSTEM!";

//??? SHOULD SUBJECT LINES HAVE ENDING NEWLINES???  WILL THAT MAKE A DIFFERENCE IN THE EMAIL TRANSMIT PROGRAM?

//??? will have to change to comma separated.

int emailrepeat = 1440 * 60;    // Repeat emails for ongoing alert situations every day
// (one day has 1440 minutes.)


int alertignore = 0;            //the clusterconf.dat file may specify that
// all alerts should be ignored (initial setup purposes)
// 0= process alerts (default) 1=ignore

int oldalert = 300;             // Send email regarding old or unreported data if it 
// is more than 5 minutes (300 seconds) old.
// This implies an initial 5 minute delay before pursuing this, 
// to first check for "no data"

int stalemail = 300;            // any email thread that does not come back in 
// five minutes is killed, and possibly resent.

//
// When an alert or shutdown situation has resolved itself, all of
// FailingNode, global_alert_flag and global_shutdown_flag are 
// reset to null, 0 and 0, respectively.


unsigned int NodeCount = 0;     // number of nodes in linked list
unsigned int ChassisCount = 0;  // number of chasses defined
unsigned int TempavCount = 0;   // number of nodes containing defined tempaverage probes.
// (if tempaverage = 0 for a node, that node is not included in tempaveragecount)


// It is also possible for the master to find out which card
// it has, by using the IP address!

char *my_null_string = "";      // this is to clean up NULL string pointers
// have them all point here instead!  cannot be extern const
// because that's not the type it's used to stand in for.

//
// Following debug function will print to the error channel the 
// listed information.

void pt_dbug(int linenmb, char *desc_string, char *the_string,
               void *val_pointer )
{
  char tempstring[400];
  if ( mydbugverbose ) {
    sprintf ( tempstring, "DEBUG LINE %d %s %s %p\n", linenmb, desc_string,
              the_string, val_pointer );
    error_msg ( tempstring );
    fflush(stdout);
  }
}

void pt_dbug2 ( int class,
                int linenmb,
                char *desc_string, int thevalue, void *val_pointer )
{
  char tempstring[400];
  if ( mydbugverbose ) {
    sprintf ( tempstring, "DEBUG LINE %d %s %d %p\n", linenmb, desc_string,
              thevalue, val_pointer );
    error_msg ( tempstring );
    fflush(stdout);
  }
}

/*void pt_dbugm ( int class, int linenmb, char *loc_strg,
                char *desc_string1, char *thestring1,
                char *desc_string2, char *thestring2,
                char *desc_int1, int theval1, char *desc_int2, int theval2,
                char *desc_pt1, int val_pt1, char *desc_pt2, int val_pt2 )
{*/
  /*
     char tempstring[8000];  //???  WATCH OUT FOR MY OWN FAULT HERE!
     sprintf(tempstring, "DEBUG STRG Ln%d %s |%s %s |%s %s |%s %d |%s %d |%s %x |%s %x\n",
     linenmb, loc_strg, 
     desc_string1, thestring1, desc_string2, thestring2,
     desc_int1, theval1, desc_int2, theval2,
     desc_pt1, val_pt1, desc_pt2, val_pt2);
     error_msg(tempstring);

     } */

/////////////////////  We need a new storage mechanism.
// that accomodates multiple chasses.

//  It takes too long to search by matching text files every single
// time the nodes database wants to seek out the limits information
// particular to its chassis.  Therefore the database on the nodes,
// should include a pointer to the entry in the chassis database 
// containing the information for that node.

// to allow cards to reset.

//
//whitespace_remove    MLJ
// Inputs:  character string pointer
// Outputs:  character string pointer with initial (left) whitespace removed.
// If string is completely blank or whitespace or comment, outputs pointer
// pointer to my_null_string



char *whitespace_remove ( char *instring )
{
  pt_dbug(__LINE__, "whitespace_remove start", instring, instring );

  if ( isNull ( instring ) )
    return ( my_null_string );  // check for null input pointer.
  while ( ( *instring == ' ' ) ||
          ( *instring == '\t' ) || ( *instring == '\n' ) )
    instring++;                 // skip over any whitespace
  if ( ( *instring == '!' ) || ( *instring == NUL ) )   // if a comment or the end of the string,
    return ( my_null_string );  // return null pointer.

  pt_dbug(__LINE__, "whitespace_remove non-null end", instring, instring );

  return ( instring );
}

//
// string_separate     MLJ
//  Inputs:   character string pointer
//  Side effect:  injects \0 null character
//     into string after first whitespace-separated entry
//  Outputs:  character string pointer pointing to next whitespace-separated
//    entry.  Outputs my_null_string pointer if no more entries or only
//    whitespace or only comments.

//  Purpose: allows arbitrary separation of space-separated line entries.

char *string_separate ( char *instring )
{
  char *dbug_storestart = instring;     // because this fcn has side effects.

  pt_dbug(__LINE__, "string_separate start", instring, instring );

  if ( isNull ( instring ) )
    return ( my_null_string );  // check for null input pointer.

  while ( ( *instring == ' ' ) ||
          ( *instring == '\t' ) || ( *instring == '\n' ) )
    instring++;                 // skip over any whitespace

  if ( ( *instring == '!' ) || ( *instring == NUL ) )   // if a comment or the end of the string,
    return ( my_null_string );  // return my_null_string pointer.

  while ( ( *instring != ' ' ) &&
          ( *instring != '\t' ) &&
          ( *instring != '\n' ) &&
          ( *instring != NUL ) && ( *instring != '!' ) )
    instring++;                 // skip over the first entry.

  if ( ( *instring == '!' ) || ( *instring == NUL ) ) {
    *instring = NUL;
    pt_dbug(__LINE__, "string_separate secondnull", dbug_storestart,
              dbug_storestart );
    return ( my_null_string );
  }                             // If a comment or the end of first string, return null pointer and 
  // shorten the existing string if necessary.

  *instring = NUL;              // terminate existing first string
  pt_dbug(__LINE__, "string_separate firstfound", dbug_storestart,
            dbug_storestart );
  instring++;                   // now go on to locate next string.

  while ( ( *instring == ' ' ) ||
          ( *instring == '\t' ) || ( *instring == '\n' ) )
    instring++;                 // skip over any whitespace

  if ( ( *instring == '!' ) || ( *instring == NUL ) )   // if a comment or the end of the string,
    return ( my_null_string );  // return null pointer.

  pt_dbug(__LINE__, "string_separate secondfound", instring, instring );

  return ( instring );          // return rest of the string (one or more entries)
}


//
// parseNumericLine    MLJ
//  inputs:    instring:   a line of space-separated numbers, can be clobbered.

//             lineNum:    line number being processed, to be used for error msgs
//             numArray:   pointer to first in a series of ints to receive the vals
//             maxNums:    maximum number of numbers that can be stored.
//             errorString: the kind of data being stored to be used in 
//                   non-numeric or overrun error messages.

char parseNumericLine ( char *instring, int lineNum, int *numArray,
                        int maxNums, char *errorString, char *fileName )
{
  char tempstring[STRING_LENGTH];
  char *curstringptr;
  while ( maxNums > 0 )         // has to be at least 1
  {
//      pt_dbug(__LINE__,"parseNumericLine", "maxNums loop:", (void *)maxNums);

    curstringptr = instring;    // remember where we are.
    instring = string_separate ( curstringptr );
    pt_dbug(__LINE__, "parseNumericLine curstringptr", curstringptr,
              curstringptr );
    pt_dbug(__LINE__, "parseNumericLine instring", instring, instring );


    // now the curstringptr should have a direct view of the first number

    if ( isNull ( curstringptr ) ) {
      return ( 2 );             // string line too short (often OK).
    }

    else if ( isNumber ( curstringptr ) ) {
      // SIGNED OR UNSIGNED???
      sscanf ( curstringptr, " %u", numArray );
      pt_dbug(__LINE__, "parseNumericLine numArray ptr", curstringptr,
                numArray );
//        pt_dbug(__LINE__,"parseNumericLine numArray val", curstringptr, (void *)*numArray);
      numArray++;
      maxNums--;
    }
    // DO NOT TRY TO USE SHORTS.  DO NOT MIX UP POINTER TYPES.
    else {
      sprintf ( tempstring, "ERROR %s line %d: %s has non-numeric entry\n",
                fileName, lineNum, errorString );
      exit_msg ( tempstring );
      return ( 1 );             // not a number
    }
  }

  return ( 0 );                 // success

}

//
signed int parseStringLine ( char *instring, int lineNum, char **stringArray,
                             int maxNums, char *errorString, char *fileName )
{
  char tempstring[STRING_LENGTH];
  char *curstringptr;
  int maxNumsSave = maxNums;
  while ( maxNums > 0 )         // has to be at least 1
  {

//      pt_dbug(__LINE__,"parseStringLine", "maxNums loop:", (void *)maxNums);

    curstringptr = instring;    // remember where we are.
    instring = string_separate ( curstringptr );        //string_separate adds \0's to mark params

    pt_dbug(__LINE__, "parseStringLine curstringptr", curstringptr,
              curstringptr );
    pt_dbug(__LINE__, "parseStringLine instring", instring, instring );


    // now the curstringptr should have a direct view of the first number

    if ( isNull ( curstringptr ) ) {
      return ( maxNumsSave - maxNums ); // string line too short (often OK).
    }

    else if ( isNull ( *stringArray ) ) {

      *stringArray =
        malloc ( sizeof ( char ) * ( strlen ( curstringptr ) + 1 ) );
      if ( *stringArray == NULL ) {
        sprintf ( tempstring, "ERROR %s line %d: %s no memory to store!\n",
                  fileName, lineNum, errorString );
        exit_msg ( tempstring );
        return ( -255 );
      } else {
        strcpy ( *stringArray, curstringptr );  // insert the string.
        pt_dbug(__LINE__, "parseStringLine stringArray val", *stringArray,
                  ( void * ) *stringArray );
        pt_dbug(__LINE__, "parseStringLine stringArray ptr", *stringArray,
                  stringArray );
      }

      stringArray++;
      maxNums--;


      /*
         sprintf(tempstring, "NON-ERROR %s line %d: %s: redundant entry %s: ignored inserting %s.\n", 
         fileName, lineNum, errorString, *stringArray, curstringptr );
         // this is necessary to avoid a memory leak!
         error_msg(tempstring);
       */

      // Above test didn't work quite as expected regarding string pointer appearance.



    } else {
      sprintf ( tempstring,
                "ERROR %s line %d: %s: redundant entry %s: ignored inserting %s.\n",
                fileName, lineNum, errorString, *stringArray, curstringptr );
      // this is necessary to avoid a memory leak!
      error_msg ( tempstring );
      return ( -254 );
    }
  }
  return ( maxNumsSave );       // success
}

//
/* Decide if a string is completely blank
 *
 *Inputs:
 *	string		string to test
 * Returns:
 *	true if the string is blank
 */

static boolean isBlank ( char *string )
{
  int i;
  i = 0;
  if ( string == NULL )
    return true;                // WHY????  WHY???????????????????!!!!!!!!!!!!!
  while ( string[i] != NUL ) {
    switch ( string[i] ) {
       case '\n':
       case ' ':
       case '\t':
         break;
       default:
         return false;
    };
    i += 1;
  };
  return true;
}

// Following routine finds out if string is null in several possible ways.
boolean isNull ( char *string )
{
  pt_dbug(__LINE__, "isNull que", string, string );
//  pt_dbug(__LINE__,"isNull ans", string, 
//        (void *)((string == NULL) ||
//              (string == my_null_string) ||
//              (*string == NUL)));
  return ( ( string == NULL ) ||
           ( string == my_null_string ) || ( *string == NUL ) );
}


/* Decide if a string contains a number possibly preceeded or followed by
 * blanks
 *
 *Inputs:
 *	string		string to test
 * Returns:
 *	true if the string is a number or empty
 ****** WHY DOES THIS FCN RETURN TRUE IF IT'S EMPTY???? WELL, CAN'T CHANGE IT.
 */
boolean isNumber ( char *string )
{
  int i;
  boolean anydigit;
  boolean midblank;
  anydigit = false;
  midblank = false;
  for ( i = 0; string[i] != NUL; i += 1 ) {
    switch ( string[i] ) {
       case '\n':
       case ' ':
       case '\t':
         if ( anydigit )
           midblank = true;
         break;
       case '0':
       case '1':
       case '2':
       case '3':
       case '4':
       case '5':
       case '6':
       case '7':
       case '8':
       case '9':
         if ( midblank )
           return false;        /* digit(s) blank(s) digit */
         anydigit = true;
         break;
       default:
         return false;
    };
  };
  return true;
}


/* Check if line is a comment
 *
 * Inputs:
 *	line		line to check
 * Returns:
 *	true if first nonblank is a '!'
 */
static boolean isComment ( char *line )
{
  int i;
  for ( i = 0; line[i]; i += 1 ) {
    if ( line[i] == '\t' )
      continue;
    if ( line[i] == ' ' )
      continue;
    if ( line[i] == '!' )
      return true;
  };
  return false;
}


// Determine if a line is a new section head. Report the number of
// leading dollar signs if it is.
static int newSection ( char *string )
{
  int i;
  int numberDollars = 0;
  for ( i = 0; string[i] != NUL; i += 1 ) {
    if ( isspace ( string[i] ) )
      continue;
    if ( string[i] != '$' )
      break;
    numberDollars += 1;
  };
  return numberDollars;
}


/************ Obsoleted

// Parse a limits line $ALERT or $SHUTDOWN in the form:
//
//  OLD FMT:    $ALERT * <type> <min> <max>
//  NEW FMT:    $ALERT|SHUTDOWN <channel> <type> <min> <max>


// Inputs:
//      line            line to parse
//      lineNumber
//      isAlert         true for ALERT line false for SHUTDOWN line
static void parseLimitsLine ( char *line, int lineNumber, boolean isAlert ) 
{
  char dummy[STRING_LENGTH];
  // This is a pointer to an automatically-allocated area 
  // of memory holding characters.
  char typeName[STRING_LENGTH];
  int minimum = 0;
  int maximum = 0;
  int channelNum;
  int machnum; // for now, slaves vs. master.

  int matches;
  struct limits *pLimits;
  matches =
    sscanf ( line, " %s %d %s %d %d %d", dummy, &channelNum, typeName, &minimum, &maximum,&machnum );

  // The following code is backwards compatible with limits file not including machnum
  // at the end of the line.  If missing, it's assumed to be 0, representing all slaves.

  if (matches == 5) // there is no machnum
    {
      machnum = 0;
    } 
  else if ( matches != 6 ) {
    char tempstring[400];
    sprintf(tempstring, "ERROR %s %d: $ALERT or $SHUTDOWN line has incorrect format\n", inFileName, lineNumber );
    error_msg(tempstring);
    return;
  };
  
  // Now store results in fanLimits, tempLimits, and voltLimits.

  //OBSOLETED
  if ( 0 == stricmp ( typeName, "fans" ) )
    {
      // might be nice to check ranges.
      // also, does the channel num start at 0 or 1?
      fanLimits[channelNum][0][isAlert][machnum] = minimum;
      fanLimits[channelNum][1][isAlert][machnum] = maximum;
    }
  else if ( 0 == stricmp ( typeName, "temperature" ) )
    {
      tempLimits[channelNum][0][isAlert][machnum] = minimum;
      tempLimits[channelNum][1][isAlert][machnum] = maximum;
    }
  else if ( 0 == stricmp ( typeName, "power" ) ) 
    {
      voltLimits[channelNum][0][isAlert][machnum] = minimum;
      voltLimits[channelNum][1][isAlert][machnum] = maximum;
    }
  else 
    {
      char tempstring[400];
      sprintf ( tempstring, "WARN %s %d: Unrecognized limit type %s\n",
		inFileName, lineNumber, typeName );
      error_msg(tempstring);
      return;
    };
  return;
}

OBSOLETED

*****/

//
// Parse file containing limit data. Callback functions to parse
// sections as they are read.
//
// Inputs:
//      inFile          open file to read if NULL try to open inName
//      inName          if the inFile is NULL try to read this one


// FORMAT OF THE LIMIT FILE:
//  ${ALERT|SHUTDOWN} {channelnum}  {fans|power|temperature} {lowerlim}  {upperlim} {machinenum}

// Note that the limit file accomodates different levels for different channels
// (fans or temps) but it doesn't accomodate special levels for the master,
// if the master is of a different case type.

// so, to recap, for every combination of case type (master case type and 
//  slaves case type):  * limits.dat file; * header file for the labels for php.

// The original version of the following function uses
// a single namespace for every kind of file, be it a limit file,
// a command file, or any other configuration file.



void parseFile ( FILE * inFile, char *inName )
{
  int lineNumber;
  int matches;
  int versionNumber = 1;
  char line[STRING_LENGTH];
  char fieldName[STRING_LENGTH];
  char nodeName[STRING_LENGTH];
  char value1[STRING_LENGTH];
  inFileName = inName;


  if ( inFile == NULL ) {
    if ( inFileName != NULL ) {

      // Quietly ignore missing file
      inFile = fopen ( inFileName, "r" );
      if ( inFile == NULL ) {
        inFile = NULL;
      };
    };
  };

  if ( inFile == NULL ) {
    // error_msg("Could not open the inFile");
    return;
  }
  lineNumber = 0;
  while ( NULL != fgets ( line, STRING_LENGTH, inFile ) ) {
    lineNumber += 1;
    if ( line[0] != NUL )
      line[strlen ( line ) - 1] = NUL;  /* remove \n */
    if ( isComment ( line ) ) {
      continue;
    };
    if ( isBlank ( line ) ) {
      continue;
    };
    if ( newSection ( line ) == 1 ) {
      serviceNewSection (  );
    };
    matches = sscanf ( line, " %s %s %s", fieldName, nodeName, value1 );

    // Please note:  this sscanf is just for initial diagnostics on
    // what the line is.  If it's a limits line for instance, 
    // a separate parseLimitsLine will be called which re-does
    // the sscanf on the whole line string

    if ( matches < 1 )
      continue;
    if ( 0 == stricmp ( fieldName, "$VERSION" ) ) {
      sscanf ( nodeName, " %d", &versionNumber );
      if ( versionNumber != 1 ) {
        char tempstring[400];
        sprintf ( tempstring, "WARN: Unexpected file version number %d\n",
                  versionNumber );
        error_msg ( tempstring );
      };
    };
    /***************** OBSOLETED 

    if ( 0 == stricmp ( fieldName, "$ALERT" ) ) {
      parseLimitsLine ( line, lineNumber, true );
    };
    if ( 0 == stricmp ( fieldName, "$SHUTDOWN" ) ) {
      parseLimitsLine ( line, lineNumber, false );
    };
 
                       OBSOLETED    ******************/

    serviceFieldLine ( line, lineNumber, matches, fieldName, nodeName,
                       value1 );
  };
  return;
}

//
////////////////// Here: the namespace for interpretation of each file
////////////////// is separated into routines that handle that file.

int parseChassisFile ( char *inName )
{
  // Mike's routine to read in entire chassis information.

  char tempstring[400];         // for errors

  FILE *inFile;                 // File descriptor

  int lineNumber;
  int versionNumber = 1;
  char origline[STRING_LENGTH]; //TEMPORARY STATIC SPACE FOR THE LINE IS HERE.

  char *line;                   //pointer to the current part of the line being analyzed.
  char *nextfield;              // pointer to the subsequent part of the line
  // from string_separate.

  int i;


  // The following keeps track of parameters that hold their value
  // until changed in the file:  (along with default values)
  int shutdowncount = 4;
  int alertcount = 4;
  int tempaverage = 1;

  CHASSIS **nextchassisptr;     // Pointer to the pointer holding current chassis info.
  CHASSIS *curchassis = NULL;   // the current chassis being stored.
  // with this approach, EVERY time curchassis is used, 
  // have to check if it's null first!  b/c the file might try to store
  // stuff in a chassis before declaring one!!

  ChassisCount = 0;             // reset chassis count

  nextchassisptr = &ChassisList;        // This allows the first new chassis
  // to be allocated and stored at the top of the list.

  // not sure why the string was copied.  
  // Could fopen be destructive on the string? No.  Got rid of it!

  inFile = NULL;                // initially mark the file as not being there.

  if ( inName != NULL ) {
    inFile = fopen ( inName, "r" );
  };

  if ( inFile == NULL ) {
    exit_msg ( "Missing chassis description file.\n" );
    return 1;
  };

  lineNumber = 0;
  while ( fgets ( origline, STRING_LENGTH, inFile ) != NULL ) {
    lineNumber += 1;
    if ( origline[0] != NUL )
      origline[strlen ( origline ) - 1] = NUL;  /* remove \n */
    if ( isComment ( origline ) ) {
      continue;
    };
    if ( isBlank ( origline ) ) {
      continue;
    };
    //  if ( newSection ( origline ) == 1 ) {
    //  serviceNewSection (  );
    // };

    // Now to use the new parsing routines to break up the string:
    //

    pt_dbug(__LINE__, "parseChassisFile origline", origline, origline );

    line = whitespace_remove ( origline );      // chop off any spaces 
    // on the left.
    // cannot use case/switch because string equality required in each case.

    nextfield = string_separate ( line );       // this causes
    // line to be only the first entry, by way of the side effect.

    pt_dbug(__LINE__, "parseChassisFile line", line, line );
    pt_dbug(__LINE__, "parseChassisFile nextfield", nextfield, nextfield );

    // At this point, line contains the first string (identifier)
    // and nextfield contains all the arguments if any.
    // Although line should not be null, (we tested for
    // blankness and comments earlier) nextfield might be.

    // Correction:  empty strings are always pointed to the global 
    // constant string called my_null_string.  This prevents dereferencing
    // a real null pointer, and allows equality testing to my_null_string.


    if ( 0 == stricmp ( line, "$VERSION" ) ) {
      parseNumericLine ( nextfield, lineNumber, &versionNumber, 1, "Version",
                         inName );

      if ( versionNumber != 1 ) {

        sprintf ( tempstring,
                  "WARN: %s line %d: Unexpected file version number %d\n",
                  inName, lineNumber, versionNumber );
        error_msg ( tempstring );
      }
    }

    else if ( 0 == stricmp ( line, "$SHUTDOWN_COUNT" ) ) {
      parseNumericLine ( nextfield, lineNumber, &shutdowncount, 1,
                         "Shutdown_count", inName );
      if ( shutdowncount < 1 ) {

        sprintf ( tempstring,
                  "%s line %d: Shutdown count must be at least 1\n", inName,
                  lineNumber );
        error_msg ( tempstring );
        shutdowncount = 1;
      }
    } else if ( 0 == stricmp ( line, "$ALERT_COUNT" ) ) {
      parseNumericLine ( nextfield, lineNumber, &alertcount, 1,
                         "Alert_count", inName );

      if ( alertcount < 1 ) {
        sprintf ( tempstring, "%s line %d: Alert count must be at least 1\n",
                  inName, lineNumber );
        error_msg ( tempstring );
        alertcount = 1;
      }
    }


    else if ( 0 == stricmp ( line, "$TEMPAVERAGE" ) ) {
      parseNumericLine ( nextfield, lineNumber, &tempaverage, 1,
                         "Tempaverage", inName );
    }

    else if ( 0 == stricmp ( line, "$CHASSISTYPE" ) ) {
      // New chassis entry!  Get things set up...
      curchassis = calloc ( 1, sizeof ( CHASSIS ) );
      pt_dbug(__LINE__, "parseChassisFile new chassis! curchassis", "",
                curchassis );
      if ( curchassis == NULL ) {
        exit_msg ( "Could not allocate memory for new chassis!\n" );
        break;
      }
      // remember that nextchassis is a pointer to a pointer to a chassis.
      else                      // memory was allocated OK, now init new chassis.
      {
        ChassisCount++;
        *nextchassisptr = curchassis;
        nextchassisptr = ( CHASSIS ** ) & ( curchassis->nextchassis );
//        pt_dbug(__LINE__,"parseChassisFile new chassis! *nextchassisptr","",(int)*nextchassisptr);
        pt_dbug(__LINE__, "parseChassisFile new chassis! nextchassisptr", "",
                  nextchassisptr );

        //?????????????????????/ should the & be there?
        // yes, we are making a pointer to the pointer to future chasses.

        curchassis->nextchassis = NULL;

        // Above: First we record within the linked list the pointer to the 
        // new chassis allocated.
        // Next we update the pointer to the pointer, to point to the link
        // to the next chassis located within the current chassis.

        //      Now initialize the allocated memory:

        curchassis->name = my_null_string;      // This was a bug. 
        // have to do this before doing the parseStringLine below!


        if ( isNull ( nextfield ) ) {
          exit_msg ( "No chassis name listed!\n" );
          //      break;  // Have to continue to initialize the chassis.
        } else {
          parseStringLine ( nextfield, lineNumber, &( curchassis->name ),
                            1, "Chassis name", inName );
        }

        // now initialize and clear things out:
        curchassis->tempaverage = tempaverage;
        curchassis->shutdowncount = shutdowncount;
        curchassis->alertcount = alertcount;
        //??? ALERTCOUNT AND SHUTDOWNCOUNT ARE NOT BEING UTILIZED
        //??? IN THE ALGORITHM!!!!!!    


        for ( i = 0; i < MAX_TACK; i++ ) {
          curchassis->fanlabels[i] = my_null_string;
          curchassis->tackarray[i] = i; // dumb default is to count up.
          curchassis->fanalertlo[i] = 0;
          curchassis->fanalerthi[i] = 25000;
          curchassis->fanshutdownlo[i] = 0;
          curchassis->fanshutdownhi[i] = 50000;
          curchassis->fangroup[i] = i;  // assign default groups.
          curchassis->fangroupfail[i] = 0;
          // The default fan group system
          // assumes that user-defined fan groups < i.
        }

        for ( i = 0; i < MAX_TEMP; i++ ) {
          curchassis->templabels[i] = my_null_string;
          curchassis->thermoarray[i] = i + 3;   // dumb default is to count up.
          // 3 is added so it is above the volt measurements.
          curchassis->tempalertlo[i] = 0;
          curchassis->tempalerthi[i] = 250;
          curchassis->tempshutdownlo[i] = 0;
          curchassis->tempshutdownhi[i] = 500;
        }

        for ( i = 0; i < MAX_VOLT; i++ ) {
          curchassis->voltlabels[i] = my_null_string;
          curchassis->voltarray[i] = i; // dumb default is to count up.
          curchassis->voltalertlo[i] = 0;
          curchassis->voltalerthi[i] = 100;
          curchassis->voltshutdownlo[i] = 0;
          curchassis->voltshutdownhi[i] = 200;
        }
        chassis_init_more ( curchassis );       // assign additional default variables.        
      }
    }
    // All the following depend on a valid chassis entry.
    // So first check that the pointer is not null!
    else if ( curchassis != NULL ) {
      if ( 0 == stricmp ( line, "$$FANLABELS" ) ) {
        curchassis->numfans =
          parseStringLine ( nextfield, lineNumber, curchassis->fanlabels,
                            MAX_TACK, "$$FANNAMES", inName );
        pt_dbug(__LINE__, "rww parse fan labels- curchassis->name",
                  curchassis->name, NULL );
        pt_dbug2(2,__LINE__, "rww parse fan labels- curchassis->numfans",
                   curchassis->numfans, NULL );
      }

      else if ( 0 == stricmp ( line, "$$TEMPLABELS" ) ) {
        parseStringLine ( nextfield, lineNumber, curchassis->templabels,
                          MAX_TEMP, "$$TEMPNAMES", inName );
      }

      else if ( 0 == stricmp ( line, "$$VOLTLABELS" ) ) {
        parseStringLine ( nextfield, lineNumber, curchassis->voltlabels,
                          MAX_VOLT, "$$VOLTNAMES", inName );
      }

      else if ( 0 == stricmp ( line, "$$TACKARRAY" ) ) {
        parseNumericLine ( nextfield, lineNumber, curchassis->tackarray,
                           MAX_TACK, "$$TACKARRAY", inName );
      }
      //???????????????????  should it be curchassis->tackarray
      //  or &(curchassis->tackarray)
      // it should be curchassis->tackarray.  Although there is no
      // pointer stored in the structure itself, curchassis->tackarray
      // generates a pointer to the start of the tackarray.


      else if ( 0 == stricmp ( line, "$$FANALERTLO" ) ) {
        parseNumericLine ( nextfield, lineNumber, curchassis->fanalertlo,
                           MAX_TACK, "$$FANALERTLO", inName );
      }

      else if ( 0 == stricmp ( line, "$$FANALERTHI" ) ) {
        parseNumericLine ( nextfield, lineNumber, curchassis->fanalerthi,
                           MAX_TACK, "$$FANALERTHI", inName );
        sprintf ( tempstring, "%s %u %u %u %u %u", curchassis->name,
                  curchassis->fanalerthi[0], curchassis->fanalerthi[1],
                  curchassis->fanalerthi[2], curchassis->fanalerthi[3],
                  curchassis->fanalerthi[4] );
        pt_dbug(__LINE__,
                  "rww reading in fanalerthi- curchassis->name and ->fanalerthi[0-4]",
                  tempstring, NULL );
      }

      else if ( 0 == stricmp ( line, "$$FANSHUTDOWNLO" ) ) {
        parseNumericLine ( nextfield, lineNumber, curchassis->fanshutdownlo,
                           MAX_TACK, "$$FANSHUTDOWNLO", inName );
      }

      else if ( 0 == stricmp ( line, "$$FANSHUTDOWNHI" ) ) {
        parseNumericLine ( nextfield, lineNumber, curchassis->fanshutdownhi,
                           MAX_TACK, "$$FANSHUTDOWNHI", inName );
      }


      else if ( 0 == stricmp ( line, "$$VOLTALERTLO" ) ) {
        parseNumericLine ( nextfield, lineNumber, curchassis->voltalertlo,
                           MAX_VOLT, "$$VOLTALERTLO", inName );
      }

      else if ( 0 == stricmp ( line, "$$VOLTALERTHI" ) ) {
        parseNumericLine ( nextfield, lineNumber, curchassis->voltalerthi,
                           MAX_VOLT, "$$VOLTALERTHI", inName );
      }

      else if ( 0 == stricmp ( line, "$$VOLTSHUTDOWNLO" ) ) {
        parseNumericLine ( nextfield, lineNumber, curchassis->voltshutdownlo,
                           MAX_VOLT, "$$VOLTSHUTDOWNLO", inName );
      }

      else if ( 0 == stricmp ( line, "$$VOLTSHUTDOWNHI" ) ) {
        parseNumericLine ( nextfield, lineNumber, curchassis->voltshutdownhi,
                           MAX_VOLT, "$$VOLTSHUTDOWNHI", inName );
      }

      else if ( 0 == stricmp ( line, "$$TEMPALERTLO" ) ) {
        parseNumericLine ( nextfield, lineNumber, curchassis->tempalertlo,
                           MAX_TEMP, "$$TEMPALERTLO", inName );
      }

      else if ( 0 == stricmp ( line, "$$TEMPALERTHI" ) ) {
        parseNumericLine ( nextfield, lineNumber, curchassis->tempalerthi,
                           MAX_TEMP, "$$TEMPALERTHI", inName );
      }

      else if ( 0 == stricmp ( line, "$$TEMPSHUTDOWNLO" ) ) {
        parseNumericLine ( nextfield, lineNumber, curchassis->tempshutdownlo,
                           MAX_TEMP, "$$TEMPSHUTDOWNLO", inName );
      }

      else if ( 0 == stricmp ( line, "$$TEMPSHUTDOWNHI" ) ) {
        parseNumericLine ( nextfield, lineNumber, curchassis->tempshutdownhi,
                           MAX_TEMP, "$$TEMPSHUTDOWNHI", inName );
      }


      else if ( 0 == stricmp ( line, "$$FANGROUP" ) ) {
        parseNumericLine ( nextfield, lineNumber, curchassis->fangroup,
                           MAX_TACK, "$$FANGROUP", inName );
      }

      else if ( 0 == stricmp ( line, "$$FANGROUPFAIL" ) ) {
        parseNumericLine ( nextfield, lineNumber, curchassis->fangroupfail,
                           MAX_TACK, "$$FANGROUPFAIL", inName );
      }
      // ---- insert more cases to be stored in a chassis here ---


      else {
        sprintf ( tempstring,
                  "WARN: %s line %d ignored -- unrecognized heading\n",
                  inName, lineNumber );
        error_msg ( tempstring );
      }
    }

    else {
      sprintf ( tempstring,
                "WARN: %s line %d ignored -- no prior chassis declaration\n",
                inName, lineNumber );
      error_msg ( tempstring );
    }


  }                             // end of main while loop

  fclose ( inFile );
  return 0;
}

//
int parseClusterFile ( char *inName )
{
  // Mike's routine to read in entire cluster information.

  char tempstring[400];         // for errors

  FILE *inFile;                 // File descriptor

  int lineNumber;
  int versionNumber = 1;
  char origline[STRING_LENGTH]; //TEMPORARY STATIC SPACE FOR THE LINE IS HERE.

  char *line;                   //pointer to the current part of the line being analyzed.
  char *nextfield;              // pointer to the subsequent part of the line
  // from string_separate.

  CLUSTERNODE **nextnodeptr;    // Pointer to the pointer holding current node info.
  CLUSTERNODE *curnode = NULL;  // the current node being stored.
  // with this approach, EVERY time curnode is used, 
  // have to check if it's null first!  b/c the file might try to store
  // stuff in a chassis before declaring one!!

  CHASSIS *chassissearch;       // variable that searches through the chasses.
//  char * charcleanout; // used to clean out the node data.
  int i;

  NodeCount = 0;                // start fresh with node count.

  nextnodeptr = &ClusterList;   // This allows the first new node
  // to be allocated and stored at the top of the list.

  NWMaster = NULL;              // clear out the idea of which machine 
  // is the nodewatch master.

  inFile = NULL;                // initially mark the file as not being there.

  if ( inName != NULL ) {
    inFile = fopen ( inName, "r" );
  }

  if ( inFile == NULL ) {
    exit_msg ( "Missing cluster inventory description file.\n" );
    return 1;
  }

  lineNumber = 0;
  while ( NULL != fgets ( origline, STRING_LENGTH, inFile ) ) {
    lineNumber += 1;
    if ( origline[0] != NUL )
      origline[strlen ( origline ) - 1] = NUL;  /* remove \n */
    if ( isComment ( origline ) ) {
      continue;
    }
    if ( isBlank ( origline ) ) {
      continue;
    }
    //  if ( newSection ( origline ) == 1 ) {
    //  serviceNewSection (  );
    // }

    // Now to use the new parsing routines to break up the string:
    //

    pt_dbug(__LINE__, "parseClusterFile origline", origline, origline );

    line = whitespace_remove ( origline );      // chop off any spaces 
    // on the left.
    // cannot use case/switch because string equality required in each case.

    nextfield = string_separate ( line );       // this causes
    // line to be only the first entry, by way of the side effect.

    pt_dbug(__LINE__, "parseClusterFile line", line, line );
    pt_dbug(__LINE__, "parseClusterFile nextfield", nextfield, nextfield );

    // At this point, line contains the first string (identifier)
    // and nextfield contains all the arguments if any.
    // Although line should not be null, (we tested for
    // blankness and comments earlier) nextfield might be.

    // Correction:  empty strings are always pointed to the global 
    // constant string called my_null_string.  This prevents dereferencing
    // a real null pointer, and allows equality testing to my_null_string.


    if ( 0 == stricmp ( line, "$VERSION" ) ) {
      parseNumericLine ( nextfield, lineNumber, &versionNumber, 1, "Version",
                         inName );

      if ( versionNumber != 1 ) {

        sprintf ( tempstring,
                  "WARN: %s line %d: Unexpected file version number %d\n",
                  inName, lineNumber, versionNumber );
        error_msg ( tempstring );
      }
    }


    else if ( 0 == stricmp ( line, "$EMAIL_ADDRESS" ) ) {
      // copy entire space-separated email list into email_addresses

      if ( isNull ( nextfield ) ) {
        error_msg
          ( "Blank $EMAIL_ADDRESS listed in cluster description file\n" );
      } else if ( isNull ( email_addresses ) ) {
        email_addresses =
          malloc ( sizeof ( char ) * ( strlen ( nextfield ) + 1 ) );
        if ( email_addresses == NULL ) {
          sprintf ( tempstring,
                    "ERROR %s line %d: no memory to store email addresses!\n",
                    inName, lineNumber );
          exit_msg ( tempstring );
          break;
        } else {
          strcpy ( email_addresses, nextfield );        // insert the string.
        }
      } else {
        error_msg
          ( "Only one $EMAIL_ADDRESS email line entry allowed in clusterconf, with multiple space-separated addresses allowed.  Subsequent lines ignored.\n" );
        // necessary to avoid a memory leak!
      }
    }




    else if ( 0 == stricmp ( line, "$EMAIL_SERVER" ) ) {
      // copy entire space-separated email list into email_addresses

      if ( isNull ( nextfield ) ) {
        error_msg
          ( "Blank $EMAIL_SERVER entry in cluster description file\n" );
      } else if ( isNull ( email_server ) ) {
        line = nextfield;       // remember where we are.
        nextfield = string_separate ( line );
        // above causes line to point to a single non-spaced name.


        email_server = malloc ( sizeof ( char ) * ( strlen ( line ) + 1 ) );
        if ( email_server == NULL ) {
          sprintf ( tempstring,
                    "ERROR %s line %d: no memory to store email server!\n",
                    inName, lineNumber );
          exit_msg ( tempstring );
          break;
        } else {
          strcpy ( email_server, line );        // insert the string.
        }
      } else {
        exit_msg
          ( "Only one $EMAIL_SERVER entry allowed in clusterconf, with multiple space-separated addresses allowed.  Subsequent lines ignored.\n" );
        // necessary to avoid a memory leak!
      }
    }


    else if ( 0 == stricmp ( line, "$EMAIL_SENDER" ) ) {
      // copy entire space-separated email list into email_addresses

      if ( isNull ( nextfield ) ) {
        error_msg
          ( "Blank $EMAIL_SENDER entry in cluster description file\n" );
      } else if ( isNull ( email_sender ) ) {
        line = nextfield;       // remember where we are.
        nextfield = string_separate ( line );
        // above causes line to point to a single non-spaced name.


        email_sender = malloc ( sizeof ( char ) * ( strlen ( line ) + 1 ) );
        if ( email_sender == NULL ) {
          sprintf ( tempstring,
                    "ERROR %s line %d: no memory to store email sender!\n",
                    inName, lineNumber );
          exit_msg ( tempstring );
          break;
        } else {
          strcpy ( email_sender, line );        // insert the string.
        }
      } else {
        exit_msg
          ( "Only one $EMAIL_SENDER entry allowed in clusterconf, with multiple space-separated addresses allowed.  Subsequent lines ignored.\n" );
        // necessary to avoid a memory leak!
      }
    }
// handle the ability to turn off alerts/shutdowns altogether
    else if ( 0 == stricmp ( line, "$ALERT_IGNORE" ) ) {
      parseNumericLine ( nextfield, lineNumber, &alertignore, 1,
                         "Alert_ignore", inName );
      if ( ( alertignore != 1 ) && ( alertignore != 0 ) ) {
        sprintf ( tempstring,
                  "%s line %d: Alert ignore must be either 0 or 1\n", inName,
                  lineNumber );
        error_msg ( tempstring );
      }
    }
    // Now allow for non-default values for $EMAIL_REPEAT and $OLD_ALERT

    else if ( 0 == stricmp ( line, "$EMAIL_REPEAT" ) ) {
      parseNumericLine ( nextfield, lineNumber, &emailrepeat, 1,
                         "$EMAIL_REPEAT", inName );
    }

    else if ( 0 == stricmp ( line, "$OLD_ALERT" ) ) {
      parseNumericLine ( nextfield, lineNumber, &oldalert, 1,
                         "$OLD_ALERT", inName );
    }



    else if ( 0 == stricmp ( line, "$NODE" ) ) {

      // New node entry!  Get things set up...
      curnode = calloc ( 1, sizeof ( CLUSTERNODE ) );

      pt_dbug(__LINE__, "parseClusterFile new node! curnode", "", curnode );

      if ( curnode == NULL ) {
        exit_msg ( "Could not allocate memory for new node!\n" );
        break;
      }
      // remember that nextnodeptr is a pointer to a pointer to a chassis.
      else {
        *nextnodeptr = curnode; // First link the previous node to 
        // the current one.

        nextnodeptr = ( CLUSTERNODE ** ) & ( curnode->nextnode );
        // then change the pointer to allow the current node
        // to be linked to the next one.




//        pt_dbug(__LINE__,"parseClusterFile new node! *nextnodeptr","",(int)*nextnodeptr);
        pt_dbug(__LINE__, "parseClusterFile new node! nextnodeptr", "",
                  nextnodeptr );

        //?????????????????????/ should the & be there?
        // yes, we are making a pointer to the pointer to future nodes.

        //      First clear some things out:
        curnode->ipaddress = my_null_string;    // this was a bug before.
        curnode->chassis = NULL;
        curnode->autoalert = 1;
        curnode->autoshutdown = 1;
        curnode->nextnode = NULL;
        curnode->serialconsole = 0;
        curnode->ismaster = 0;

        // Now clear out the data for the node:

        // charcleanout = (char *) (&(curnode->nwdata) + sizeof(NODE));

        // start it out pointing one past the end of the data for the node.

        /*  The following didn't work so try a different approach below.

           while (charcleanout > (char *)(&(curnode->nwdata)))
           {
           charcleanout--; // subtract first because it starts out 

           // pointing one byte AFTER the end of the data!
           // it will automatically subtract based on sizeof(char).
           // because it was cast to a character pointer.

           *charcleanout = 0; // set data to zero.
           }

         */

        // Clean out the node the hard way:

        curnode->nwdata.alive = 0;
        curnode->nwdata.timestamp = 0;  //3;
        curnode->nwdata.powerstat = 0;
        curnode->nwdata.power_status = 0;
        curnode->nwdata.shutdowncount = 0;
        curnode->nwdata.alertcount = 0;
        curnode->nwdata.version.ucMajor = 10;
        curnode->nwdata.version.ucMinor = 11;
        curnode->power_cmd_state = POWER_CMD_IDLE_ON;

        // DATA INIT CHANGED
        for ( i = 0; i < MAX_TACK; i++ )
          curnode->nwdata.uaRpm[i] = i;

        for ( i = 0; i < MAX_CHAN; i++ ) {
          curnode->nwdata.uaA2D[i] = i;
          curnode->nwdata.bLower[i] = 0;
          curnode->nwdata.bUpper[i] = 0;
        }

        for ( i = 0; i < MAX_TACK; i++ )
          curnode->fan_limit_new_state[i] = NEW_LIM_IDLE;

        for ( i = 0; i < MAX_VOLT; i++ )
          curnode->volt_limit_new_state[i] = NEW_LIM_IDLE;

        for ( i = 0; i < MAX_TEMP; i++ )
          curnode->thermo_limit_new_state[i] = NEW_LIM_IDLE;


        // Above: First we record within the linked list the pointer to the 
        // new node allocated.
        // Next we update the pointer to the pointer, to point to the link
        // to the next node located within the current node.
        cluster_init_more ( curnode );
        //      Now initialize the allocated memory:
        if ( isNull ( nextfield ) ) {
          exit_msg ( "No node address information!\n" );

          //  break; // cannot break-- still have to clear things out.
        } else {
          unsigned int temprww;
          NodeCount++;
          line = nextfield;     // remember where we are.
          nextfield = string_separate ( line );
          sscanf ( line, " %u", &temprww );
          curnode->cardaddress = temprww;
          parseStringLine ( nextfield, lineNumber, &( curnode->ipaddress ),
                            1, "Node IP Address", inName );

          // Tricky -- it's not an array of strings, so it needs a pointer generated.
        }
      }
    } else if ( curnode != NULL ) {
      // Following sub-definitions require there to be a node:

      if ( 0 == stricmp ( line, "$$CHASSIS" ) ) {
        // Now we have to search the chassis database and 
        // add a pointer to the one whose name matches

        if ( curnode->chassis == NULL ) // not more than one chassis entry!
        {

          chassissearch = ChassisList;  // start the chassis list at the beginning.

          line = nextfield;     // remember where we are.
          nextfield = string_separate ( line );
          // above causes line to point to a single non-spaced name.

          while ( chassissearch != NULL ) {
            if ( 0 == stricmp ( chassissearch->name, line ) ) {
              curnode->chassis = chassissearch; // Found it!
              break;            // without this it searches all anyway.
            }
            chassissearch = chassissearch->nextchassis; // move on to the next.
          }
          if ( curnode->chassis == NULL ) {
            sprintf ( tempstring,
                      "ERROR %s line %d: %s unrecognized chassis name!\n",
                      inName, lineNumber, line );
            exit_msg ( tempstring );
            //    break;  // just keep going.  
            //   Watch out for null chassis pointer in result!
          } else if ( ( curnode->chassis )->tempaverage ) {
            TempavCount++;      // If the node has an inlet temp probe,
            // then increase the total count of such probes.
          }
        } else {
          sprintf ( tempstring,
                    "ERROR %s line %d: %s Extra chassis ID ignored.\n",
                    inName, lineNumber, line );
          error_msg ( tempstring );
        }
      }

      else if ( 0 == stricmp ( line, "$$AUTOALERT" ) ) {
        parseNumericLine ( nextfield, lineNumber, &( curnode->autoalert ), 1,
                           "$$AUTOALERT", inName );
      }


      else if ( 0 == stricmp ( line, "$$AUTOSHUTDOWN" ) ) {
        parseNumericLine ( nextfield, lineNumber, &( curnode->autoshutdown ),
                           1, "$$AUTOSHUTDOWN", inName );
      }

      else if ( 0 == stricmp ( line, "$$CARDVERSION" ) ) {
        // do nothing -- we don't need to know about the card version.
      }

      else if ( 0 == stricmp ( line, "$$SERIALCONSOLE" ) ) {
        curnode->serialconsole=1;
      }

      else if ( 0 == stricmp ( line, "$$MASTER" ) ) {
        // It's indicating that this node is the Nodewatch master -- IE 
        // has the converter box
        curnode->ismaster=1;
        if ( NWMaster == NULL ) {
          NWMaster = curnode;
        } else {
          sprintf ( tempstring,
                    "WARN: %s line %d ignored -- duplicate master declaration -- there can only be one Nodewatch master node.\n",
                    inName, lineNumber );
          error_msg ( tempstring );

          // Yeah, yeah, I know what you're thinking... the nwserver
          // could use the IP address to figure out which one it is,
          // and there could be duplicate redundant nodewatch masters,
          // which come on line when the line goes down, etc.etc. 

        }

      }

      else {
        sprintf ( tempstring,
                  "WARN: %s line %d ignored -- unrecognized heading\n",
                  inName, lineNumber );
        error_msg ( tempstring );
      }
    }

    else {
      sprintf ( tempstring,
                "WARN: %s line %d ignored -- no prior node declaration\n",
                inName, lineNumber );
      error_msg ( tempstring );
    }


  }                             // end of main while loop

  fclose ( inFile );

  // NOW CHECK IF ANY CRITICAL ENTRY WAS NOT CREATED:

  if ( email_addresses == NULL )
    exit_msg
      ( "No email address listed in clusterconf.dat.  Use $EMAIL_ADDRESS\n" );

  if ( email_server == NULL )
    exit_msg
      ( "No email server listed in clusterconf.dat.  Use $EMAIL_SERVER\n" );

  if ( email_sender == NULL )
    exit_msg
      ( "No email sender listed in clusterconf.dat.  Use $EMAIL_SENDER\n" );



  return 0;
}


//
// The following routine takes the cluster and chassis data in memory, and
// outputs a autonodeconf.dat file suitable for use with the php code.

int outNodeFile ( char *outName )
{

  FILE *outFile;                // file descriptor.
  CLUSTERNODE *clustersearch;   // pointer going through the list of nodes.
  CLUSTERNODE *nextsearch;      // pointer going through the next node.

  CHASSIS *chas;                // temporary pointer to the chassis.

  char tmpstring[400];          // for errors

  int i;

  outFile = fopen ( outName, "w" );
  if ( outFile == NULL ) {
    sprintf ( tmpstring, "ERROR: Unable to open output file %s\n", outName );
    error_msg ( tmpstring );
    return 1;                   //if file cannot be opened, exit
  }

  clustersearch = ClusterList;  // start the cluster search at the beginning.


  fprintf ( outFile, "\n! ***************\n! %s\n", outName );
  fprintf ( outFile, "! This file automatically generated by %s\n",
            program_name );
  fprintf ( outFile, "! DO NOT EDIT THIS FILE DIRECTLY!\n" );
  fprintf ( outFile, "! Edit the following files instead:\n" );
  fprintf ( outFile, "! %s\n", clusterfilename );
  fprintf ( outFile, "! %s\n", chassisfilename );
  fprintf ( outFile, "! (Restart %s to start using changes.)\n\n",
            program_name );
  fprintf ( outFile, "! THIS CONFIGURATION:\n" );
  fprintf ( outFile, "!  %u nodes in this cluster.\n", NodeCount );
  fprintf ( outFile,
            "!  %u temperature probes to be averaged in this cluster. (inlet)\n",
            TempavCount );
  fprintf ( outFile, "!  %u defined chassis types.\n", ChassisCount );
  fprintf ( outFile, "!  Alerts will be emailed to:  %s\n", email_addresses );

  fprintf ( outFile, "\n! ***************\n\n$NEWCHASSIS\n" );

  while ( clustersearch != NULL ) {
    chas = clustersearch->chassis;

    fprintf ( outFile, "\n$NODE %u %s\n", clustersearch->cardaddress,
              clustersearch->ipaddress );
    fprintf ( outFile, "$$AUTOALERT %u\n", clustersearch->autoalert );
    fprintf ( outFile, "$$AUTOSHUTDOWN %u\n", clustersearch->autoshutdown );
    if ( chas != NULL ) {
      fprintf ( outFile, "$$CHASSIS %s\n", chas->name );

      fprintf ( outFile, "$$FANLABELS " );
      for ( i = 0; i < MAX_TACK; i++ )
        fprintf ( outFile, " %s", chas->fanlabels[i] );

      fprintf ( outFile, "\n$$TEMPLABELS " );
      for ( i = 0; i < MAX_TEMP; i++ )
        fprintf ( outFile, " %s", chas->templabels[i] );

      fprintf ( outFile, "\n$$VOLTLABELS " );
      for ( i = 0; i < MAX_VOLT; i++ )
        fprintf ( outFile, " %s", chas->voltlabels[i] );


      fprintf ( outFile, "\n$$TACKARRAY " );
      for ( i = 0; i < MAX_TACK; i++ )
        fprintf ( outFile, " %u", chas->tackarray[i] );


      fprintf ( outFile, "\n$$VOLTALERTLO " );
      for ( i = 0; i < MAX_VOLT; i++ )
        fprintf ( outFile, " %d", chas->voltalertlo[i] );

      fprintf ( outFile, "\n$$VOLTALERTHI " );
      for ( i = 0; i < MAX_VOLT; i++ )
        fprintf ( outFile, " %d", chas->voltalerthi[i] );

      fprintf ( outFile, "\n$$VOLTSHUTDOWNLO " );
      for ( i = 0; i < MAX_VOLT; i++ )
        fprintf ( outFile, " %d", chas->voltshutdownlo[i] );

      fprintf ( outFile, "\n$$VOLTSHUTDOWNHI " );
      for ( i = 0; i < MAX_VOLT; i++ )
        fprintf ( outFile, " %d", chas->voltshutdownhi[i] );


      fprintf ( outFile, "\n$$TEMPALERTLO " );
      for ( i = 0; i < MAX_TEMP; i++ )
        fprintf ( outFile, " %d", chas->tempalertlo[i] );

      fprintf ( outFile, "\n$$TEMPALERTHI " );
      for ( i = 0; i < MAX_TEMP; i++ )
        fprintf ( outFile, " %d", chas->tempalerthi[i] );

      fprintf ( outFile, "\n$$TEMPSHUTDOWNLO " );
      for ( i = 0; i < MAX_TEMP; i++ )
        fprintf ( outFile, " %d", chas->tempshutdownlo[i] );

      fprintf ( outFile, "\n$$TEMPSHUTDOWNHI " );
      for ( i = 0; i < MAX_TEMP; i++ )
        fprintf ( outFile, " %d", chas->tempshutdownhi[i] );



      fprintf ( outFile, "\n$$FANALERTLO " );
      for ( i = 0; i < MAX_TACK; i++ )
        fprintf ( outFile, " %d", chas->fanalertlo[i] );

      fprintf ( outFile, "\n$$FANALERTHI " );
      for ( i = 0; i < MAX_TACK; i++ )
        fprintf ( outFile, " %d", chas->fanalerthi[i] );

      fprintf ( outFile, "\n$$FANSHUTDOWNLO " );
      for ( i = 0; i < MAX_TACK; i++ )
        fprintf ( outFile, " %d", chas->fanshutdownlo[i] );

      fprintf ( outFile, "\n$$FANSHUTDOWNHI " );
      for ( i = 0; i < MAX_TACK; i++ )
        fprintf ( outFile, " %d", chas->fanshutdownhi[i] );


      fprintf ( outFile, "\n$$FANGROUP " );
      for ( i = 0; i < MAX_TACK; i++ )
        fprintf ( outFile, " %d", chas->fangroup[i] );

      fprintf ( outFile, "\n$$FANGROUPFAIL " );
      for ( i = 0; i < MAX_TACK; i++ )
        fprintf ( outFile, " %d", chas->fangroupfail[i] );

      fprintf ( outFile, "\n" );

      if ( clustersearch->serialconsole == 1 ) {
        fprintf ( outFile, "$$SERIALCONSOLE\n");
      }

      if ( clustersearch->ismaster == 1 ) {
        fprintf ( outFile, "$$MASTER\n");
      } 
    }
    // Following is a convenience to the PHP script.
    // it indicates if a change in headings (change in chassis)
    // is required.

    nextsearch = clustersearch->nextnode;
    if ( nextsearch != NULL )   // if structured this way to not deref. null ptr.
    {
      if ( clustersearch->chassis != nextsearch->chassis )
        fprintf ( outFile, "\n$NEWCHASSIS\n" );
    }

    clustersearch = nextsearch; // update the search regardless if null or
    // not.  While loop needs to stop someday.

  }                             // end of main while loop
  fclose ( outFile );
  return 0;
}

//

void command_process (  )       // called by main loop to check file for command
{
  if ( isTimeout ( TIMER_CMDFILE ) ) {
    if ( ( pausecount-- ) < 1 ) // timeout on pause counts
    {
      pausecount = PAUSECOUNTS; // reset it
      sleep ( PAUSESLEEPS );    // wait for about 15 seconds.
      setTimeout ( TIMER_CMDFILE, TIMEOUT_PAUSE_CMDFILE );
    } else                      // don't run the command processing right after sleeping!
    {
      setTimeout ( TIMER_CMDFILE, TIMEOUT_CMDFILE );    // timeout done first
      unlink ( movedfilename );
      rename ( cmdfilename, movedfilename );
      parseFile ( NULL, movedfilename );        // try to parse the command file

      //  parseFile(cmdfile,NULL);  // try to parse the command file
      // was cmdfile,cmdfile
      // fifo is blocking...

      // which is supposed to be a named pipe.  If it cannot be opened,
      // it will try to open it as a regular file.
      if ( alertignore == 0 )   //allow limits to be skipped altogether
        genLimitsCheck (  );    // handle all the limits issues.

      //      setTimeout ( TIMER_CMDFILE, TIMEOUT_CMDFILE );
    }
  }
}

//
void command_open (  )
{

  /*  FOR LATER 
     int status;
     status =  unlink(cmdfilename);
     printf("Removed file %s with status %d \n",cmdfilename,status);

     status =  mkfifo(cmdfilename,(mode_t)(S_IRUSR|S_IWUSR));
     printf("Removed file %s with status %d \n",cmdfilename,status);


     status =  mkfifo(cmdfilename,(mode_t)(S_IRUSR|S_IWUSR));
     printf("Removed file %s with status %d \n",cmdfilename,status);
   */

  // Debug the globals:

  pt_dbug(__LINE__, "initial global my_null_string", my_null_string,
            my_null_string );
  pt_dbug(__LINE__, "initial global", " ClusterList", ClusterList );
//  pt_dbug(__LINE__,"initial global","&ClusterList",&ClusterList);
  pt_dbug(__LINE__, "initial global", " ChassisList", ChassisList );
//  pt_dbug(__LINE__,"initial global","&ChassisList",(int)&ChassisList);



  // Now read in the limits file...


  if ( nodaemon )
    printf ( "Reading limits file...\n" );

  // OBSOLETE:
  //  parseFile(NULL, limitsfilename);  // read in entire limits file


  parseChassisFile ( chassisfilename );
  parseClusterFile ( clusterfilename );
  outNodeFile ( autonodefilename );

  email_init (  );              // set up variables associated with emailing.
  //  shutdown_init();  // set up variables associated with shutdown
  //??? does this have to be done per node... if so do it in parseClusterFile

  //  OK, limits file read in, now start the command processing loop:

  pausecount = INITIALPAUSECOUNTS;      // reset the pause counter.

  unlink ( movedfilename );
  unlink ( cmdfilename );

  cmdfile = NULL;               // let the backup system do the work.


  setTimeout ( TIMER_CMDFILE, TIMEOUT_CMDFILE );

  //  while(1); /// get stuck here after initializing things.

}

//
void timecountcheck ( CLUSTERNODE * thechecknode, int *thecheckcount,
                      long *thechecktime, char *actual_limit_state,
                      char proposed_limit_state )
     // This function is called for every over measurement.
     // accepts time and count parameters.
     // if these are exceeded, the appropriate node and measurement is marked 
     // for evaluation of the power situation.
{
  long thetime;
  CHASSIS *thenodechassis;
  thetime = mygetcurtime (  );
  pt_dbug2(2,__LINE__, "timecountcheck checkcount", *thecheckcount, NULL );
  pt_dbug2(2,__LINE__, "timecountcheck limitstates", proposed_limit_state,
             actual_limit_state );
  thenodechassis = thechecknode->chassis;

  if ( ( *thecheckcount >
         thenodechassis->measure_confirm_count ) &&
       ( ( thetime - ( *thechecktime ) ) >
         thenodechassis->measure_confirm_time ) )
    // OK we've got an exceeded limit!
  {
    if ( *actual_limit_state == NEW_LIM_IDLE )
      // Do not proceed if already in progress from before.
    {
      start_power_measure ( thechecknode );
      *actual_limit_state = proposed_limit_state;       // marks the kind of error waiting for.
      // NOW reset the counts and times of the limit:

      *thecheckcount = 0;
      *thechecktime = 0;

    }
    // If there is an alert condition, and it escalates
    // to a shutdown condition, the shutdown condition 
    // will have to wait briefly for the alert condition to finish
    // being handled.

    // If the new_limit_state is NOT idle, we can and will
    // safely ignore the entire situation.  This is because
    // 1.  the limit_state is already handling the situation
    // and 2.  the limit_state will shortly become idle again,
    // at which point it will be handled.
  }
}

//
void start_power_measure ( CLUSTERNODE * thenodechassis )
{
  long thetime;

  // If the power measurement is not already in progress,
  // reset the timers and start it for this node.

  //??? WRONG DATA TYPE FOR TIME.  WILL NEED TO BE FIXED
  // TO INCLUDE ENTIRE TIME STRUCTURE, SECONDS AND MILLISECONDS.
  pt_dbug2(2,__LINE__, "start_power_measure",
             thenodechassis->power_supply_state_timeout, thenodechassis );
  if ( thenodechassis->power_supply_state_timeout == 0 )
    // there is no current active power status search
  {
    thetime = mygetcurtime (  );
    thenodechassis->power_supply_state_timeout =
      thetime + ( thenodechassis->chassis )->power_valid_timeout;
    thenodechassis->power_supply_state_count = 0;
    thenodechassis->power_supply_state_time = 0;
  }
}

//
void checkThermoLimits ( CLUSTERNODE * thechecknode )
     // Called in execnode.c when new analog info received.
     // How does it remember the breakdown of 
     // analog information into temps and volts?
     // ANSWER:  it refers to:
     //  - voltarray
     //  - thermoarray ... both in the chassis data structure
     // and it refers to MAX_VOLT and MAX_TACK to 
     // determine how far to count.
     // For now, the voltarray and thermoarray are initialized
     // to hardcoded values:  0,1,2 and 3,4,5,6,7... 
     // Note -- it does not count thru all nodes.  It is called from 
     // execnode with a pointer to a particular node to check.
     // In other words, there is no point in checking a node which
     // has not reported data, because the data would be the same as 
     // last reported, and because if not reported the default data
     // of all 0 does not make sense for limits anyway.
{

  int i;
  CHASSIS *thenodechassis = thechecknode->chassis;
  pt_dbug2(2,__LINE__, "checkThermoLimits", thechecknode->cardaddress,
             thenodechassis );
  for ( i = 0; i < MAX_TEMP; i++ ) {
    //??? watch out to average the averaged temperature.
    //??? watch out for disconnected probes... the way datafiles does

    //??? minor bug:  
    // check each against alert lo& hi, and shutdown lo&hi

    // check alert lo
    if ( thenodechassis->tempalertlo[i] >
         thechecknode->nwdata.uaA2D[thenodechassis->thermoarray[i]] ) {
      // measured temperature is below low alert level.

      // First check to see if this is the first continuous time
      // this occured:
      if ( thechecknode->thermo_alert_lo_count[i] == 0 ) {      // first time overrun; mark the time received.
        thechecknode->thermo_alert_lo_time[i] =
          thechecknode->nwdata.timestamp;
      }
      thechecknode->thermo_alert_lo_count[i]++; // now increase the count.
      global_alert_flag = 1;

      timecountcheck ( thechecknode,
                       &( thechecknode->thermo_alert_lo_count[i] ),
                       &( thechecknode->thermo_alert_lo_time[i] ),
                       &( thechecknode->thermo_limit_new_state[i] ),
                       NEW_LIM_ALERT_LO_WAITING_PWR );
    } else
      thechecknode->thermo_alert_lo_count[i] = 0;       // nothing wrong, so clear the count.


    // check alert HI
    if ( thenodechassis->tempalerthi[i] <
         thechecknode->nwdata.uaA2D[thenodechassis->thermoarray[i]] ) {
      // measured temperature is above high alert level!

      // First check to see if this is the first continuous time
      // this occured:
      if ( thechecknode->thermo_alert_hi_count[i] == 0 ) {      // first time overrun; mark the time received.
        thechecknode->thermo_alert_hi_time[i] =
          thechecknode->nwdata.timestamp;
      }
      thechecknode->thermo_alert_hi_count[i]++; // now increase the count.
      global_alert_flag = 1;
      timecountcheck ( thechecknode,
                       &( thechecknode->thermo_alert_hi_count[i] ),
                       &( thechecknode->thermo_alert_hi_time[i] ),
                       &( thechecknode->thermo_limit_new_state[i] ),
                       NEW_LIM_ALERT_HI_WAITING_PWR );


    } else
      thechecknode->thermo_alert_hi_count[i] = 0;       // nothing wrong, so clear the count.


    // check shutdown  lo
    if ( thenodechassis->tempshutdownlo[i] >
         thechecknode->nwdata.uaA2D[thenodechassis->thermoarray[i]] ) {
      // measured temperature is below low shutdown level!

      // First check to see if this is the first continuous time
      // this occured:
      if ( thechecknode->thermo_shutdown_lo_count[i] == 0 ) {   // first time overrun; mark the time received.
        thechecknode->thermo_shutdown_lo_time[i] =
          thechecknode->nwdata.timestamp;
      }
      thechecknode->thermo_shutdown_lo_count[i]++;      // now increase the count.
      global_shutdown_flag = 1;
      timecountcheck ( thechecknode,
                       &( thechecknode->thermo_shutdown_lo_count[i] ),
                       &( thechecknode->thermo_shutdown_lo_time[i] ),
                       &( thechecknode->thermo_limit_new_state[i] ),
                       NEW_LIM_SHUTDOWN_LO_WAITING_PWR );


    } else
      thechecknode->thermo_shutdown_lo_count[i] = 0;    // nothing wrong, so clear the count.




    // check shutdown  hi
    if ( thenodechassis->tempshutdownhi[i] <
         thechecknode->nwdata.uaA2D[thenodechassis->thermoarray[i]] ) {
      // measured temperature is above high shutdown level!

      // First check to see if this is the first continuous time
      // this occured:
      if ( thechecknode->thermo_shutdown_hi_count[i] == 0 ) {   // first time overrun; mark the time received.
        thechecknode->thermo_shutdown_hi_time[i] =
          thechecknode->nwdata.timestamp;
      }
      thechecknode->thermo_shutdown_hi_count[i]++;      // now increase the count.
      global_shutdown_flag = 1;
      timecountcheck ( thechecknode,
                       &( thechecknode->thermo_shutdown_hi_count[i] ),
                       &( thechecknode->thermo_shutdown_hi_time[i] ),
                       &( thechecknode->thermo_limit_new_state[i] ),
                       NEW_LIM_SHUTDOWN_HI_WAITING_PWR );

    } else
      thechecknode->thermo_shutdown_hi_count[i] = 0;    // nothing wrong, so clear the count.

  }
}

//
void checkVoltLimits ( CLUSTERNODE * thechecknode )
     // Called in execnode.c when new analog info received.
     // This routine will also process the power on-off analysis.  DONE NOW
{

  int newpowerstate = 0;        // used to accumulate calculation of power 
  // switch state.  wired-or:  starts as 0, 
  // brought up for any voltage indicating power attempt. 

  int i;
  CHASSIS *thenodechassis = thechecknode->chassis;

  long thetime;
  thetime = mygetcurtime (  );
  pt_dbug2(2,__LINE__, "checkVoltLimits", thechecknode->cardaddress,
             thenodechassis );
  for ( i = 0; i < MAX_VOLT; i++ ) {

    //??? watch out for power supplies that are truly _off_

    // check each against alert lo& hi, and shutdown lo&hi

    // check alert lo
    if ( thenodechassis->voltalertlo[i] >
         thechecknode->nwdata.uaA2D[thenodechassis->voltarray[i]] ) {
      // measured volt is below low alert level.

      // First check to see if this is the first continuous time
      // this occured:
      if ( thechecknode->volt_alert_lo_count[i] == 0 ) {        // first time overrun; mark the time received.
        thechecknode->volt_alert_lo_time[i] = thechecknode->nwdata.timestamp;
      }
      thechecknode->volt_alert_lo_count[i]++;   // now increase the count.
      global_alert_flag = 1;
      timecountcheck ( thechecknode,
                       &( thechecknode->volt_alert_lo_count[i] ),
                       &( thechecknode->volt_alert_lo_time[i] ),
                       &( thechecknode->volt_limit_new_state[i] ),
                       NEW_LIM_ALERT_LO_WAITING_PWR );



    } else
      thechecknode->volt_alert_lo_count[i] = 0; // nothing wrong, so clear the count.


    // check alert HI
    if ( thenodechassis->voltalerthi[i] <
         thechecknode->nwdata.uaA2D[thenodechassis->voltarray[i]] ) {
      // measured volt is above high alert level!

      // First check to see if this is the first continuous time
      // this occured:
      if ( thechecknode->volt_alert_hi_count[i] == 0 ) {        // first time overrun; mark the time received.
        thechecknode->volt_alert_hi_time[i] = thechecknode->nwdata.timestamp;
      }
      thechecknode->volt_alert_hi_count[i]++;   // now increase the count.
      global_alert_flag = 1;
      timecountcheck ( thechecknode,
                       &( thechecknode->volt_alert_hi_count[i] ),
                       &( thechecknode->volt_alert_hi_time[i] ),
                       &( thechecknode->volt_limit_new_state[i] ),
                       NEW_LIM_ALERT_HI_WAITING_PWR );



    } else
      thechecknode->volt_alert_hi_count[i] = 0; // nothing wrong, so clear the count.


    // check shutdown  lo
    if ( thenodechassis->voltshutdownlo[i] >
         thechecknode->nwdata.uaA2D[thenodechassis->voltarray[i]] ) {
      // measured volt is below low shutdown level!

      // First check to see if this is the first continuous time
      // this occured:
      if ( thechecknode->volt_shutdown_lo_count[i] == 0 ) {     // first time overrun; mark the time received.
        thechecknode->volt_shutdown_lo_time[i] =
          thechecknode->nwdata.timestamp;
      }
      thechecknode->volt_shutdown_lo_count[i]++;        // now increase the count.
      global_shutdown_flag = 1;

      timecountcheck ( thechecknode,
                       &( thechecknode->volt_shutdown_lo_count[i] ),
                       &( thechecknode->volt_shutdown_lo_time[i] ),
                       &( thechecknode->volt_limit_new_state[i] ),
                       NEW_LIM_SHUTDOWN_LO_WAITING_PWR );


    } else
      thechecknode->volt_shutdown_lo_count[i] = 0;      // nothing wrong, so clear the count.




    // check shutdown  hi
    if ( thenodechassis->voltshutdownhi[i] <
         thechecknode->nwdata.uaA2D[thenodechassis->voltarray[i]] ) {
      // measured volt is above high shutdown level!

      // First check to see if this is the first continuous time
      // this occured:
      if ( thechecknode->volt_shutdown_hi_count[i] == 0 ) {     // first time overrun; mark the time received.
        thechecknode->volt_shutdown_hi_time[i] =
          thechecknode->nwdata.timestamp;
      }
      thechecknode->volt_shutdown_hi_count[i]++;        // now increase the count.
      global_shutdown_flag = 1;

      timecountcheck ( thechecknode,
                       &( thechecknode->volt_shutdown_hi_count[i] ),
                       &( thechecknode->volt_shutdown_hi_time[i] ),
                       &( thechecknode->volt_limit_new_state[i] ),
                       NEW_LIM_SHUTDOWN_HI_WAITING_PWR );


    } else
      thechecknode->volt_shutdown_hi_count[i] = 0;      // nothing wrong, so clear the count.


    // check power on/off situation.

    if ( thenodechassis->volt_powerstate_level[i] <
         thechecknode->nwdata.uaA2D[thenodechassis->voltarray[i]] ) {
      newpowerstate = 1;        // at least one of the voltages is minimally 
      // alive -- we conclude that the power supply is trying to be on        
    }
  }
  // Now figure if the power state is the same as before.
  // Note that the initial value of power_supply_state will need
  // to be -1 or something like that so this will always trigger a new
  // event and a new timestamp.

  // Knowing how long and for how many measurements, 
  // the power has been on or off, is the first step
  // in identifying poweroff-related secondary failures.

  if ( newpowerstate == thechecknode->power_supply_state ) {
    thechecknode->power_supply_state_count++;
  } else {
    thechecknode->power_supply_state_count = 0;
    thechecknode->power_supply_state_time = thechecknode->nwdata.timestamp;
    thechecknode->power_supply_state = newpowerstate;
  }
}

//
// this function checks a single node
void checkTackLimits ( CLUSTERNODE * thechecknode )
     // Called in execnode.c when new tack info received.
{

  int i;
  CHASSIS *thenodechassis = thechecknode->chassis;

  long thetime;
  thetime = mygetcurtime (  );
  pt_dbug2(2,__LINE__, "checkTackLimits", thechecknode->cardaddress,
             thenodechassis );
  for ( i = 0; i < thenodechassis->numfans; i++ ) {

//??? watch out for power supplies that are truly _off_

    // check each against alert lo& hi, and shutdown lo&hi

    // check alert lo
    if ( thenodechassis->fanalertlo[i] >
         thechecknode->nwdata.uaRpm[thenodechassis->tackarray[i]] ) {
      // measured fan is below low alert level.
      pt_dbug(__LINE__, "rww checkTackLimits checking thenodechassis->name",
                thenodechassis->name, NULL );
      pt_dbug2(2,__LINE__, "rww checkTackLimits checking on i=", i, NULL );
      pt_dbug2(2,__LINE__,
                 "rww checkTackLimits alertlo thechecknode->nwdata.uaRpm[thenodechassis->tackarray[i]]",
                 thechecknode->nwdata.uaRpm[thenodechassis->tackarray[i]],
                 NULL );
      pt_dbug2(2,__LINE__,
                 "rww checkTackLimits alertlo thenodechassis->fanalertlo[i]",
                 thenodechassis->fanalertlo[i], NULL );
      // First check to see if this is the first continuous time
      // this occured:
      if ( thechecknode->fan_alert_lo_count[i] == 0 ) { // first time overrun; mark the time received.
        thechecknode->fan_alert_lo_time[i] = thechecknode->nwdata.timestamp;
      }
      thechecknode->fan_alert_lo_count[i]++;    // now increase the count.
      global_alert_flag = 1;

      timecountcheck ( thechecknode, &( thechecknode->fan_alert_lo_count[i] ),
                       &( thechecknode->fan_alert_lo_time[i] ),
                       &( thechecknode->fan_limit_new_state[i] ),
                       NEW_LIM_ALERT_LO_WAITING_PWR );

    } else
      thechecknode->fan_alert_lo_count[i] = 0;  // nothing wrong, so clear the count.


    // check alert HI
    if ( thenodechassis->fanalerthi[i] <
         thechecknode->nwdata.uaRpm[thenodechassis->tackarray[i]] ) {
      // measured fan is above high alert level!
      pt_dbug(__LINE__, "rww checkTackLimits checking thenodechassis->name",
                thenodechassis->name, NULL );
      pt_dbug2(2,__LINE__, "rww checkTackLimits checking on i=", i, NULL );
      pt_dbug2(2,__LINE__,
                 "rww checkTackLimits alerthi thechecknode->nwdata.uaRpm[thenodechassis->tackarray[i]]",
                 thechecknode->nwdata.uaRpm[thenodechassis->tackarray[i]],
                 NULL );
      pt_dbug2(2,__LINE__,
                 "rww checkTackLimits alerthi thenodechassis->fanalerthi[i]",
                 thenodechassis->fanalerthi[i], NULL );

      // First check to see if this is the first continuous time
      // this occured:
      if ( thechecknode->fan_alert_hi_count[i] == 0 ) { // first time overrun; mark the time received.
        thechecknode->fan_alert_hi_time[i] = thechecknode->nwdata.timestamp;
      }
      thechecknode->fan_alert_hi_count[i]++;    // now increase the count.
      global_alert_flag = 1;

      timecountcheck ( thechecknode, &( thechecknode->fan_alert_hi_count[i] ),
                       &( thechecknode->fan_alert_hi_time[i] ),
                       &( thechecknode->fan_limit_new_state[i] ),
                       NEW_LIM_ALERT_HI_WAITING_PWR );


    } else
      thechecknode->fan_alert_hi_count[i] = 0;  // nothing wrong, so clear the count.


    // check shutdown  lo
    if ( thenodechassis->fanshutdownlo[i] >
         thechecknode->nwdata.uaRpm[thenodechassis->tackarray[i]] ) {
      // measured fan is below low shutdown level!
      pt_dbug(__LINE__, "rww checkTackLimits checking thenodechassis->name",
                thenodechassis->name, NULL );
      pt_dbug2(2,__LINE__, "rww checkTackLimits checking on i=", i, NULL );
      pt_dbug2(2,__LINE__,
                 "rww checkTackLimits shutdownlo thechecknode->nwdata.uaRpm[thenodechassis->tackarray[i]]",
                 thechecknode->nwdata.uaRpm[thenodechassis->tackarray[i]],
                 NULL );
      pt_dbug2(2,__LINE__,
                 "rww checkTackLimits alertlo thenodechassis->fanshutdownlo[i]",
                 thenodechassis->fanshutdownlo[i], NULL );

      // First check to see if this is the first continuous time
      // this occured:
      if ( thechecknode->fan_shutdown_lo_count[i] == 0 ) {      // first time overrun; mark the time received.
        thechecknode->fan_shutdown_lo_time[i] =
          thechecknode->nwdata.timestamp;
      }
      thechecknode->fan_shutdown_lo_count[i]++; // now increase the count.
      global_shutdown_flag = 1;

      timecountcheck ( thechecknode,
                       &( thechecknode->fan_shutdown_lo_count[i] ),
                       &( thechecknode->fan_shutdown_lo_time[i] ),
                       &( thechecknode->fan_limit_new_state[i] ),
                       NEW_LIM_SHUTDOWN_LO_WAITING_PWR );


    } else
      thechecknode->fan_shutdown_lo_count[i] = 0;       // nothing wrong, so clear the count.


    // check shutdown  hi
    if ( thenodechassis->fanshutdownhi[i] <
         thechecknode->nwdata.uaRpm[thenodechassis->tackarray[i]] ) {
      // measured fan is above high shutdown level!
      pt_dbug(__LINE__, "rww checkTackLimits checking thenodechassis->name",
                thenodechassis->name, NULL );
      pt_dbug2(2,__LINE__, "rww checkTackLimits checking on i=", i, NULL );
      pt_dbug2(2,__LINE__,
                 "rww checkTackLimits shutdownhi thechecknode->nwdata.uaRpm[thenodechassis->tackarray[i]]",
                 thechecknode->nwdata.uaRpm[thenodechassis->tackarray[i]],
                 NULL );
      pt_dbug2(2,__LINE__,
                 "rww checkTackLimits shutdownhi thenodechassis->fanshutdownhi[i]",
                 thenodechassis->fanshutdownhi[i], NULL );

      // First check to see if this is the first continuous time
      // this occured:
      if ( thechecknode->fan_shutdown_hi_count[i] == 0 ) {      // first time overrun; mark the time received.
        thechecknode->fan_shutdown_hi_time[i] =
          thechecknode->nwdata.timestamp;
      }
      thechecknode->fan_shutdown_hi_count[i]++; // now increase the count.
      global_shutdown_flag = 1;

      timecountcheck ( thechecknode,
                       &( thechecknode->fan_shutdown_hi_count[i] ),
                       &( thechecknode->fan_shutdown_hi_time[i] ),
                       &( thechecknode->fan_limit_new_state[i] ),
                       NEW_LIM_SHUTDOWN_HI_WAITING_PWR );


    } else
      thechecknode->fan_shutdown_hi_count[i] = 0;       // nothing wrong, so clear the count.
  }

}

//
// The following function is used to confirm, after seeing a fan 
// or voltage failure, that the power supply continues to be on.
// It is possible for a power supply failure or a manual turn-off
// to cause apparent fan failures visible before the power
// supply is seen to be off.
//=================
// Algorithm:  

// A power level determination is accomplished by means of a minimum number
// of measurements and a minimum time in a particular power state, AFTER
// the request for a determination has been started.
// Once a power level determination has been started for a particular node,
// it cannot be stopped, for instance by other measurements subsequently
// going out-of-bounds.  However, the power level determination will
// need to maintain two sets of counts and times:  One to find the minimum
// continuous measurement within a count and time interval, and the other
// (larger) to timeout if too many inconsistent measurements are received.

// The power level determination is used within the main algorithm
// for verifying a failure.
// Unfortunately, since we are not using true OOP, the state of
// progress through the main algorithm will be maintained for
// each measurement within an array in the CLUSTERNODE structure.


// why do we have a minimum time requirement on 
// all these measurements?  multiple measurements are only needed
// because of possible comm or physical measurement errors.
// minimum time only makes it more complicated.
//  Two measurements in a row should be more than enough confirmation...

// Reason:  because if the measurements come very quickly,
// want to make sure they were MEASURED stable.

// So in summary, state is stored in four ways:  There are configurable 
// parameters associated with much of the state.
//  - the counts and times used to initially double-check level violations
//  - the counts and times used to determine power level (min time and timeout)
//  - the state maintained for each MEASUREMENT, showing its progress
//    to being marked overlimit
//  - the state maintained for each NODE, showing its progress and countdown
//    towards a shutdown 

// what happens if an alert was just triggered before
// a shutdown?  it will email both...

// Here are all the states:
// Counts and times:  one measurement count and one time-duration 
// for alert hi, alert lo, shutdown hi, and shutdown lo.
// It's done this way because shutdown is a superset of alert.

// counts and times for power level:  count, time and state for minimum 
//  time/count in one state, and count, time for timeout trying to find
//  duration in one state.

// state for each measurement:  
//    idle - nothing is happening.
//    An overlimit condition found, waiting for power level determination.
//     (during this time the power level determination algorithm
//     starts counting states and time)
//    Power level determination complete:
//      either the measurement is overlevel, in which case 
//      the node is marked for email and/or shutdown action,
//      and the measurement is marked for notification.  
//     or
//      the power is off, in which case the node is marked
//      for email action, and the measurement returns to the idle state.

// State maintained for each node:  related to power and shutdown.
//  - idle, power on.
//  - lost contact with nodewatch, power unknown
//  - power off (spontaneously turned off)
//  - emergency shutdown in progress.  OS notified. waiting to power off. 
//  - emergency shutdown in progress.  shutdown command given.  waiting to confirm.
//      ABOVE STATE CAUSES POWER DETERMINATION ALGORITHM TO ADD EXTRA DELAY
//      BEFORE TRYING TO RESPOND.
// - Emergency shutdown completed.  poweroff was confirmed.


//  separate time maintained to countdown to poweroff.



//  Time maintained for each node:  time of last known transition.
// For a more detailed chronology of the activity, please check your emails
//  or the system logger.



// state maintained for emailing:  simple timeout.
//  It works like this:  anytime a new situation is discovered,
//   an email is immediately started that first describes the 
//   new situation.  Then a general email handler is called that
//   writes up all currently listed problems (based on 
//   the states of the measurements)

// Every time the email handler is called, it sets a timeout of a day,
// after which it is called again to send out a reminder email.

// The result is that the user receives a fresh email every time there
// is a new valid situation; each email lists all situations, and
// the user also gets a reminder emails every day after the last email,
// reminder or not, was sent.

//

/*void send_test_email() // for benefit of -m option, sends email and quits.
{
  //  email_init();
  
  while
    (email_build
     (snprintf
      (email_buf,email_mem,	      
       "Test email from Nodewatch monitoring system.\n\n")));

  sleep(10);
  exit(0);

}*/


void genLimitsCheck (  )
     // checks for pending power determinations, upgrades the
     // state for the appropriate measurement, and takes appropriate
     // alert or shutdown action:
     // - checks power determinations and 
     //   double checks that a power determination has not timed out.
     // - checks for old data.
     // - sweeps the nodes looking for pending email notifications, etc.,
     //   and sends them.
     // - initiates, monitors and concludes shutdowns if need be.
     // - sends daily (by default) reminder emails for ongoing situations
     //   if appropriate.
     // - monitors already sent emails for success status 
     // - logs appropriate information to syslog.
     // should be called periodically, for instance when checking for commands
     // or when checking to poll more data.
{
  long thetime;

  long timedifference;

  CHASSIS *chas;                // temporary pointer to the chassis.
  CLUSTERNODE *clustersearch;   // pointer going through the list of nodes.
  CLUSTERNODE *nextsearch;      // pointer going through the next node.

  clustersearch = ClusterList;  // start the cluster search at the beginning.
  thetime = mygetcurtime (  );

  while ( clustersearch != NULL ) {
    if ( clustersearch->autoalert == 0 ) {
      clustersearch = clustersearch->nextnode;
      continue;
    }
    chas = clustersearch->chassis;
    // Loop thru all clusters, searching for situations listed above

    // Start by checking for old data.

    timedifference = thetime - clustersearch->nwdata.timestamp;

    if ( ( timedifference > oldalert )
         && ( ( thetime - mytimestart ) > oldalert ) )  //waiting initial period
      // mytimestart refers to when the daemon program first started.
    {
      // THIS NODE HAS OLD DATA.  DO SOMETHING.
      // We will need a fcn that sends email alerts 
      // and executes shutdowns.

      //////////////////////////////////// cannot run out of memory!
      //// we will be using asprintf to allocate and print to string
      // in one operation, without worrying about how big the string
      // is.


      /* CONVENTION FOR PRINTING

         if (email_restrict (clustersearch, EHASH_OLD))
         while
         (email_build
         (snprintf
         (email_buf,email_mem,
         "STRING", ARGUMENTS))));

       */

      // Now tell whether it's "old data" or "no data at all":

      if ( clustersearch->nwdata.timestamp > 0 )        // it's old data.
      {
        pt_dbug2(2,__LINE__,
                   "genlimitscheck considering sending old data email",
                   clustersearch->email_hash, clustersearch );

        ctime_r ( (const time_t *)&( clustersearch->nwdata.timestamp ), &mytimestring[0] );
        if ( email_restrict ( clustersearch, 1 ) )      //  ehash_old
          while
            ( email_build
              ( snprintf
                ( email_buf, email_mem,
                  "Node %s has not been providing Nodewatch data\n"
                  " for %d minutes.  Last contact at %s\n\n",
                  clustersearch->ipaddress, ( int ) ( timedifference / 60 ),
                  // ctime(&(clustersearch->nwdata.timestamp)))));
                  mytimestring ) ) );
        else {
          pt_dbug2(2,__LINE__, "genlimitscheck RESTRICTED THE EMAIL",
                     clustersearch->email_hash, clustersearch );

        }

      } else {
        pt_dbug2(2,__LINE__,
                   "genlimitscheck considering sending no data email",
                   clustersearch->cardaddress, clustersearch );

        ctime_r ( &mytimestart, &mytimestring[0] );
        if ( email_restrict ( clustersearch, 2 ) )      // ehash_none
          while
            ( email_build
              ( snprintf
                ( email_buf, email_mem,
                  "Node %s has never provided Nodewatch data\n"
                  " since the last restart of the monitoring service at %s\n\n",
                  clustersearch->ipaddress,
                  //  ctime(&mytimestart))));
                  &mytimestring[0] ) ) );
      }
      cluster_reset_limits ( clustersearch );
      // prevents the system from acting on old data.

    }
    //???  TYPE MISMATCH WITH TIME.
    //??? NEED TO FIX.

    // Now check for pending power detects,
    // which, if present, means there was an out-of-range
    // situation.

    // Do temperature failures really require 
    // power measurement?

    if ( clustersearch->power_supply_state_timeout != 0 )
      // There is a power measurement happening now.

    {
      // First check to see if the power measurement is finished:
      // It can be finished either because it timed out (ie unsuccessful
      // measurement) or because it had a successful run of consistent
      // measurements (successful)
      // In any event, when a power measurement is finished for 
      // either of the above reasons, it's time to 
      // take action:  emails or shutdowns.

      if ( ( clustersearch->power_supply_state_count
             > chas->power_valid_count )
           && ( ( thetime - clustersearch->power_supply_state_time )
                > chas->power_valid_time ) )
        // We have a valid measurement:

      {
        // Take action:
        action_limits ( clustersearch, 1 );     // take action now, valid value.
      } else if ( thetime > clustersearch->power_supply_state_timeout ) {
        action_limits ( clustersearch, 0 );     // take action now, invalid value.
      }
    }
    // Last step:  check on any in-progress auto shutdowns:

    shutdown_check ( clustersearch );

    nextsearch = clustersearch->nextnode;
    clustersearch = nextsearch; // update the search regardless if null or
    // not.  While loop needs to stop someday.
  }

// if (not in progress sending AND has msg to be sent)
  email_send (  );              // final command to actually push emails out.
//else continue to build up message for next chance to send
  email_mon (  );               // check on any in-progress emails.

}

//
int email_alloc ( int asksize )
{
  // accepts a size in bytes as a minimum for the remaining available space.
  // makes sure there is enough room at the end of email_buffer_start (after email_buf) for that amount.
  // if not, allocates more memory.
  // assumes the initial allocation has already taken place.
  void *temp_ptr;
  int new_mem_total;

  pt_dbug2(2,__LINE__, "email_alloc asksize,email_buf_start: ", asksize,
             email_buf_start );
  if ( email_mem < asksize ) {
    // some judgement/prediction is needed here:
    new_mem_total = 2 * email_mem_total + asksize;      // This is a bit more than was asked for.
    // we technically only need email_mem_total + asksize - email_mem.

    temp_ptr = realloc ( email_buf_start, new_mem_total * sizeof ( char ) );

    if ( temp_ptr == NULL ) {
      exit_msg ( "Realloc is out of memory!\n" );
      //??? SHOULD THIS EXIT?    - RW - yes
      exit ( -1 );
    } else {
      // everything OK, new memory added; update values:

      pt_dbug2(2,__LINE__, "email_alloc did it  new_mem_total,email_buf:",
                 new_mem_total, email_buf );

      email_buf_start = temp_ptr;       // in case realloc moved everything.
      email_buf = email_buf_start + email_mem_total - email_mem;        // in case realloc moved everything!!!

      email_mem += ( new_mem_total - email_mem_total );
      email_mem_total = new_mem_total;

      pt_dbug2(2,__LINE__, "email_alloc did it  email_mem, email_buf_total:",
                 email_mem, email_buf_start );

      return ( 1 );
    }
  } else {
    return ( 1 );
  }
}

//
int email_build ( int printedsize )
{
  // this function is designed to be called with the output of snprintf,
  // and to be surrounded by a while loop.
  // it checks that the string was successfully printed to email_build
  // if so, it then updates the email_alloc_remain and the email_build_end pointer
  // if snprintf failed, it allocates more memory, updates the appropriate variables,
  //  and returns 1 so that the calling while loop will cause the snprintf to try again. 

  pt_dbug2(2,__LINE__, "email_build printedsize", printedsize, 0 );
  if ( ( printedsize > -1 ) && ( printedsize <= email_mem ) )
    // print was successful.
  {
    email_mem -= printedsize;
    email_buf += printedsize;   // points to the terminating null.
    return ( 0 );               // cause surrounding while loop to exit.
  } else {
    if ( printedsize > -1 ) {
      email_alloc ( printedsize );
      return ( 1 );
    } else {
      email_alloc ( email_mem_total );
      return ( 1 );
    }
  }
}

//
int email_restrict ( CLUSTERNODE * thenode, unsigned long long newhashvalue )
{
  // Helps decide whether to actually send a message.
  // returns 1 if it is a good idea, and 0 if it recommends skipping the message.

  // First, if this is an urgent message (shutdown pending), it gets sent regardless
  // of the following.  Urgent messages have zero as a newhashvalue.

  // Otherwise, decide whether the segment should be sent in the first place:
  // bitwise-compare the newhashvalue with that already stored
  // in the node. If they compare, then that particular warning
  // has already been sent, within the past 24 [specified] hours. 

  // Then check the last email time; if the time is up, set it to 0 
  // and erase the hash in the node.  If there is no time, 
  // then this is a new message; mark the time in last email time.

  // the resulting behavior is that a new message is sent for
  // each new problem or set of new problems encountered;
  // however, repeats of ongoing problems are not stated until 24 hours
  // after the *first* problem was encountered.

  // For example, if a fan goes out it gets reported right away 
  // (fans are immune to fan groups) and the time is noted.
  // If a second fan goes out, only that gets reported.
  // 24 hours after the first fan went out, both fans get reported
  // if they are still an issue.

  // BUG:  there may be a bug in the granularity of the 
  // state machine that counts and times out multiple measurements.
  // this may result in separate messages for each separate problem,
  // even if email_send correctly combines messages accumulated
  // over one sweep.  SO WHAT.

  // Possible correction:  accumulate potential messages over 
  // maximum of 4 sweeps

  // combine messages from one scan
  // supress duplicate messages (except as occasional reminders)

  long thetime;
  thetime = mygetcurtime (  );

  // now check the time and see if it is expired:
  pt_dbug2(2,__LINE__, "email_restrict newhashvalue:", newhashvalue, 0 );

  if ( ( thenode->last_email_time != 0 )        //has an email about this node been sent?
       && ( ( thenode->last_email_time + emailrepeat ) < thetime ) ) {  // emailrepeat time has passed?
    // time's up... lift all restrictions on messages:
    thenode->last_email_time = 0;
    thenode->email_hash = 0;
  }
  // compare hash value with previous value to node, and check/set time:
  if ( ( newhashvalue & ( thenode->email_hash ) ) == 0 ) {
    // this is a new message or an urgent one.  set the time if necessary
    // and send it.

    if ( newhashvalue != 0 ) {
      // nonurgent message.  set time and update hash value.

      thenode->email_hash |= newhashvalue;      // mark the message as sent.

      if ( thenode->last_email_time == 0 ) {
        thenode->last_email_time = thetime;     // mark the time.
      }
    }
    // queue the message for sending now.
    return ( 1 );
  } else {
    pt_dbug2(2,__LINE__, "email_restrict DID NOT SEND EMAIL newhashvalue:",
               newhashvalue, 0 );
    pt_dbug2(2,__LINE__, "email_restrict DID NOT SEND EMAIL CLUSTER:",
               thenode->cardaddress, 0 );

    return ( 0 );               // do NOT send the message at this time.
  }
}

//
void email_send (  )
{
  // Check on threads and consider sending accumulated msg

  // 1.  check that it's OK to send: 

  // It becomes time to send the email when 
  // (a sweep is done or when time goes by) and after the
  // previous email has been sent, and when there is content to send.

  // If it is time to send:
  // 2.  add a dot on a line by itself.
  // 3.  transfer email_buf_start (and email_mem_total?) to the_email_body
  // 4.  then set up new email_buf_start for future messages
  // 5.  then spin off thread to send email.

  pt_dbug2(2,__LINE__, "email_send", 0, 0 );

  if ( ( ( the_emailmsg.status_flag == ALERT_STATUS_IDLE )      // the thread is done
         //  || (email_status_flag == ALERT_STATUS_DONE)
         // Done is NOT an option!  have to wait for memory to be freed!
        ) && ( email_mem < email_mem_total ) )  // There is content to be sent.
    //??? Does it matter whether the thread is idle vs. done??
    //??? IS PREVIOUS SENT MESSAGE FREED FROM MEMORY?
    //??? MAYBE DONE MEANS MONITOR DID NOT YET DO THAT!
  {
    // Okay, time to send:

    // add a dot on a line by itself to the end of the buffer:
    email_alloc ( 4 );
    while ( email_build ( snprintf ( email_buf, email_mem, "\n.\n" ) ) );

    the_emailmsg.body = email_buf_start;

    //email_init (  );            // allocates scratchpad space for building emails.
    ////////////////// above line was the bug...
    // Finally, we can spin off the thread:

    email_transmit_time = mygetcurtime (  ) + stalemail;
    // mark time when thread started.
    the_emailmsg.status_flag = ALERT_STATUS_SPUN;
    if ( pthread_create ( &email_thread, NULL, ( void * ) nodewatch_email,
                          &the_emailmsg ) == 0 ) {
      //??? DO I HAVE TO SET ANY STATES HERE TO MARK THE FACT
      //??? THAT THE THREAD WAS ACTUALLY STARTED AT THIS POINT?
      pthread_detach ( email_thread );
    }

    else {
      // error message.
      error_msg ( "Unable to spin off email thread.\n" );
      the_emailmsg.status_flag = ALERT_STATUS_IDLE;
      email_init (  );
      //free ( the_emailmsg.body );
    }


  }
}

//
//??? not finished
void email_mon (  )
{
  // Check on any email thread, and cancel if overtime
  // assume this function is being called periodically 

  long thetime;
  char mon_error_msg[400];
  switch ( the_emailmsg.status_flag ) {

     case ALERT_STATUS_IDLE:
       break;                   // all is OK.

     case ALERT_STATUS_DONE:
       pt_dbug2(2,__LINE__, "email_mon found ALERT_STATUS_DONE", 0, 0 );
       email_init (  );
       //free ( the_emailmsg.body );     // this will free the memory AND reallocate it.
       //      free(email_buf_start);  //??? SHOULD THIS BE DONE HERE???
       //??? SHOULD EMAIL_INIT BE CALLED HERE?
       the_emailmsg.status_flag = ALERT_STATUS_IDLE;
       break;                   // reset everything
       // Note: this had to be done here so that it is confirmed that the
       // thread has exited.?

     case ALERT_STATUS_ERROR:

       // what else can be done here?
       // try to send the message multiple times?
       // would need a counter not reset by email_send,
       // then would need to call email_send again.
       //
       // the error was presumably documented in error log.
       error_msg ( "Email will not be re-attempted.\n" );
       email_init (  );
       //free ( the_emailmsg.body );
       // free(email_buf_start);
       the_emailmsg.status_flag = ALERT_STATUS_IDLE;
       break;

     case ALERT_STATUS_SPUN:
     case ALERT_STATUS_BUSY:

       thetime = mygetcurtime (  );
       if ( thetime > email_transmit_time ) {
         // the email has timed out.  Cancel it immediately.
         pthread_cancel ( email_thread );
         if ( the_emailmsg.status_flag == ALERT_STATUS_BUSY )
           error_msg ( "Email message timed out and not sent.\n" );
         else
           error_msg
             ( "Email message thread never got started and timed out.\n" );
         //??? fix error message?
         //??? should the email be retransmitted or not?  
         //
         email_init (  );
         //free ( the_emailmsg.body );
         //free(email_buf_start);

         the_emailmsg.status_flag = ALERT_STATUS_IDLE;
       }
       break;
     default:
       sprintf ( mon_error_msg,
                 "This area of code (limits.c ~line 2840) should never be reached\n"
                 "status flag: %u \n", the_emailmsg.status_flag );
       error_msg ( mon_error_msg );
       exit ( 2 );
       break;
  }
}

//email_buf_start is the entire body that is yet to be sent
//email_buf is a pointer into the end of email_buf_start- where new info should be added
//
//??? not finished
void email_init (  )
{
  // initially set up variables for email system.  
  // allocate initial memory for messages, set up email address pointers.


  //??? assume the values have been brought in from the files!
  //???     the_emailmsg.server = ;
  //??? subject?  is that subject to change?

  pt_dbug2(2,__LINE__, "email_init", 0, 0 );

  the_emailmsg.addresses = email_addresses;
  the_emailmsg.server = email_server;
  the_emailmsg.source_email = email_sender;
  the_emailmsg.subject = ( char * ) email_default_subject;

  if ( email_buf_start != NULL )
    free ( email_buf_start );   //???  SHOULD THIS BE DONE HERE?
  email_buf_start = malloc ( EMAIL_INITIAL_BUF * sizeof ( char ) );
  if ( email_buf_start == NULL ) {
    error_msg ( "Could not allocate memory for email\n" );
    email_mem_total = 0;
    email_mem = 0;
    return;
    //??? what should be done when mem not alloc?  i think exit!
  }
  email_buf = email_buf_start;  // start filling in messages at the beginning.
  email_mem_total = EMAIL_INITIAL_BUF;
  email_mem = EMAIL_INITIAL_BUF;

}

//
      /////////////////////////////////////CURRENT:

      // No matter how much memory we assign to deal with multiple 
      // messages, or how long we wait to allow messages to be combined
      // there can always come a message while the previous message
      // is in the process of being sent.

      // What is needed here is two email buffers.
      // one is the email being sent, and the other
      // is the email being assembled.
      // emails are assembled by storing a linked list of strings.
      // emails have no length limit and can be composed of
      // arbitrary number of segments.

      // It becomes time to send the email when 
      // (a sweep is done or when time goes by) and after the
      // previous email has been sent, and when there is content to send.

      // The linked list of strings are possibly sorted and
      // reassembled into a single string, the email 
      // info resource is freed for use by the next email, the time
      // is noted, and the email thread is started.

      // How to make just a linked list of strings?  Does this
      // require allocating the string and separately the pointer
      // to the string and the pointer to the next link?
      // is there a way to allocate the string itself 
      // at the same time as the pointer to it?

      // Well, where does the segment come from?  it comes 
      // from the asprintfs when the segment was made.
      // so either copy, reallocate, or allocate a second segment.
      // -------------:
      // or create a structure with 16 string pointers in it, the 
      // last of which points to another such structure!!!

      // To review:  there should -never- be any arbitrary 
      // size limits or restrictions, or risk of seg faulting,
      // no matter how extremely remote the restriction seems for now.

      // The algorithm continues to collect new messages while checking on the email thread.

      //??? what happens if the email is not sent?
      //??? should it try again with previous email?

      // Here are my options for allocating memory for email messages:
      // 1.  spin off a separate thread for each segment (individual error)
      //     - requires thread database + management loop + allocating memory

      // 2.  queue up a separate pending email for each segment
      //     - requires queue of separate emails allocated.

      // 3.  have 2 emails, one being sent and one being accumulated
      //     with linked lists of segments.

      // 4.  have 2 emails, one being sent and one being accumulated
      //     with a growing memory region.
      // every time a segment comes in, it decides whether to send it.
      // if it wants to send it, it sets up the appropriate
      // variables (body, information) and spins off a thread.
      // 
      // If it does not want to send it, it notes the size of the 
      // new segment and the size of the existing email (if any)
      // it then either appends (copies) the new segment and adjusts
      // the size values, or it first reallocs the message space.
      //
      // since we never know how often email_send can be called, we need
      // a separate fcn to check on the email thread regardless
      // of additional messages wanting to be sent.
      //
      // VARIABLES NEEDED:
      //
      // for the email being sent:
      //  from, to, body strings.
      //  structure with email status etc.
      //  value/handle to the thread created, or null if none active.
      //  count of how many times tried to send it.
      //
      // for the upcoming email:
      //  pointer to body string.
      //  amount of room left in body string. (watch out for off-by-one)
      //

//
// function meant to work on a single node
void action_limits ( CLUSTERNODE * thenode, char powertruth )
{
  // sends emails or initiates shutdowns based on state in node.
  // assumes all delays and power checks have been performed.
  // emails are sent using email_send() which accumulates the email and
  // opens a thread
  CHASSIS *chas;
  long thetime;
  int i;
  thetime = mygetcurtime (  );
// this was in below if statement as an OR to == POWER_SUPPLY_ON         || ( powertruth == 0 ) 
  if ( ( thenode->power_supply_state  == POWER_SUPPLY_ON )
       && (thenode->power_cmd_state == POWER_CMD_IDLE_ON)
       && (noalertshutdown == 0)
		 && ( (thetime - thenode->power_supply_state_time ) >  POWER_ON_STABLE_MEASUREMENT ) )
    // ALSO CHECK FOR POWER_CMD_STATE???? WHAT IF IT IS SHUTTING DOWN?

  {
    chas = thenode->chassis;

    //???? HERE IS WHERE THE POWER MUST BE CHECKED AND REACTED TO.
    pt_dbug2(2,__LINE__, "action_limits powertruth", powertruth, thenode );
    for ( i = 0; i < chas->numfans ; i++ )
      // REMINDER tackarray only used when accessing actual measurements.
    {
      switch ( thenode->fan_limit_new_state[i] ) {
           // FOR EACH OF THESE, SEND EMAIL and possibly INIT SHUTDOWN NOW.

         case NEW_LIM_ALERT_LO_WAITING_PWR:
           pt_dbug2(2,__LINE__, "rww alert_email_send fanalertlo", chas->fanalertlo[i], NULL );
           alert_email_send ( thenode, LIM_ALERT, LIM_LO, LIM_FAN, i,
                              chas->fanlabels[i],
                              ( int ) thenode->nwdata.uaRpm[chas->tackarray[i]],
                              ( int ) chas->fanalertlo[i],
                              ( int ) chas->fanalerthi[i] );

           thenode->fan_limit_new_state[i] = NEW_LIM_IDLE;
           // thenode->fan_limit_recent_state[i] = RECENT_LIM_ALERT_LO;
           //??? above two lines may have been obsoleted by the email hash code number.

           break;


         case NEW_LIM_ALERT_HI_WAITING_PWR:

           pt_dbug2(2,__LINE__, "rww alert_email_send fanalerthi", chas->fanalerthi[i], NULL );
           alert_email_send ( thenode, LIM_ALERT, LIM_HI, LIM_FAN, i,
                              chas->fanlabels[i],
                              thenode->nwdata.uaRpm[chas->tackarray[i]],
                              chas->fanalertlo[i],
                              chas->fanalerthi[i] );

           thenode->fan_limit_new_state[i] = NEW_LIM_IDLE;
           // thenode->fan_limit_recent_state[i] = RECENT_LIM_ALERT_HI;
           //??? above two lines may have been obsoleted by the email hash code number.

           break;



         case NEW_LIM_SHUTDOWN_LO_WAITING_PWR:

           // strategy is to send email, and configure state[i] to 
           // perform shutdown.

           pt_dbug2(2,__LINE__, "rww auto_shutdown_node fanshutdownlo", chas->fanshutdownlo[i], NULL );
           alert_email_send ( thenode,
                              LIM_SHUTDOWN,
                              LIM_LO,
                              LIM_FAN,
                              i,
                              chas->fanlabels[i],
                              thenode->nwdata.uaRpm[chas->tackarray[i]],
                              chas->fanshutdownlo[i],
                              chas->fanshutdownhi[i] );

           thenode->fan_limit_new_state[i] = NEW_LIM_IDLE;
           // thenode->fan_limit_recent_state[i] = RECENT_LIM_SHUTDOWN_LO;
           //??? above two lines may have been obsoleted by the email hash code number.

           auto_shutdown_node ( thenode );      // set up for that node to go down.

           break;



         case NEW_LIM_SHUTDOWN_HI_WAITING_PWR:

           // strategy is to send email, and configure state[i] to 
           // perform shutdown.

           pt_dbug2(2,__LINE__, "rww auto_shutdown_node fanshutdownhi", chas->fanshutdownhi[i], NULL );
           alert_email_send ( thenode,
                              LIM_SHUTDOWN,
                              LIM_HI,
                              LIM_FAN,
                              i,
                              chas->fanlabels[i],
                              thenode->nwdata.uaRpm[chas->tackarray[i]],
                              chas->fanshutdownlo[i],
                              chas->fanshutdownhi[i] );

           thenode->fan_limit_new_state[i] = NEW_LIM_IDLE;
           //          thenode->fan_limit_recent_state[i] = RECENT_LIM_SHUTDOWN_HI;
           //??? above two lines may have been obsoleted by the email hash code number.

           auto_shutdown_node ( thenode );      // set up for that node to go down.

           break;



      }


    }

    for ( i = 0; i < MAX_TEMP; i++ ) {
      switch ( thenode->thermo_limit_new_state[i] ) {
           // FOR EACH OF THESE, SEND EMAIL and possibly INIT SHUTDOWN NOW.

         case NEW_LIM_ALERT_LO_WAITING_PWR:

           pt_dbug2(2,__LINE__, "rww alert_email_send tempalertlo", chas->tempalertlo[i], NULL );
           alert_email_send ( thenode, LIM_ALERT, LIM_LO, LIM_THERMO, i,
                              chas->templabels[i],
                              thenode->nwdata.uaA2D[chas->thermoarray[i]],
                              chas->tempalertlo[i],
                              chas->tempalerthi[i] );

           thenode->thermo_limit_new_state[i] = NEW_LIM_IDLE;
           //          thenode->thermo_limit_recent_state[i] = RECENT_LIM_ALERT_LO;
           //??? above two lines may have been obsoleted by the email hash code number.

           break;


         case NEW_LIM_ALERT_HI_WAITING_PWR:

           pt_dbug2(2,__LINE__, "rww alert_email_send tempalerthi", chas->tempalerthi[i], NULL );
           alert_email_send ( thenode, LIM_ALERT, LIM_HI, LIM_THERMO, i,
                              chas->templabels[i],
                              thenode->nwdata.uaA2D[chas->thermoarray[i]],
                              chas->tempalertlo[i],
                              chas->tempalerthi[i] );

           thenode->thermo_limit_new_state[i] = NEW_LIM_IDLE;
           //          thenode->thermo_limit_recent_state[i] = RECENT_LIM_ALERT_LO;
           //??? above two lines may have been obsoleted by the email hash code number.

           break;



         case NEW_LIM_SHUTDOWN_LO_WAITING_PWR:

           pt_dbug2(2,__LINE__, "rww auto_shutdown_node tempshutdownlo", chas->fanshutdownlo[i], NULL );
           // strategy is to send email, and configure state[i] to 
           // perform shutdown.

           alert_email_send ( thenode, LIM_SHUTDOWN, LIM_LO, LIM_THERMO, i,
                              chas->templabels[i],
                              thenode->nwdata.uaA2D[chas->thermoarray[i]],
                              chas->tempshutdownlo[i],
                              chas->tempshutdownhi[i] );

           thenode->thermo_limit_new_state[i] = NEW_LIM_IDLE;
           //          thenode->thermo_limit_recent_state[i] = RECENT_LIM_SHUTDOWN_LO;
           //??? above two lines may have been obsoleted by the email hash code number.

           auto_shutdown_node ( thenode );      // set up for that node to go down.

           break;


         case NEW_LIM_SHUTDOWN_HI_WAITING_PWR:
           pt_dbug2(2,__LINE__, "rww auto_shutdown_node tempshutdownhi", chas->fanshutdownhi[i], NULL );

           // strategy is to send email, and configure state[i] to 
           // perform shutdown.

           alert_email_send ( thenode, LIM_SHUTDOWN, LIM_HI, LIM_THERMO, i,
                              chas->templabels[i],
                              thenode->nwdata.uaA2D[chas->thermoarray[i]],
                              chas->tempshutdownlo[i],
                              chas->tempshutdownhi[i] );

           thenode->thermo_limit_new_state[i] = NEW_LIM_IDLE;
           //          thenode->thermo_limit_recent_state[i] = RECENT_LIM_SHUTDOWN_HI;
           //??? above two lines may have been obsoleted by the email hash code number.

           auto_shutdown_node ( thenode );      // set up for that node to go down.

           break;


      }

    }

    for ( i = 0; i < MAX_VOLT; i++ ) {
      switch ( thenode->volt_limit_new_state[i] ) {
           // FOR EACH OF THESE, SEND EMAIL and possibly INIT SHUTDOWN NOW.

         case NEW_LIM_ALERT_LO_WAITING_PWR:
           pt_dbug2(2,__LINE__, "rww alert_email_send voltalertlo", chas->voltalertlo[i], NULL );

           alert_email_send ( thenode, LIM_ALERT, LIM_LO, LIM_VOLT, i,
                              chas->voltlabels[i],
                              thenode->nwdata.uaA2D[chas->voltarray[i]],
                              chas->voltalertlo[i],
                              chas->voltalerthi[i] );

           thenode->volt_limit_new_state[i] = ( char ) NEW_LIM_IDLE;
           //          thenode->volt_limit_recent_state[i] = RECENT_LIM_ALERT_LO;
           //??? above two lines may have been obsoleted by the email hash code number.

           break;


         case NEW_LIM_ALERT_HI_WAITING_PWR:

           pt_dbug2(2,__LINE__, "rww alert_email_send voltalerthi", chas->voltalerthi[i], NULL );
           alert_email_send ( thenode, LIM_ALERT, LIM_HI, LIM_VOLT, i,
                              chas->voltlabels[i],
                              thenode->nwdata.uaA2D[chas-> voltarray[i]],
                              chas->voltalertlo[i],
                              chas->voltalerthi[i] );

           thenode->volt_limit_new_state[i] = NEW_LIM_IDLE;
           //          thenode->volt_limit_recent_state[i] = RECENT_LIM_ALERT_HI;
           //??? above two lines may have been obsoleted by the email hash code number.

           break;


         case NEW_LIM_SHUTDOWN_LO_WAITING_PWR:

           pt_dbug2(2,__LINE__, "rww auto_shutdown_node voltshutdownlo", chas->voltshutdownlo[i], NULL );
           // strategy is to send email, and configure state[i] to 
           // perform shutdown.

           alert_email_send ( thenode, LIM_SHUTDOWN, LIM_LO, LIM_VOLT, i,
                              chas->voltlabels[i],
                              thenode->nwdata.uaA2D[chas->voltarray[i]],
                              chas->voltshutdownlo[i],
                              chas->voltshutdownhi[i] );

           thenode->volt_limit_new_state[i] = NEW_LIM_IDLE;
           //          thenode->volt_limit_recent_state[i] = RECENT_LIM_SHUTDOWN_LO;
           //??? above two lines may have been obsoleted by the email hash code number.

           auto_shutdown_node ( thenode );      // set up for that node to go down.

           break;


         case NEW_LIM_SHUTDOWN_HI_WAITING_PWR:
           // strategy is to send email, and configure state[i] to 
           // perform shutdown.

           pt_dbug2(2,__LINE__, "rww auto_shutdown_node voltshutdownhi", chas->voltshutdownhi[i], NULL );
           alert_email_send ( thenode, LIM_SHUTDOWN, LIM_HI, LIM_VOLT, i,
                              chas->voltlabels[i],
                              thenode->nwdata.uaA2D[chas->voltarray[i]],
                              chas->voltshutdownlo[i],
                              chas->voltshutdownhi[i] );

           thenode->volt_limit_new_state[i] = NEW_LIM_IDLE;
           //          thenode->volt_limit_recent_state[i] = RECENT_LIM_SHUTDOWN_HI;
           //??? above two lines may have been obsoleted by the email hash code number.

           auto_shutdown_node ( thenode );      // set up for that node to go down.

           break;

      }

    }
    // Here is where power detection is reset:  however, what about problems appearing on powerup?
    thenode->power_supply_state_count = 0;
    thenode->power_supply_state_time = 0;
    thenode->power_supply_state_timeout = 0;



  } else {
    for ( i = 0; i < MAX_TACK; i++ )
      thenode->fan_limit_new_state[i] = NEW_LIM_IDLE;

    for ( i = 0; i < MAX_TEMP; i++ )
      thenode->thermo_limit_new_state[i] = NEW_LIM_IDLE;

    for ( i = 0; i < MAX_VOLT; i++ )
      thenode->volt_limit_new_state[i] = NEW_LIM_IDLE;

  }
}

//
void alert_email_send ( CLUSTERNODE * clustersearch, char alert_shutdn, // is it alert(0) or shutdown(1) msg?
                        // for reference, if a shutdown, the hash ID 
                        // becomes 0, which means always send
                        // (There can't be two shutdowns of the same
                        // node because the node's shutdown state can only
                        // be marked for shutdown once.)
                        char lim_direction,     // is it too low(0) or too high(1)?
                        char measure_type,      // fans(0), thermo(1) or volts(2)?
                        int measure_index,      // index (pre-array) into physical measurement
                        // right now used to calculate email hash ID code number,
                        // which is needed to prevent duplicate email messages.
                        // could/should also be used to retrieve measure_label and values!
                        char *measure_label,    // the actual measurement label
                        int the_measurement,    // actual value before normalizing.
                        int lim_lo_val, // measurement alert lo limit
                        int lim_hi_val )        // measurement alert hi limit
     // formats an email message segment to report an over/underlimit situation.
{
  // First generate the hash and subject fields:
  unsigned long long thehash;


  if ( alert_shutdn == 1 )
    // this is a shutdown message. 
  {
    thehash = 0;                // force it to be sent.
    the_emailmsg.subject = ( char * ) email_shutdown_subject;
  } else {
    // build the hash from all the arguments:
    thehash = pow ( 2, ( 3 +    // 2 low bits for old data, no data and failed shutdown, see above.
                         lim_direction +        // 0,1
                         measure_type * 2 +     // 0,1,2
                         measure_index * 2 * 3  // +  //
                     ) );
    //     alert_shutdn  * 2 * 3 * 2   ));  // 53 bits consumed!
    //???  double check this calculation.
  };

  pt_dbug2(2,__LINE__, "alert_email_send thehash:", thehash, 0 );

  if ( email_restrict ( clustersearch, thehash ) ) {
    assert ( ( alert_shutdn == 0 ) || ( alert_shutdn == 1 ) );  //are we hitting an invalid array index?
    while
      ( email_build
        ( snprintf
          ( email_buf, email_mem,
            "%sNode %s has %s %s %s\n Current Reading: %7.1f %s.\n %s: %7.1f - %7.1f %s.\n\n",
            limit_alert_shutdown[( int ) alert_shutdn],
            clustersearch->ipaddress,
            limit_direction[( int ) lim_direction],
            limit_measurement_type[( int ) measure_type],
            measure_label,
            ( float ) the_measurement / limit_val_div[( int ) measure_type],
            limit_units[( int ) measure_type],
            limit_range_id[( int ) alert_shutdn],
            ( float ) lim_lo_val / limit_val_div[( int ) measure_type],
            ( float ) lim_hi_val / limit_val_div[( int ) measure_type],
            limit_units[( int ) measure_type] ) ) ) {
    };
  }
}


//
//??? not finished
void auto_shutdown_node ( CLUSTERNODE * thenode )
{
  // Called by action_limits to initiate an automatic shutdown
  // of a node.  This should start by initiating the o/s shutdown,
  // marking new state, and noting the time
  // 
  long thetime;
  thetime = mygetcurtime (  );


  //??? does this set the right state
  //??? DOES THE STATE OF IT GOING DOWN SOON/NOW
  //??? get considered by the other state machines
  //??? such as initial power detect?  it should!


  // sets the state to initiate auto shut down the node
  // starts process of shutting down the operating system.
  // sends out another email stating that the node is going out.
  // checks to see if this is the master going down.
  // if so, sets up all the nodes to go down,
  //???   bug:  will all the nodes have OS shutdown message in time before this service times out?
  //??? bug:  what about dual-redundant master systems?
  // they do not need to shutdown the cluster if the master is going
  // out and if the other master is ready to take over automatically...
  //??? bug:  what about the max number of threads allowed by the system?

  pt_dbug2(2,__LINE__, "auto_shutdown_node: ", thenode->cardaddress, thenode );
  if ( noshutdown == 0 ) {
    if ( thenode == NWMaster ) {
		 pt_dbug(__LINE__,"Master's special shutdown case not handled"," yet",NULL);
      //??? does not handle master correctly yet!
      //??? decide what to do, probably shut down the whole cluster.
    } else if ( thenode->power_cmd_state == POWER_CMD_IDLE_ON )
      // the primary state for being able to shut down.
    {
      // now start the shutdown in motion, set a timeout for it,
      // and set the power cmd state to reflect new state.


      // be sure to make state changes
      // before starting off any threads:

      thenode->power_cmd_state = POWER_CMD_SHUTDOWN_OS_TRY;
      thenode->power_cmd_delay = thetime;

      // now start off the thread:

      //??? do we need to check the value of status_flag before setting it?

      thenode->os_shutdown_status_flag = OS_SHUTDOWN_IDLE;


      pt_dbug(__LINE__,"About to spinoff pthread for safeshutdown"," now", NULL);
      if ( pthread_create ( &( thenode->os_shutdown_thread ), NULL,
                            ( void * ) safeShutdown, thenode ) == 0 ) {
        pt_dbug(__LINE__,"Just to spunoff pthread for safeshutdown"," about to detach\n", NULL);
        pthread_detach ( thenode->os_shutdown_thread );
        pt_dbug(__LINE__,"Just detached pthread for safeshutdown"," now\n", NULL);
		}


      else {
        // error message.
        error_msg ( "Unable to spin off o/s shutdown thread!\n" );

        thenode->power_cmd_state = POWER_CMD_SHUTDOWN_OS_MSG_FAIL;
      }


      //??? INITIATE SHUTDOWN O/S FOR THAT NODE.

      //??? RICK to write function --
      // I call as a new thread, giving it a pointer to the 
      // power_cmd_state byte where the results should be stored.
      // it tries to shut down the node's OS.  It keeps trying
      // till it either succeeds or fails.  
      // The timeout for powering off does not start until
      // it succeeds.  And if it fails, maybe poweroff happens immediately!





      //??? OKAY if this is where the o/s shutdown takes place,
      //??? where does the actual poweroff take place?
      //??? (there must be another state machine somewhere that
      //??? monitors power_cmd_state and moves things along!)
      //??? also if thread has difficulty what happens
      //??? also what about keeping tabs of separate threads?

      // spins off a thread to try to shutdown a node's OS, returning
      // immediately.
      // Thread reports back success or failure to deliver the shutdown
      // request over the network.          

    } else {
      pt_dbug2(2,__LINE__,
                 "auto_shutdown_node: not in idle state to begin with!",
                 thenode->cardaddress, thenode );
    }
  }
  pt_dbug(__LINE__,"Leaving auto_node_shutdown"," now", NULL);
}

//
//??? not finished  --  issue of what to do if failure. syncing up with other power state functions.
// meant to be passed head of cluster linked list
void shutdown_check ( CLUSTERNODE * thenode )
{
  // Checks for and monitors the progress of automatic shutdowns.
  // called at the end of genlimitscheck, once for each node.

  //??? in addition to checking for timing, it has to 
  //??? cancel os threads that are taking to long, as well as
  //??? reset the state of the os_shutdown_status_flag
  //??? and reconcile the os_shutdown_status_flag with the 
  //??? shutdown_cmd_state

  long thetime;
  thetime = mygetcurtime (  );

  switch ( thenode->power_cmd_state ) {

     case POWER_CMD_IDLE_ON:
       break;                   // nothing special.

     case POWER_CMD_IDLE_OFF:
       break;                   // nothing special.
       /// believe this state is obsolete.


       // nothing to do but to wait
       // for the os shutdown thread to 
       // change state to one of two following:


     case POWER_CMD_SHUTDOWN_OS_MSG_FAIL:
       // turn power off immediately on this node, set state, mark time:
       pt_dbug2(2,__LINE__, "shutdown_check MSG_FAIL", 0, 0 );
       pt_dbug2(2,__LINE__, "POWERING OFF NOW POWER_CMD_SHUTDOWN_OS_MSG_FAIL",
                  0, 0 );

       power_ctl_fromnode ( thenode, NODE_OFF );
       thenode->power_cmd_state = POWER_CMD_DO_POWEROFF;
       thenode->power_cmd_delay = thetime;

       break;

     case POWER_CMD_SHUTDOWN_OS_TRY:
       // the timeout should happen whether or not 
       // the message successfully sent:
       if ( thetime > ( thenode->power_cmd_delay +
                        ( thenode->chassis )->shutdown_os_time ) ) {
         pthread_cancel ( thenode->os_shutdown_thread );
         error_msg ( "OS shutdown thread timed out.\n" );
         // time to turn off power and move on to next step.
         pt_dbug2(2,__LINE__, "POWERING OFF NOW POWER_CMD_SHUTDOWN_OS_TRY", 0,
                    0 );

         power_ctl_fromnode ( thenode, NODE_OFF );      // in power_ctl.c
         thenode->power_cmd_state = POWER_CMD_DO_POWEROFF;
         thenode->power_cmd_delay = thetime;
       }
       break;

     case POWER_CMD_SHUTDOWN_OS_MSG_SENT:

       // Waiting for OS to die.  

       // separate function -- have to get thetime again.
       if ( thetime > ( thenode->power_cmd_delay +
                        ( thenode->chassis )->shutdown_os_time ) )
         //??? why use the shutdown time as a timeout for the messaging?
         //??? shouldn't it be a shorter timeout?
       {
         // time to turn off power and move on to next step.
         pt_dbug2(2,__LINE__, "shutdown_check MSG_SENT turning off power",
                    thenode->cardaddress, 0 );
         pt_dbug2(2,__LINE__,
                    "POWERING OFF NOW POWER_CMD_SHUTDOWN_OS_MSG_SENT", 0, 0 );

         power_ctl_fromnode ( thenode, NODE_OFF );      // in power_ctl.c
         thenode->power_cmd_state = POWER_CMD_DO_POWEROFF;
         thenode->power_cmd_delay = thetime;
       }
       break;


     case POWER_CMD_DO_POWEROFF:
       // one question: how long should we wait 
       // before checking the success?
       // it's in POWEROFF_DELAY, a function of the network and the button.
       // check success later

       if ( thetime > ( thenode->power_cmd_delay + POWEROFF_DELAY ) ) {
         // check to see if it was successful.
         if ( thenode->power_supply_state == POWER_SUPPLY_OFF ) {
           thenode->power_cmd_state = POWER_CMD_CONFIRM_POWEROFF;
           thenode->power_cmd_delay = thetime;
         }

         else {
           thenode->power_cmd_state = POWER_CMD_POWEROFF_FAIL;
           thenode->power_cmd_delay = thetime;
         }

       }


       break;

     case POWER_CMD_CONFIRM_POWEROFF:
       //???  what yet needs to be done?
       // Wait for the power to come back on manually, and then reset the state.
       if ( thenode->power_supply_state == POWER_SUPPLY_ON ) {
         thenode->power_cmd_state = POWER_CMD_IDLE_ON;
         thenode->power_cmd_delay = thetime;
       }
       break;

       //??? what about after the problem is addressed.
       //??? does the software need to be reset in order to recognize
       //??? the node again???

       //??? what about notifying the user or the website of the action taken?


     case POWER_CMD_POWEROFF_FAIL:
       //???  ...what should be done here?
       // alert the user via email, using a restrictive code to prevent spam..
       // then reset the state, to allow the mechanism to retry if it wants.


       if ( email_restrict ( thenode, 4 ) )
         while
           ( email_build
             ( snprintf
               ( email_buf, email_mem,
                 "Node %s refuses to poweroff automatically.\n",
                 thenode->ipaddress ) ) );

       thenode->power_cmd_state = POWER_CMD_IDLE_ON;
       thenode->power_cmd_delay = thetime;
       break;
  }



}

//
void chassis_init_more ( CHASSIS * thechassis )
{
  thechassis->measure_confirm_count = 4;
  thechassis->measure_confirm_time = 20;
  thechassis->shutdown_os_time = 40;
  thechassis->powerswitch_off_time = 6;
  thechassis->power_valid_count = 3;
  thechassis->power_valid_time = 10;
  thechassis->power_valid_timeout = 20;
  thechassis->volt_powerstate_level[0] = 200;   // 12 volts below  2 volts
  thechassis->volt_powerstate_level[1] = 100;   //  5 volts below  1 volts
  thechassis->volt_powerstate_level[2] = 070;   //3.3 volts below .7 volts

}

//
void cluster_init_more ( CLUSTERNODE * thenode )
{
  // by using calloc to allocate memory, all unallocated values become zero.
  int i;
  thenode->power_supply_state = 0;      // default is the machine is off.

  for ( i = 0; i < MAX_TEMP; i++ )
    thenode->thermo_limit_new_state[i] = NEW_LIM_IDLE;
  for ( i = 0; i < MAX_VOLT; i++ )
    thenode->volt_limit_new_state[i] = NEW_LIM_IDLE;
  for ( i = 0; i < MAX_TACK; i++ )
    thenode->fan_limit_new_state[i] = NEW_LIM_IDLE;

  thenode->power_cmd_state = POWER_CMD_IDLE_ON;
  //???  is POWER_CMD_IDLE_ON right?   how does it transition between
  //??? IDLE_ON and IDLE_OFF?
  thenode->os_shutdown_thread = 0;
  thenode->os_shutdown_status_flag = OS_SHUTDOWN_IDLE;

}

void cluster_reset_limits ( CLUSTERNODE * thenode )
{

  int i;

  for ( i = 0; i < MAX_TEMP; i++ ) {
    thenode->thermo_limit_new_state[i] = NEW_LIM_IDLE;

    thenode->thermo_shutdown_lo_count[i] = 0;
    thenode->thermo_shutdown_hi_count[i] = 0;
    thenode->thermo_alert_lo_count[i] = 0;
    thenode->thermo_alert_hi_count[i] = 0;
  }

  for ( i = 0; i < MAX_VOLT; i++ ) {
    thenode->volt_limit_new_state[i] = NEW_LIM_IDLE;

    thenode->volt_shutdown_lo_count[i] = 0;
    thenode->volt_shutdown_hi_count[i] = 0;
    thenode->volt_alert_lo_count[i] = 0;
    thenode->volt_alert_hi_count[i] = 0;

  }

  for ( i = 0; i < MAX_TACK; i++ ) {
    thenode->fan_limit_new_state[i] = NEW_LIM_IDLE;

    thenode->fan_shutdown_lo_count[i] = 0;
    thenode->fan_shutdown_hi_count[i] = 0;
    thenode->fan_alert_lo_count[i] = 0;
    thenode->fan_alert_hi_count[i] = 0;
  }

  // here is where the power detection is reset:
  thenode->power_supply_state_count = 0;
  thenode->power_supply_state_time = 0;
  thenode->power_supply_state_timeout = 0;

}

//

/*          **********
	    Generic method for looping thru nodes:
	    {
	    
  CHASSIS *chas; // temporary pointer to the chassis.
  CLUSTERNODE *clustersearch; // pointer going through the list of nodes.
  CLUSTERNODE *nextsearch; // pointer going through the next node.
  
  
  while (clustersearch != NULL)
    {
      chas = clustersearch->chassis;
      
      
      
      nextsearch = clustersearch->nextnode;
      clustersearch = nextsearch; // update the search regardless if null or
                                  // not.  While loop needs to stop someday.
}

               ********** */
