#!/bin/bash
# Script written by Richard Warner, Microway Inc., 4/2003

rwwerror() {
   echo $1
   bash
   exit 1
}

usage() {
  echo "Usage: " `basename $0` "<config file to be used during restore>"
  if [ x"$1" = "x" ] ; then
   exit 2
  else
   exit 1
  fi
}

if [ -f /var/lock/prepare-restore ] ; then
  echo "Failed to prepare image.  One copy of prepare-restore already running."
  exit 6
fi

if [ $# -ne 2 ] ; then
  echo "Invalid parameters."
  usage
fi

if [ ! -f $1 ] ; then
  echo "$1 does not exist"
  usage
fi

if [ $UID -ne 0 ] ; then
  echo "You must be root to run this."
  exit 5
fi

  touch /var/lock/prepare-restore

tempdir=`/bin/mktemp -d /tmp/loopXXXXXX`
echo "Created temp directory $tempdir"
echo "Extracting initrd to $tempdir"
cd $tempdir
zcat /tftpboot/initrd.img.gz | cpio -id &>/dev/null
echo "Copying current config file into initrd image"
cp -f $1 .
find . | cpio --quiet -H newc -o | gzip -9 > /tftpboot/initrd.img.gz
cd - >&/dev/null
rm -rf $tempdir

echo "Removing lockfiles"
rm /var/lock/mcms-nodebackup-$2
rm /var/lock/prepare-restore
